Unit BPNav;

{ Class TBPNavigator }
{ Rewritten to ScanLines 27.01.03 }

Interface

Uses Windows,SysUtils,Classes,Controls,Graphics,Forms;

Type  TBPColor=packed record { TColor representation }
       r:Byte;
       g:Byte;
       b:Byte;
       a:Byte;
      end;

      TMyColor=packed record { 24-bit bitmap entry }
        r:Byte;
        g:Byte;
        b:Byte;
      end;

      PMyColorArray=^TMyColorArray;
      TMyColorArray=Array[0..1600] of TMyColor;

      TAlphaArray=Array[0..255] of Byte;


      TBPNavigator=class(TCustomControl)
        private
         FNavigateObject:TScrollingWinControl;
         FNavigateBitmap:TBitmap;
         FOutputBitmap:TBitmap;
         FNavigateFullWidth:Integer;
         FNavigateFullHeight:Integer;
         FNavigateClientWidth:Integer;
         FNavigateClientHeight:Integer;
         FNavigatePosLeft:Integer;
         FNavigatePosTop:Integer;
         FRatio:Double; { Scale Ratio }
         FBackColor:TColor;
         FReEnter:Boolean;
         FMyDrag:Boolean;
         FOldX:Integer;
         FOldY:Integer;
         FOldCursor:TCursor;
         FAlpha:Double;
         FShadowColor:TColor;
         FBitmapRect:TRect;
         FRamkaRect:TRect;
         FAlphaAr:TAlphaArray;
         F1MinusAlphaAr:TAlphaArray;
         procedure SetNavigateObject(Value:TScrollingWinControl);
         procedure MyMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
         procedure MyMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
         procedure MyMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
         procedure SetBackColor(Value:TColor);
         procedure CreateOutputBitmap(x,y,x1,y1:Integer);
         procedure SetAlpha(Value:Double);
         procedure SetShadowColor(Value:TColor);
         procedure CreateAlphaArray;
        protected
         procedure Paint; override;
        public
        {}
         procedure PaintImmediate;
         constructor Create(AOwner: TComponent); override;
         destructor  Destroy; override;
         procedure ScrollBarChanged(Sender:TObject);
         procedure ClearBitmap;
         property Bitmap:TBitmap read FNavigateBitmap;
         property OnMouseDown;
         property OnMouseMove;
         property OnMouseUp;
         property ReEnter:Boolean read FReEnter write FReEnter default False;
        published
         property NavigateObject:TScrollingWinControl read FNavigateObject write SetNavigateObject;
         property BackColor:TColor read FBackColor write SetBackColor default clBackground;
         property Align;
         property ShowHint;
         property Alpha:Double read FAlpha write SetAlpha;
         property ShadowColor:TColor read FShadowColor write SetShadowColor;
      end;

procedure Register;

Implementation

procedure TBPNavigator.CreateAlphaArray; { Checked }
Var i:Integer;
 begin
   for i:=0 to 255 do FAlphaAr[i]:=Round(i*FAlpha);
   for i:=0 to 255 do F1MinusAlphaAr[i]:=Round(i*(1-FAlpha));
 end;

procedure TBPNavigator.ClearBitmap;
 begin
   FNavigateBitmap.Canvas.Brush.Color:=clWhite;
   FNavigateBitmap.Canvas.FillRect(FNavigateBitmap.Canvas.ClipRect);
 end;

procedure TBPNavigator.SetShadowColor(Value:TColor);
 begin
   FShadowColor:=Value;
   Paint;
 end;

procedure TBPNavigator.SetAlpha(Value:Double);
 begin
   FAlpha:=Value;
   if FAlpha<0 then FAlpha:=0;
   if FAlpha>1 then FAlpha:=1;
   CreateAlphaArray;
   Paint;
 end;

procedure TBPNavigator.CreateOutputBitmap(x,y,x1,y1:Integer);
Var i,j:Integer;
    w,h:Integer;
    y2,y3,x2,x3:Integer;
    var col1:TBPColor;
         col:TMyColor;
    SLP,DLP:PMyColorArray;
 begin
   x2:=x;
   if (x2<0) then x2:=0;
   y2:=y;
   if (y2<0) then y2:=0;
   w:=FOutputBitmap.Width-1;
   h:=FOutputBitmap.Height-1;
   x3:=x1;
   y3:=y1;
   if x3>w then x3:=w;
   if y3>h then y3:=h;
   { Copy Navigate to Output Bitmap }
   for j:=0 to h do
     begin
       SLP:=FNavigateBitmap.ScanLine[j];
       DLP:=FOutputBitmap.ScanLine[j];
       System.Move(SLP^,DLP^,w*sizeof(TMyColor));
     end;
   { Draw Alpha Shadow }
   col1:=TBPColor(FShadowColor);
   with FOutputBitmap do
    begin
      for j:=y2 to y3 do
        begin
          SLP:=ScanLine[j];
           for i:=0 to (x2-1) do
             begin
             col:=SLP^[i];
             col.b:=FAlphaAr[col.b]+F1MinusAlphaAr[col1.b];
             col.r:=FAlphaAr[col.r]+F1MinusAlphaAr[col1.r];
             col.g:=FAlphaAr[col.g]+F1MinusAlphaAr[col1.g];
             SLP^[i]:=col;
             end;
        end;
      for j:=y2 to y3 do
        begin
          SLP:=ScanLine[j];
          for i:=(x3+1) to w do
           begin
             col:=SLP^[i];
             col.b:=FAlphaAr[col.b]+F1MinusAlphaAr[col1.b];
             col.r:=FAlphaAr[col.r]+F1MinusAlphaAr[col1.r];
             col.g:=FAlphaAr[col.g]+F1MinusAlphaAr[col1.g];
             SLP^[i]:=col;
           end;
        end;
      for j:=0 to (y2-1) do
       begin
           SLP:=ScanLine[j];
           for i:=0 to w do
            begin
             col:=SLP^[i];
             col.b:=FAlphaAr[col.b]+F1MinusAlphaAr[col1.b];
             col.r:=FAlphaAr[col.r]+F1MinusAlphaAr[col1.r];
             col.g:=FAlphaAr[col.g]+F1MinusAlphaAr[col1.g];
             SLP[i]:=col;
            end;
        end;
      for j:=(y3+1) to h do
       begin
         SLP:=ScanLine[j];
         for i:=0 to w do
          begin
             col:=SLP^[i];
             col.b:=FAlphaAr[col.b]+F1MinusAlphaAr[col1.b];
             col.r:=FAlphaAr[col.r]+F1MinusAlphaAr[col1.r];
             col.g:=FAlphaAr[col.g]+F1MinusAlphaAr[col1.g];
             SLP^[i]:=col;
          end;
       end;
   end;
 end;

procedure TBPNavigator.MyMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
Var x1,y1:Integer;
    hw,hh:Integer;
 begin
   if Button=mbLeft then
     begin
       if FNavigateObject=nil then Exit;
       FOldX:=X;
       FOldY:=Y;
       FMyDrag:=True;
       FOldCursor:=Cursor;
       Cursor:=crDrag;
       Application.ProcessMessages;
     end;
   if Button=mbRight then
     begin
       if FNavigateObject=nil then Exit;
       hw:=((FRamkaRect.Right-FRamkaRect.Left) div 2);
       hh:=((FRamkaRect.Bottom-FRamkaRect.Top) div 2);
       if (x<FBitmapRect.Left) then Exit;
       if (x>FBitmapRect.Right) then Exit;
       if (y<FBitmapRect.Top) then Exit;
       if (y>FBitmapRect.Bottom) then Exit;
       x1:=Round((X-FBitmapRect.Left-FRamkaRect.Left-hw)/FRatio);
       y1:=Round((Y-FBitmapRect.Top-FRamkaRect.Top-hh)/FRatio);
      if (x1<>0) then
       FNavigateObject.HorzScrollBar.Position:=FNavigateObject.HorzScrollBar.Position+
            x1;
      if (y1<>0) then
       FNavigateObject.VertScrollBar.Position:=FNavigateObject.VertScrollBar.Position+
              y1;
     end;
 end;

procedure TBPNavigator.MyMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
var x1,y1:Integer;
    x2,y2:Integer;
 begin
   if not(FMyDrag) then Exit;
   if FNavigateObject=nil then Exit;
   x1:=Round((X-FOldX)/FRatio);
   y1:=Round((Y-FOldY)/FRatio);
   if (x1<>0) then
     FNavigateObject.HorzScrollBar.Position:=FNavigateObject.HorzScrollBar.Position+
              x1;
   if (y1<>0) then
     FNavigateObject.VertScrollBar.Position:=FNavigateObject.VertScrollBar.Position+
              y1;
   FOldX:=X;
   FOldY:=Y;
 end;

procedure TBPNavigator.MyMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
 begin
   FMyDrag:=False;
   Cursor:=FOldCursor;
   Application.ProcessMessages;
 end;

procedure TBPNavigator.PaintImmediate;
 begin
   FReEnter:=False;
   Paint;
 end;

procedure TBPNavigator.Paint;
Var r1,r2,r3,r4:TRect;
    w,h:Integer;
 begin
    if FReEnter then Exit;
    FReEnter:=True;
    FBitmapRect.Left:=(ClientWidth-Round(FNavigateFullWidth*FRatio)) div 2;
    FBitmapRect.Top:=(ClientHeight-Round(FNavigateFullHeight*FRatio)) div 2;
    FBitmapRect.Right:=FBitmapRect.Left+FNavigateBitmap.Width;
    FBitmapRect.Bottom:=FBitmapRect.Top+FNavigateBitmap.Height;
    FRamkaRect.Left:=Round(FNavigatePosLeft*FRatio);
    FRamkaRect.Top:=Round(FNavigatePosTop*FRatio);
    FRamkaRect.Right:=FRamkaRect.Left+Round(FNavigateClientWidth*FRatio);
    FRamkaRect.Bottom:=FRamkaRect.Top+Round(FNavigateClientHeight*FRatio);
    w:=ClientWidth;
    h:=ClientHeight;
    { Top }
    r1.Left:=0;
    r1.Right:=w;
    r1.Top:=0;
    r1.Bottom:=FBitmapRect.Top;
    { Right }
    r2.Left:=FBitmapRect.Right;
    r2.Right:=w;
    r2.Top:=FBitmapRect.Top;
    r2.Bottom:=FBitmapRect.Bottom;
    { Left }
    r3.Left:=0;
    r3.Right:=FBitmapRect.Left;
    r3.Top:=FBitmapRect.Top;
    r3.Bottom:=FBitmapRect.Bottom;
    { Bottom }
    r4.Left:=0;
    r4.Right:=w;
    r4.Top:=FBitmapRect.Bottom;
    r4.Bottom:=h;
    with Canvas do
    begin
     Brush.Color:=FBackColor;
     FillRect(r1);
     FillRect(r2);
     FillRect(r3);
     FillRect(r4);
     CreateOutputBitmap(FRamkaRect.Left,FRamkaRect.Top,FRamkaRect.Right,FRamkaRect.Bottom-1);
     Draw(FBitmapRect.Left,FBitmapRect.Top,FOutputBitmap);
     Pen.Color:=clBlack;
     MoveTo(FBitmapRect.Left+FRamkaRect.Left,FRamkaRect.Top+FBitmapRect.Top);
     LineTo(FBitmapRect.Left+FRamkaRect.Right,FRamkaRect.Top+FBitmapRect.Top);
     LineTo(FBitmapRect.Left+FRamkaRect.Right,FRamkaRect.Bottom+FBitmapRect.Top);
     LineTo(FBitmapRect.Left+FRamkaRect.Left,FRamkaRect.Bottom+FBitmapRect.Top);
     LineTo(FBitmapRect.Left+FRamkaRect.Left,FRamkaRect.Top+FBitmapRect.Top);
    end;
    FReEnter:=False;
 end;

constructor TBPNavigator.Create(AOwner: TComponent);
 begin
   inherited Create(AOwner);
   Parent:=TWinControl(AOwner);
   FBackCOlor:=clBackground;
   FReEnter:=False;
   FNavigateObject:=nil;
   FNavigateBitmap:=TBitmap.Create;
   FOutputBitmap:=TBitmap.Create;
   FBackColor:=clBackground;
   FNavigateFullWidth:=ClientWidth;
   FNavigateFullHeight:=ClientHeight;
   FNavigateClientHeight:=ClientHeight;
   FNavigateClientWidth:=ClientWidth;
   FNavigatePosLeft:=0;
   FNavigatePosTop:=0;
   FRatio:=1;
   FAlpha:=0.7;
   CreateAlphaArray;
   FShadowCOlor:=clGray;
   FNavigateBitmap.Width:=ClientWidth;
   FNavigateBitmap.Height:=ClientHeight;
   FNavigateBitmap.PixelFormat:=pf24bit;
   FOutputBitmap.Width:=ClientWidth;
   FOutputBitmap.Height:=ClientHeight;
   FOutputBitmap.PixelFormat:=FNavigateBitmap.PixelFormat;
   FMyDrag:=False;
   OnMouseDown:=MyMouseDOwn;
   OnMouseMove:=MyMouseMove;
   OnMouseUp:=MyMouseUp;
{   FNavigateBitmap.Canvas.FillRect(FNavigateBitmap.Canvas.ClipRect);}
   Width:=100;
   Height:=100;
 end;

destructor  TBPNavigator.Destroy;
 begin
   if (FNavigateObject<>nil) then FNavigateObject:=nil;
   FOutputBitmap.Destroy;
   FNavigateBitmap.Destroy;
   OnMouseDown:=nil;
   OnMouseMove:=nil;
   OnMouseUp:=nil;
   inherited Destroy;
 end;

procedure TBPNavigator.SetNavigateObject(Value:TScrollingWinControl);
var r1,r2:Double;
 begin
    if Value=nil then Exit;
    FNavigateObject:=Value;
    ScrollBarChanged(Value);
 end;

procedure TBPNavigator.ScrollBarChanged(Sender:TObject);
var r1,r2:Double;
 begin
    if FNavigateObject=nil then Exit;
    FNavigateFullWidth:=FNavigateObject.HorzScrollBar.Range;
    FNavigateFullHeight:=FNavigateObject.VertScrollBar.Range;
    if FNavigateFullWidth=0 then FNavigateFullWidth:=FNavigateObject.ClientWidth;
    if FNavigateFUllHeight=0 then FNavigateFullHeight:=FNavigateObject.ClientHeight;
    FNavigateClientWidth:=FNavigateObject.ClientWidth;
    FNavigateClientHeight:=FNavigateObject.ClientHeight;
    FNavigatePosLeft:=FNavigateObject.HorzScrollBar.Position;
    FNavigatePosTop:=FNavigateObject.VertScrollBar.Position;
    r1:=ClientWidth/FNavigateFullWidth;
    r2:=ClientHeight/FNavigateFullHeight;
    if r1<r2 then FRatio:=r1 else FRatio:=r2;
    FNavigateBitmap.Width:=Round(FNavigateFullWidth*FRatio);
    FNavigateBitmap.Height:=Round(FNavigateFullHeight*FRatio);
    FNavigateBitmap.PixelFormat:=pf24bit;
    FOutputBitmap.Width:=FNavigateBitmap.Width;
    FOutputBitmap.Height:=FNavigateBitmap.Height;
    FOutputBitmap.PixelFormat:=FNavigateBitmap.PixelFormat;
    Paint;
 end;

procedure TBPNavigator.SetBackColor(Value:TColor);
 begin
   FBackColor:=Value;
   Paint;
 end;

procedure Register;
begin
  RegisterComponents('DeadSouls', [TBPNavigator]);
end;

end.