Unit BPStream;

{ Class TBPFileStream }
{ Corrected 24-12-98 }
{ Corrected 28-09-98 }
{ Rewritten 30-09-98 }
{ Rewritten 22-12-99 }
{ Rewritten 16-11-00 }
{ Modified  06-08-02 }
{ Rewritten 12-01-03 }

Interface

Uses Windows,SysUtils,BPDef;

   {  BPD File structure:
          Header
          NodesTypesInfo
          LinksTypesInfo
          Tree
          Data1
          Data2
          Data3
          Data4
          .....
          OpTree
          OpData1
          OpData2
          OpData3
          .......
          InProcessData;
          EntryArray:ver 6.0
          CheckGraphArray:ver 6.0 }


Type

 { TBPFileStream }
    TBPReadType = (rt_500,rt_600);

    PBPFileHeader =^TBPFileHeader;
    TBPFileHeader = packed record
      dwSignature:DWORD;      { File signature }
      dwSize:DWORD;           { Self size - offset to tree root }
      ssModel:ShortString;    { Model name }
      ssCompany:ShortString;  { Company name }
      ssAuthor:ShortString;   { Author }
      dfDateTime:TDateTime;   { Date and Time creating model }
      dwRevision:DWORD;       { Document revision }
      dwRootEntries:DWORD;    { Tree entries count }
      dwNodesTypesCount:DWORD;
      dwLinksTypesCount:DWORD;
      dwOpEntries:DWORD;      { OpTree entries count }
      dwOpTreeOffset:DWORD;
{@@@}
      dwInProcessEntries:DWORD; { InProcess entries count }
      dwInProcessOffset:DWORD;
{/@@@}
    end;

    PBPRootEntry =^TBPRootEntry;
    TBPRootEntry = packed record
      dwDataOffset:DWORD;
      dwLevel:Integer;          { Opened level e.q. 123 = 1.2.3 }
      dwParent:Integer;
      ssLevel:String[64];
      ssCaption:ShortString;  { Level Caption }
      ssNotes:ShortString;
      dwRevision:DWORD;       { Shit Revision }
      dwNodesCount:DWORD;     { Nodes count   }
      dwLinksCount:DWORD;     { Links count   }
      dwStatus:TDraftStatus;  { Draft status  }
      dwPageNumber:DWORD;     { Page Number   }
      dwLevelType:DWORD;      { New Default Level Type  }
      dwNodesFontSize:Integer;
      dbOverrideNodesFS:Boolean;
      dwDNodesFontSize:Integer;
      dbOverrideDNodesFS:Boolean;
      dwDProjectionFontSize:Integer;
      dwInProcessFontSize:Integer;
      dbOverrideInProcessFS:Boolean;
      dwLabelsFontSize:Integer;
      dbOverrideLabelsFS:Boolean;
      ssReserved:TReservedType; { New Reserved Block }
    end;

    PBPOPEntry=^TBPOpEntry;
    TBPOpEntry= packed record
      dwDataOffset:DWORD;
      dwLevel:Integer;
      dwParent:Integer;
      ssCaption:ShortString;
      dwOpDetailCount:DWORD;
      dwNodeIndex:DWORD;
    end;

      TBPStreamOpenMode = ( stRead,stWrite);

      TBPFileStream = class
       private
        FFileName:String;                { File Name }
        FFile:File;                      { Internal variable }
        FBuf:PByte;                      { Internal variable }
        FFileSize:DWORD;
        FOpened:Boolean;
        FMode:TBPStreamOpenMode;
        FRootEntries:DWORD;
        FNodesTypesCount:DWORD;
        FLinksTypesCount:DWORD;
        FReadType:TBPReadType;
        FOpTreeEntries:DWORD;
        FOpTreeOffset:DWORD;
{@@@}
        FInProcessEntries:DWORD;
        FInProcessOffset:DWORD;
{/@@@}
        function    ReadFileHeader:TBPFileHeader;
        procedure   WriteFileHeader(Value: TBPFileHeader);
        procedure   ReadNodesArray(Var data:PBPNode;nodesCount,offset:DWORD);
        procedure   ReadLinksArray(Var data:PBPLink;linksCount,offset:DWORD);
{@@@}
        procedure   ReadInProcessArray(Var data:PBPInProcessNode;nodesCount,offset:DWORD);
{/@@@}
        procedure   WriteNodesArray(Const data:PBPNode;nodesCount,offset:DWORD);
        procedure   WriteLinksArray(Const data:PBPLink;linksCount,offset:DWORD);
{@@@}
        procedure   WriteInProcessArray(Const data:PBPInProcessNode;nodesCount,offset:DWORD);
{/@@@}
        function    ReadRootEntry(entryNumber:DWORD):TBPRootEntry;
        function    ReadOpTreeEntry(entryNumber:DWORD):TBPOpEntry;
        { New }
        procedure    ReadNodesTypes(Var ntTree:PBPNodeTypes;Count,offset:DWORD);
        procedure    WriteNodesTypes(Const ntTree:PBPNodeTypes;Count,offset:DWORD);
        procedure    ReadLinksTypes(Var ltTree:PBPLinkTypes;Count,offset:DWORD);
        procedure    WriteLinksTypes(Const ltTree:PBPLinkTypes;Count,offset:DWORD);
        procedure    ReadOpDetail(var data:PBPOpDetail;Count,offset:DWORD);
        procedure    WriteOpDetail(Const data:PBPOpDetail;Count,offset:DWORD);
        procedure    ReadMSDAddon(var EntryArray:PBPEntryArray;var CheckGraphArray:PBPCheckGraphArray;Offset:DWORD);
        procedure    WriteMSDAddon(Const EntryArray:PBPEntryArray;Const CheckGraphArray:PBPCheckGraphArray;Offset:DWORD);
        { End new }
        function    ReadTree(Var bpTree:PBPTree;RootEntries:DWORD):Boolean;
        function    WriteTree(Const bpTree:PBPTree;Count:DWORD):Boolean;
        function    ReadOpTree(Var OpTree:PBPOpTree;Count:DWORD):Boolean;
        function    WriteOpTree(Const OpTree:PBPOpTree;Count:DWORD):Boolean;
        procedure   WriteRootEntry(Const bpRootEntry:TBPRootEntry;entryNumber:DWORD);
        procedure   WriteOpTreeEntry(Const bpOpEntry:TBPOpEntry;entryNumber:DWORD);

       public
        constructor Create;
        destructor  Destroy; override;
        function    OpenStream(oFileName:String;mode:TBPStreamOpenMode):Boolean;
{@@@}
        function    ReadBP(Var bpHeader:TBPHeader;Var bpTree:PBPTree;
                     Var ntTree:PBPNodeTypes;Var ltTree:PBPLinkTypes;var OpTree:PBPOpTree;
                     Var EntryArray:PBPEntryArray;Var CheckGraphArray:PBPCheckGraphArray;
                     Var IPArray:PBPInProcessNode):Boolean;
        function    WriteBP(Const bpHeader:TBPHeader;Const bpTree:PBPTree;
                     Const ntTree:PBPNodeTypes;Const ltTree:PBPLinkTypes;Const OpTree:PBPOpTree;
                     Const EntryArray:PBPEntryArray;Const CheckGraphArray:PBPCheckGraphArray;
                     Const IPArray:PBPInProcessNode):Boolean;
{/@@@}
        procedure   CloseStream;
       published
        property FileName:String read FFileName;
        property FileSize:DWORD read FFileSize;
        property RootEntries:DWORD read FRootEntries;
        property OpTreeEntries:DWORD read FOpTreeEntries;
      end;

Const     BPSignature500 = $35445042; { 5.00 }
          BPSignature600 = $36445042; { 6.00 }

Implementation


constructor TBPFileStream.Create;
 begin
    inherited Create;
    FReadType:=rt_500;
    FFileName:='';
    FFileSize:=0;
    FOpened:=False;
    FBuf:=nil;
    FMode:=stRead;
    FRootEntries:=0;
    FOpTreeEntries:=0;
    FOpTreeOffset:=0;
{@@@}
    FInProcessEntries:=0;
    FInProcessOffset:=0;
{/@@@}
    FNodesTypesCount:=0;
    FLinksTypesCount:=0;
 end;

destructor  TBPFileStream.Destroy;
 begin
   if FOpened then CloseStream;
   inherited Destroy;
 end;

function    TBPFileStream.OpenStream(oFileName:String;mode:TBPStreamOpenMode):Boolean;
 begin
 FFileName:=oFileName;
 Result:=False;
 AssignFile(FFile,FFileName);
 {$I-}
 FMode:=mode;
 case (mode) of
  stRead: Reset(FFile,1);
  stWrite: Rewrite(FFile,1);
 end;
 {$I+}
 if IOResult<>0 then Exit;
 FFileSize:=System.FileSize(FFile);
 if mode = stRead then
     begin
       if (FBuf<>nil) then FreeMem(FBuf);
       FBuf:=PByte(AllocMem(FFileSize));
       System.Seek(FFile,0);
       BlockRead(FFile,FBuf^,FFileSize);
     end
   else
    begin
     FBuf:=nil;
     FFileSize:=0;
    end;
 FOpened:=True;
 Result:=True;
 end;

procedure   TBPFileStream.CloseStream;
 begin
 if FOpened then
   begin
     if (FMode=stWrite) and (FBuf<>nil) then
        begin
         System.Seek(FFile,0);
         BlockWrite(FFile,FBuf^,FFileSize);
         FreeMem(FBuf);
         FBuf:=nil;
        end;
     if (FMode=stRead) and (FBuf<>nil) then
        begin
          FreeMem(FBuf);
          FBuf:=nil;
        end;
     CloseFile(FFile);
     FOpened:=False;
   end;
 end;

function   TBPFileStream.ReadFileHeader:TBPFileHeader;
Var tmpPtr:PBPFileHeader;
 begin
   tmpPtr:=PBPFileHeader(FBuf);
   System.Move(tmpPtr^,Result,sizeOf(TBPFileHeader));
   FRootEntries:=Result.dwRootEntries;
   FNodesTypesCount:=Result.dwNodesTypesCount;
   FLinksTypesCount:=Result.dwLinksTypesCount;
   FOpTreeEntries:=Result.dwOpEntries;
   FOpTreeOffset:=Result.dwOpTreeOffset;
 end;

procedure   TBPFileStream.WriteFileHeader(Value: TBPFileHeader);
Var tmpPtr:PBPFileHeader;
 begin
   tmpPtr:=PBPFileHeader(FBuf);
   System.Move(Value,tmpPtr^,sizeOf(TBPFileHeader));
 end;

procedure  TBPFileStream.ReadNodesTypes(Var ntTree:PBPNodeTypes;Count,offset:DWORD);
Var tmpPtr:PBPNodeTypes;
    nodePtr:PByte;
 begin
   ntTree:=PBPNodeTypes(AllocMem(sizeof(TBPNodeTypes)*FNodesTypesCount));
   nodePtr:=FBuf;
   Inc(nodePtr,offset);
   tmpPtr:=PBPNodeTypes(nodePtr);
   System.Move(tmpPtr^,ntTree^,Count*sizeof(TBPNodeTypes));
 end;

procedure  TBPFileStream.WriteNodesTypes(Const ntTree:PBPNodeTypes;Count,offset:DWORD);
Var tmpPtr:PBPNodeTypes;
    nodePtr:PByte;
 begin
   nodePtr:=FBuf;
   Inc(nodePtr,offset);
   tmpPtr:=PBPNodeTypes(nodePtr);
   System.Move(ntTree^,tmpPtr^,Count*sizeof(TBPNodeTypes));
 end;

procedure  TBPFileStream.ReadMSDAddon(var EntryArray:PBPEntryArray;var CheckGraphArray:PBPCheckGraphArray;Offset:DWORD);
Var tmpPtr:PBPEntryArray;
    tmpPtr1:PBPCheckGraphArray;
    nodePtr:PByte;
 begin
   EntryArray:=PBPEntryArray(AllocMem(sizeof(TBPEntryArray)));
   CheckGraphArray:=PBPCheckGraphArray(AllocMem(sizeof(TBPCheckGraphArray)));
   case FReadType of
    rt_500: begin
              EntryArray^.NodesIndexCount:=0;
              CheckGraphArray^.NodesCount:=0;
              CheckGraphArray^.LinksCount:=0;
            end;
    rt_600: begin
             nodePtr:=FBuf;
             Inc(nodePtr,offset);
             tmpPtr:=PBPEntryArray(nodePtr);
             System.Move(tmpPtr^,EntryArray^,sizeof(TBPEntryArray));
             nodePtr:=FBuf;
             Inc(nodePtr,offset+sizeof(TBPEntryArray));
             tmpPtr1:=PBPCheckGraphArray(nodePtr);
             System.Move(tmpPtr1^,CheckGraphArray^,sizeof(TBPCheckGraphArray));
            end;
   end;
 end;

procedure  TBPFileStream.WriteMSDAddon(Const EntryArray:PBPEntryArray;Const CheckGraphArray:PBPCheckGraphArray;Offset:DWORD);
Var tmpPtr:PBPEntryArray;
    tmpPtr1:PBPCheckGraphArray;
    nodePtr:PByte;
 begin
   nodePtr:=FBuf;
   Inc(nodePtr,offset);
   tmpPtr:=PBPEntryArray(nodePtr);
   System.Move(EntryArray^,tmpPtr^,sizeof(TBPEntryArray));
   nodePtr:=FBuf;
   Inc(nodePtr,offset+sizeof(TBPEntryArray));
   tmpPtr1:=PBPCheckGraphArray(nodePtr);
   System.Move(CheckGraphArray^,tmpPtr1^,sizeof(TBPCheckGraphArray));
 end;

procedure  TBPFileStream.ReadLinksTypes(Var ltTree:PBPLinkTypes;Count,offset:DWORD);
Var tmpPtr:PBPLinkTypes;
    nodePtr:PByte;
 begin
   ltTree:=PBPLinkTypes(AllocMem(sizeof(TBPLinkTypes)*FLinksTypesCount));
   nodePtr:=FBuf;
   Inc(nodePtr,offset);
   tmpPtr:=PBPLinkTypes(nodePtr);
   System.Move(tmpPtr^,ltTree^,Count*sizeof(TBPLinkTypes));
 end;

procedure  TBPFileStream.WriteLinksTypes(Const ltTree:PBPLinkTypes;Count,offset:DWORD);
Var tmpPtr:PBPLinkTypes;
    nodePtr:PByte;
 begin
   nodePtr:=FBuf;
   Inc(nodePtr,offset);
   tmpPtr:=PBPLinkTypes(nodePtr);
   System.Move(ltTree^,tmpPtr^,Count*sizeof(TBPLinkTypes));
 end;

procedure TBPFileStream.ReadOpDetail(var data:PBPOpDetail;Count,offset:DWORD);
Var tmpPtr:PBPOpDetail;
    nodePtr:PByte;
 begin
   data:=PBPOpDetail(AllocMem(sizeof(TBPOpDetail)*Count));
   nodePtr:=FBuf;
   Inc(nodePtr,offset);
   tmpPtr:=PBPOpDetail(nodePtr);
   System.Move(tmpPtr^,data^,Count*sizeof(TBPOpDetail));
 end;

procedure TBPFileStream.WriteOpDetail(Const data:PBPOpDetail;Count,offset:DWORD);
Var tmpPtr:PBPOpDetail;
    nodePtr:PByte;
 begin
   nodePtr:=FBuf;
   Inc(nodePtr,offset);
   tmpPtr:=PBPOpDetail(nodePtr);
   System.Move(data^,tmpPtr^,Count*sizeof(TBPOpDetail));
 end;

{@@@}
procedure   TBPFileStream.ReadInProcessArray(Var data:PBPInProcessNode;nodesCount,offset:DWORD);
Var tmpPtr:PBPInProcessNode;
    nodePtr:PByte;
    i:Integer;
 begin
    data:=PBPInProcessNode(AllocMem(sizeof(TBPInProcessNode)*nodesCount));
    nodePtr:=FBuf;
    Inc(nodePtr,offset);
    tmpPtr:=PBPInProcessNode(nodePtr);
    System.Move(tmpPtr^,data^,nodesCount*sizeof(TBPInProcessNode));
 end;

procedure   TBPFileStream.WriteInProcessArray(Const data:PBPInProcessNode;nodesCount,offset:DWORD);
Var  tmpPtr:PBPInProcessNode;
     nodePtr:PByte;
begin
   nodePtr:=FBuf;
   Inc(nodePtr,offset);
   tmpPtr:=PBPInProcessNode(nodePtr);
   System.Move(data^,tmpPtr^,nodesCount*sizeof(TBPInProcessNode));
 end;
{/@@@}

procedure   TBPFileStream.ReadNodesArray(Var data:PBPNode;nodesCount,offset:DWORD);
Var tmpPtr:PBPNode;
    nodePtr:PByte;
    i:Integer;
begin
   nodePtr:=FBuf;
   Inc(nodePtr,offset);
   tmpPtr:=PBPNode(nodePtr);
   System.Move(tmpPtr^,data^,nodesCount*sizeof(TBPNode));
end;

procedure   TBPFileStream.ReadLinksArray(Var data:PBPLink;linksCount,offset:DWORD);
Var tmpPtr:PBPLink;
    nodePtr:PByte;
    i:Integer;
begin
   nodePtr:=FBuf;
   Inc(nodePtr,offset);
   tmpPtr:=PBPLink(nodePtr);
   System.Move(tmpPtr^,data^,linksCount*sizeof(TBPLink));
end;

procedure   TBPFileStream.WriteNodesArray(Const data:PBPNode;nodesCount,offset:DWORD);
Var  tmpPtr:PBPNode;
     nodePtr:PByte;
begin
   nodePtr:=FBuf;
   Inc(nodePtr,offset);
   tmpPtr:=PBPNode(nodePtr);
   System.Move(data^,tmpPtr^,nodesCount*sizeof(TBPNode));
end;

procedure   TBPFileStream.WriteLinksArray(Const data:PBPLink;linksCount,offset:DWORD);
Var tmpPtr:PBPLink;
    nodePtr:PByte;
begin
   nodePtr:=FBuf;
   Inc(nodePtr,offset);
   tmpPtr:=PBPLink(nodePtr);
   System.Move(data^,tmpPtr^,linksCount*sizeof(TBPLink));
end;

function    TBPFileStream.ReadOpTreeEntry(entryNumber:DWORD):TBPOpEntry;
Var tmpPtr:PBPOpEntry;
    nodePtr:PByte;
begin
   nodePtr:=FBuf;
   Inc(nodePtr,FOpTreeOffset+entryNumber*sizeof(TBPOpEntry));
   tmpPtr:=PBPOpEntry(nodePtr);
   System.Move(tmpPtr^,Result,sizeof(TBPOpEntry));
end;

function    TBPFileStream.ReadRootEntry(entryNumber:DWORD):TBPRootEntry;
Var tmpPtr:PBPRootEntry;
    nodePtr:PByte;
    tmpOffset:DWORD;
begin
   nodePtr:=FBuf;
   tmpOffset:=sizeof(TBPFileHeader)+FLinksTypesCount*SizeOf(TBPLinkTypes)+
                                    FNodesTypesCount*SizeOf(TBPNodeTypes);
   Inc(nodePtr,tmpOffset+entryNumber*sizeof(TBPRootEntry));
   tmpPtr:=PBPRootEntry(nodePtr);
   System.Move(tmpPtr^,Result,sizeof(TBPRootEntry));
 end;

function    TBPFileStream.ReadOpTree(Var OpTree:PBPOpTree;Count:DWORD):Boolean;
Var i:Integer;
    tmp:TBPOpEntry;
    tmpTree,oldTmpTree:PBPOpTree;
 begin
   Result:=False;
   oldTmpTree:=nil;
   for i:=0 to (OpTreeEntries-1) do
    begin
      tmp:=ReadOpTreeEntry(i); { ⠫ । entry }
      tmpTree:=PBPOpTree(AllocMem(sizeof(TBPOpTree)));
      tmpTree^.dwDataOffset:=tmp.dwDataOffset;
      tmpTree^.dwLevel:=tmp.dwLevel;
      tmpTree^.dwParent:=tmp.dwParent;
      tmpTree^.ssCaption:=tmp.ssCaption;
      tmpTree^.dwOpDetailCount:=tmp.dwOpDetailCount;
      tmpTree^.dwNodeIndex:=tmp.dwNodeIndex;
      tmpTree^.lpOpDetail:=PBPOpDetail(AllocMem(tmp.dwOpDetailCount*sizeof(TBPOpDetail)));
      ReadOpDetail(tmpTree^.lpOpDetail,tmpTree^.dwOpDetailCount,tmpTree^.dwDataOffset);
      tmpTree^.next:=nil;
      if (i=0) then opTree:=tmpTree;
      if (oldTmpTree<>nil) then oldTmpTree^.next:=tmpTree;
      oldTmpTree:=tmpTree;
    end;
   Result:=True;
 end;

function    TBPFileStream.ReadTree(Var bpTree:PBPTree;RootEntries:DWORD):Boolean;
Var i:Integer;
    tmp:TBPRootEntry;
    tmpTree,oldTmpTree:PBPTree;
 begin
   Result:=False;
   oldTmpTree:=nil;
   for i:=0 to (RootEntries-1) do
    begin
      tmp:=ReadRootEntry(i); { ⠫ । entry }
      tmpTree:=PBPTree(AllocMem(sizeof(TBPTree)));
      tmpTree^.dwDataOffset:=tmp.dwDataOffset;
      tmpTree^.dwLevel:=tmp.dwLevel;
      tmpTree^.dwParent:=tmp.dwParent;
      tmpTree^.ssLevel:=tmp.ssLevel;
      tmpTree^.ssCaption:=tmp.ssCaption;
      tmpTree^.ssNotes:=tmp.ssNotes;
      tmpTree^.dwRevision:=tmp.dwRevision;
      tmpTree^.dwNodesCount:=tmp.dwNodesCount;
      tmpTree^.dwLinksCount:=tmp.dwLinksCount;
      tmpTree^.dwStatus:=tmp.dwStatus;
      tmpTree^.dwPageNumber:=tmp.dwPageNumber;
      tmpTree^.dwNodesMemRequired:=tmp.dwNodesCount*sizeof(TBPNode);
      tmpTree^.dwLinksMemRequired:=tmp.dwLinksCount*sizeof(TBPLink);
      tmpTree^.lpNodes:=PBPNode(AllocMem(tmpTree^.dwNodesMemRequired));
      tmpTree^.lpLinks:=PBPLink(AllocMem(tmpTree^.dwLinksMemRequired));
      tmpTree^.dwNodesFontSize:=tmp.dwNodesFontSize;
      tmpTree^.dbOverrideNodesFS:=tmp.dbOverrideNodesFS;
      tmpTree^.dwDNodesFontSize:=tmp.dwDNodesFontSize;
      tmpTree^.dbOverrideDNodesFS:=tmp.dbOverrideDNodesFS;
      tmpTree^.dwDProjectionFontSize:=tmp.dwDProjectionFontSize;
      tmpTree^.dwInProcessFontSize:=tmp.dwInProcessFontSize;
      tmpTree^.dbOverrideInProcessFS:=tmp.dbOverrideInProcessFS;
      tmpTree^.dwLabelsFontSize:=tmp.dwLabelsFontSize;
      tmpTree^.dbOverrideLabelsFS:=tmp.dbOverrideLabelsFS;
      tmpTree^.ssReserved:=tmp.ssReserved;
      ReadNodesArray(tmpTree^.lpNodes,tmpTree^.dwNodesCount,tmpTree^.dwDataOffset);
      ReadLinksArray(tmpTree^.lpLinks,tmpTree^.dwLinksCount,tmpTree^.dwDataOffset+tmpTree^.dwNodesMemRequired);
      tmpTree^.next:=nil;
      if (i=0) then bpTree:=tmpTree;
      if (oldTmpTree<>nil) then oldTmpTree^.next:=tmpTree;
      oldTmpTree:=tmpTree;
    end;
   Result:=True;
 end;

function    TBPFileStream.ReadBP(Var bpHeader:TBPHeader;Var bpTree:PBPTree;
                                 Var ntTree:PBPNodeTypes;Var ltTree:PBPLinkTypes;Var OpTree:PBPOpTree;
                                 Var EntryArray:PBPEntryArray;Var CheckGraphArray:PBPCheckGraphArray;
                                 Var IPArray:PBPInProcessNode):Boolean;
Var tmp:TBPFileHeader;
    tmpOffset:DWORD;
    msdOffset:DWORD;
 begin
    Result:=False;
    if not(FOpened) then Exit;
    tmp:=ReadFileHeader;
    if (tmp.dwSignature<>BPSignature500) and
       (tmp.dwSignature<>BPSignature600) then Exit;
    if (tmp.dwSignature=BPSignature500) then FReadType:=rt_500 else FReadType:=rt_600;
    bpHeader.ssModel   := tmp.ssModel;
    bpHeader.ssCompany := tmp.ssCompany;
    bpHeader.ssAuthor  := tmp.ssAuthor;
    bpHeader.dfDateTime:= tmp.dfDateTime;
    bpHeader.dwRevision:= tmp.dwRevision;
    bpHeader.dwNodesTypesCount:=tmp.dwNodesTypesCount;
    bpHeader.dwLinksTypesCount:=tmp.dwLinksTypesCount;
    bpHeader.dwOpEntries:=tmp.dwOpEntries;
{@@@}
    bpHeader.dwIPEntries:=tmp.dwInProcessEntries;
    FInProcessEntries:=tmp.dwInProcessEntries;
    FInProcessOffset:=tmp.dwInProcessOffset;
{/@@@}
    FNodesTypesCount:=tmp.dwNodesTypesCount;
    FLinksTypesCount:=tmp.dwLinksTypesCount;
    FOpTreeEntries:=tmp.dwOpEntries;
{@@@}
    ReadInProcessArray(IPArray,FInProcessEntries,FInProcessOffset);
{/@@@}
    tmpOffset:=sizeOf(TBPFileHeader);
    ReadNodesTypes(ntTree,FNodesTypesCount,tmpOffset);
    tmpOffset:=sizeOf(TBPFileHeader)+FNodesTypesCount*sizeof(TBPNodeTypes);
    ReadLinksTypes(ltTree,FLinksTypesCount,tmpOffset);
    Result:=ReadTree(bpTree,tmp.dwRootEntries);
    if (FOpTreeEntries>0) then Result:=Result and ReadOpTree(opTree,tmp.dwOpEntries);
    MSDOffset:=FFileSize-sizeof(TBPEntryArray)-sizeof(TBPCheckGraphArray);
    ReadMSDAddon(EntryArray,CheckGraphArray,MSDOffset);
 end;

function    TBPFileStream.WriteBP(Const bpHeader:TBPHeader;Const bpTree:PBPTree;
                                  Const ntTree:PBPNodeTypes;Const ltTree:PBPLinkTypes; Const OpTree:PBPOpTree;
                                  Const EntryArray:PBPEntryArray;Const CheckGraphArray:PBPCheckGraphArray;
                                  Const IPArray:PBPInProcessNode):Boolean;
Var tmp:TBPFileHeader;
    tmpTree:PBPTree;
    totalNodes:DWORD;
    totalLinks:DWORD;
    totalOp:DWORD;
    tmpOffset:DWORD;
    tmpOpTree:PBPOpTree;
    msdoffset:DWORD;
{@@@}
    ipoffset:DWORD;
{/@@@}
 begin
   Result:=False;
   if FBuf<>nil then FreeMem(FBuf);
   FillChar(tmp,sizeof(TBPFileHeader),0);
   tmp.dwRootEntries:=0;
   tmp.dwOpEntries:=0;
   tmp.dwInProcessEntries:=0;
   totalNodes:=0;
   totalLinks:=0;
   tmpTree:=bpTree;
   while (tmpTree<>nil) do
    begin
     inc(tmp.dwRootEntries);
     totalNodes:=totalNodes+tmpTree^.dwNodesCount;
     totalLinks:=totalLinks+tmpTree^.dwLinksCount;
     tmpTree:=tmpTree^.next;
    end;
   tmpOpTree:=OpTree;
   totalOp:=0;
   while (tmpOpTree<>nil) do
    begin
      inc(tmp.dwOpEntries);
      totalOp:=totalOp+tmpOpTree^.dwOpDetailCount;
      tmpOpTree:=tmpOpTree^.Next;
    end;
{@@@}
   IPOFFSET:=sizeof(TBPFileHeader)
              +tmp.dwRootEntries*sizeof(TBPRootEntry)
              +totalNodes*sizeof(TBPNode)
              +totalLinks*sizeof(TBPLink)+
              bpHeader.dwNodesTypesCount*sizeof(TBPNodeTypes)+
              bpHeader.dwLinksTypesCount*sizeof(TBPLinkTypes)+
              tmp.dwOpEntries*sizeof(TBPOpEntry)+
              totalOp*sizeof(TBPOpDetail);
{/@@@}
   MSDOffset:=sizeof(TBPFileHeader)
              +tmp.dwRootEntries*sizeof(TBPRootEntry)
              +totalNodes*sizeof(TBPNode)
              +totalLinks*sizeof(TBPLink)+
              bpHeader.dwNodesTypesCount*sizeof(TBPNodeTypes)+
              bpHeader.dwLinksTypesCount*sizeof(TBPLinkTypes)+
              bpHeader.dwIPEntries*sizeof(TBPInProcessNode)+
              tmp.dwOpEntries*sizeof(TBPOpEntry)+
              totalOp*sizeof(TBPOpDetail);
   FFileSize:=MSDOffset+sizeof(TBPEntryArray)+sizeof(TBPCheckGraphArray);
   FOpTreeOffset:=sizeof(TBPFileHeader)
                 +tmp.dwRootEntries*sizeof(TBPRootEntry)
                 +totalNodes*sizeof(TBPNode)
                 +totalLinks*sizeof(TBPLink)+
                 bpHeader.dwNodesTypesCount*sizeof(TBPNodeTypes)+
                 bpHeader.dwLinksTypesCount*sizeof(TBPLinkTypes);
   FBuf:=PByte(AllocMem(FFileSize));
   if FBuf=nil then Exit;
   tmp.dwSignature:=BPSignature600;
   tmp.dwSize:=sizeof(TBPFileHeader);
   tmp.ssModel:= bpHeader.ssModel;
   tmp.ssCompany:= bpHeader.ssCompany;
   tmp.ssAuthor:= bpHeader.ssAuthor;
   tmp.dfDateTime:= bpHeader.dfDateTime;
   tmp.dwRevision:= bpHeader.dwRevision;
   tmp.dwNodesTypesCount:= bpHeader.dwNodesTypesCount;
   tmp.dwLinksTypesCount:= bpHeader.dwLinksTypesCount;
   tmp.dwOpTreeOffset:= FOpTreeOffset;
{@@@}
   tmp.dwInProcessEntries:=bpHeader.dwIPEntries;
   tmp.dwInProcessOffset:=ipoffset;
{/@@@}
   FNodesTypesCount:=bpHeader.dwNodesTypesCount;
   FLinksTypesCount:=bpHeader.dwLinksTypesCount;
   FOpTreeEntries:=tmp.dwOpEntries;
   WriteFileHeader(tmp);
   tmpOffset:=sizeOf(TBPFileHeader);
   WriteNodesTypes(ntTree,FNodesTypesCount,tmpOffset);
   tmpOffset:=sizeOf(TBPFileHeader)+FNodesTypesCount*sizeof(TBPNodeTypes);
   WriteLinksTypes(ltTree,FLinksTypesCount,tmpOffset);
   Result:=WriteTree(bpTree,tmp.dwRootEntries);
   if (FOpTreeEntries>0) then Result:=Result and WriteOpTree(OpTree,tmp.dwOpEntries);
{@@@}
   if (bpHeader.dwIPEntries>0) then WriteInProcessArray(IPArray,bpHeader.dwIPEntries,ipoffset);
{/@@@}
   WriteMSDAddon(EntryArray,CheckGraphArray,MSDOffset);
 end;

function    TBPFileStream.WriteOpTree(Const OpTree:PBPOpTree;Count:DWORD):Boolean;
Var tmpTree:PBPOpTree;
    tmp:TBPOpEntry;
    i:Integer;
    offset:DWORD;
 begin
  Result:=False;
  tmpTree:=opTree;
   i:=0;
   offset:=FOpTreeOffset+sizeof(TBPOpEntry)*count;
   while (tmpTree<>nil) do
    begin
     FillChar(tmp,sizeof(TBPOpEntry),0);
     tmp.dwDataOffset:=offset;
     tmp.dwLevel:=tmpTree^.dwLevel;
     tmp.dwParent:=tmpTree^.dwParent;
     tmp.ssCaption:=tmpTree^.ssCaption;
     tmp.dwOpDetailCount:=tmpTree^.dwOpDetailCount;
     tmp.dwNodeIndex:=tmpTree^.dwNodeIndex;
     WriteOpTreeEntry(tmp,i);
     WriteOpDetail(tmpTree^.lpOpDetail,tmpTree^.dwOpDetailCount,offset);
     offset:=offset+tmpTree^.dwOpDetailCount*sizeof(TBPOpDetail);
     tmpTree:=tmpTree^.next;
     Inc(i);
    end;
  Result:=True;
 end;

function    TBPFileStream.WriteTree(Const bpTree:PBPTree;Count:DWORD):Boolean;
Var tmpTree:PBPTree;
    tmp:TBPRootEntry;
    i,j:Integer;
    offset:DWORD;
 begin
  Result:=False;
  tmpTree:=bpTree;
   i:=0;
   offset:=sizeof(TBPFileHeader)+FNodesTypesCount*sizeOf(TBPNodeTypes)+
                                 FLinksTypesCount*sizeOf(TBPLinkTypes)+
                                 sizeof(TBPRootEntry)*count;
   while (tmpTree<>nil) do
    begin
     FillChar(tmp,sizeof(TBPRootEntry),0);
     tmp.dwDataOffset:=offset;
     tmp.dwLevel:=tmpTree^.dwLevel;
     tmp.dwParent:=tmpTree^.dwParent;
     tmp.ssLevel:=tmpTree^.ssLevel;
     tmp.ssCaption:=tmpTree^.ssCaption;
     tmp.ssNotes:=tmpTree^.ssNotes;
     tmp.dwRevision:=tmpTree^.dwRevision;
     tmp.dwNodesCount:=tmpTree^.dwNodesCount;
     tmp.dwLinksCount:=tmpTree^.dwLinksCount;
     tmp.dwNodesFontSize:=tmpTree^.dwNodesFontSize;
     tmp.dbOverrideNodesFS:=tmpTree^.dbOverrideNodesFS;
     tmp.dwDNodesFontSize:=tmpTree^.dwDNodesFontSize;
     tmp.dbOverrideDNodesFS:=tmpTree^.dbOverrideDNodesFS;
     tmp.dwDProjectionFontSize:=tmpTree^.dwDProjectionFontSize;
     tmp.dwInProcessFontSize:=tmpTree^.dwInProcessFontSize;
     tmp.dbOverrideInProcessFS:=tmpTree^.dbOverrideInProcessFS;
     tmp.dwLabelsFontSize:=tmpTree^.dwLabelsFontSize;
     tmp.dbOverrideLabelsFS:=tmpTree^.dbOverrideLabelsFS;
     tmp.dwStatus:=tmpTree^.dwStatus;
     tmp.ssReserved:=tmpTree^.ssReserved;
     tmp.dwPageNumber:=i+1;
     tmp.dwLevelType:=0;
{     if tmp.ssReserved[0]=0 then
       for j:=0 to 231 do  tmp.ssReserved[j]:=0;}
     WriteRootEntry(tmp,i);
     WriteNodesArray(tmpTree^.lpNodes,tmpTree^.dwNodesCount,offset);
     offset:=offset+tmpTree^.dwNodesCount*sizeof(TBPNode);
     WriteLinksArray(tmpTree^.lpLinks,tmpTree^.dwLinksCount,offset);
     offset:=offset+tmpTree^.dwLinksCount*sizeof(TBPLink);
     tmpTree:=tmpTree^.next;
     Inc(i);
    end;
  Result:=True;
 end;

procedure   TBPFileStream.WriteRootEntry(Const bpRootEntry:TBPRootEntry;entryNumber:DWORD);
Var  tmpPtr:PBPRootEntry;
     nodePtr:PByte;
begin
   nodePtr:=FBuf;
   Inc(nodePtr,sizeof(TBPFileHeader)+FNodesTypesCount*sizeOf(TBPNodeTypes)+
                                 FLinksTypesCount*sizeOf(TBPLinkTypes)+entryNumber*sizeof(TBPRootEntry));
   tmpPtr:=PBPRootEntry(nodePtr);
   System.Move(bpRootEntry,tmpPtr^,sizeof(TBPRootEntry));
 end;

procedure   TBPFileStream.WriteOpTreeEntry(Const bpOpEntry:TBPOpEntry;entryNumber:DWORD);
Var  tmpPtr:PBPOpEntry;
     nodePtr:PByte;
begin
   nodePtr:=FBuf;
   Inc(nodePtr,FOpTreeOffset+entryNumber*sizeof(TBPOpEntry));
   tmpPtr:=PBPOpEntry(nodePtr);
   System.Move(bpOpEntry,tmpPtr^,sizeof(TBPOpEntry));
 end;

end.

