unit TextExportUnit;

interface

Uses Windows,SysUtils,Graphics,Classes,BPDef;

Type   TBPTextExport=class
         private
           FTree:PBPTree;
           FNodes:PBPNode;
           FLinks:PBPLink;
           FNTTree:PBPNodeTypes; { NodeTypes Tree }
           FLTTree:PBPLinkTypes; { LinkTypes Tree }
           FOPTree:PBPOpTree;    { Operation Tree }
           FIPNode:PBPInProcessNode;
           FHeader:TBPHeader;    { Header for model }
           FLevel:Integer;       { Level for model, model must have level -1 }
           FParentLevel:Integer; { Parent level for model }
           FNodesCount:DWORD;
           FLinksCount:DWORD;
           FFileName:String;
           FDest:TStringList;
           function  GetMultipleFirst(index:Integer;MultipleFrom:Boolean):PBPLink;
           function  GetIndexedLink(index:Integer):PBPLink;
           function  GetNodeNumber(nodenum:DWORD):String;
           function  GetParentLevel(s:String):String;
           { String functions }
           function  GetPenStyle(ps:TPenStyle):String;
           function  GetBoolean(val:Boolean):String;
           function  GetShadowStyle(ss:DWORD):String;
           function  GetTColor(val:TColor):String;
           function  DeleteQuotes(s:String):String;
           { Translate functions }
           procedure TranslateFileToken;
           procedure TranslateModelToken;
           procedure TranslateNTTree;
           procedure TranslateLTTree;
           procedure TranslateTree;
           procedure TranslateNodes;
           procedure TranslateLinks;
           procedure TranslateIPNode;
           procedure TranslateOpTree;
         public
           constructor Create;
           destructor  Destroy; override;
           procedure   InitTranslator(Tree:PBPTree;NTTree:PBPNodeTypes;LTTree:PBPLinkTypes;OPTree:PBPOpTree;Header:TBPHeader;IPArray:PBPInProcessNode;FileName:String);
           function    Translate:Boolean;
           procedure   DoneTranslator;
           procedure   SaveResults(fileName:String);
       end;

implementation

function  TBPTextExport.GetParentLevel(s:String):String;
var lev:Integer;
    s1:String;
begin
   Result:='-2';
   lev:=StrToIntDef(s,-2);
   if lev=-1 then Exit;
   if lev=0 then  begin Result:='-1'; Exit; end;
   if lev<>-2 then begin Result:='0'; Exit; end;
   s1:=s;
   while (length(s1)>0) and (s1[length(s1)]<>'.') do delete(s1,length(s1),1);
   if (length(s1)>0) and (s1[length(s1)]='.') then delete(s1,length(s1),1);
   Result:=s1;
end;

function  TBPTextExport.GetPenStyle(ps:TPenStyle):String;
 begin
   Result:='Solid';
   case ps of
     psSolid:      Result:='Solid';
     psDash:       Result:='Dash';
     psDot:        Result:='Dot';
     psDashDot:    Result:='DashDot';
     psDashDotDot: Result:='DashDotDot';
   end;
 end;

function  TBPTextExport.GetBoolean(val:Boolean):String;
 begin
   if val then
     Result:='True'
   else
     Result:='False';
 end;

function  TBPTextExport.GetShadowStyle(ss:DWORD):String;
 begin
   Result:='RightFramed';
   case ss of
     0:      Result:='None';
     1:      Result:='LeftBoxed';
     2:      Result:='LeftFramed';
     3:      Result:='TopPerspective';
     4:      Result:='InvPerspective';
     5:      Result:='RightBoxed';
     6:      Result:='RightFramed';
     7:      Result:='BottomPerspective';
   end;
 end;

function  TBPTextExport.GetTColor(val:TColor):String;
 begin
   Result:='$'+IntToHex(DWORD(val),8);
 end;



function  TBPTextExport.DeleteQuotes(s:String):String;
 begin
   Result:=s;
   while (pos('"',Result)<>0) do
      delete(Result,pos('"',Result),1);
   while (pos(',',Result)<>0) do
      delete(Result,pos(',',Result),1);
 end;

function  TBPTextExport.GetMultipleFirst(index:Integer;MultipleFrom:Boolean):PBPLink; { Checked }
 Var linkP:PBPLink;
     f:Boolean;
     linkT:PBPLink;
     i:Integer;
  begin
   Result:=nil;
   if FLinksCount=0 then Exit;
   i:=0;
   linkT:=FLinks;
   inc(linkT,index);
   linkP:=FLinks;
   f:=True;
   while (i<FLinksCount) and (f) do
    begin
      if MultipleFrom then
         begin
           if (linkP^.dwToNodeNumber=linkT^.dwToNodeNumber) and
              ((linkP^.dwFlags and BP_TOMASK)=(linkT^.dwFlags and BP_TOMASK)) and
              (linkP^.dwToLinkNumber=linkT^.dwToLinkNumber) then
              begin
                Result:=linkP;
                f:=False;
              end;
         end
        else
         begin
           if (linkP^.dwFromNodeNumber=linkT^.dwFromNodeNumber) and
              ((linkP^.dwFlags and BP_FROMMASK)=(linkT^.dwFlags and BP_FROMMASK)) and
              (linkP^.dwFromLinkNumber=linkT^.dwFromLinkNumber) then
              begin
                Result:=linkP;
                f:=False;
              end;
         end;
      inc(linkP);
      inc(i);
    end;
  end;


function  TBPTextExport.GetNodeNumber(nodenum:DWORD):String;
Var s:String;
   i,cnt:Integer;
   node:PBPNode;
 begin
   s:='0';
   if (FNodesCount=0) then Exit;
   cnt:=FNodesCount-1;
   node:=FNodes;
   for i:=0 to cnt do
    begin
      if (node^.dwNodeNumber=nodenum) then
          begin
            s:=Trim(node^.ssNodeNumber);
            delete(s,1,1);
          end;
      inc(node);
    end;
   Result:=s;
 end;

procedure TBPTextExport.TranslateFileToken;
Var s:String;
 begin
   s:='File,"'+DeleteQuotes(ExtractFileName(FFileName))+'"';
   FDest.Append(s);
 end;

procedure TBPTextExport.TranslateModelToken;
Var s:String;
 begin
   s:='Model,"'+DeleteQuotes(FHeader.ssModel)+'","'+
      DeleteQuotes(FHeader.ssCompany)+'","'+
      DeleteQuotes(FHeader.ssAuthor)+'",'+IntToStr(FHeader.dwRevision);
   FDest.Append(s);
 end;

procedure TBPTextExport.TranslateNTTree;
Var s:String;
    i,cnt:Integer;
    nt:PBPNodeTypes;
 begin
 {}
  cnt:=FHeader.dwNodesTypesCount-1;
  nt:=FNTTree;
  for i:=0 to cnt do
   begin
     s:='NodeType,'+IntToStr(nt^.dwNodeIndex)+
        ',"'+DeleteQuotes(nt^.ssCaption)+'"'+
        ','+GetTColor(nt^.dwFontColor)+
        ','+GetTColor(nt^.dwPaintColor)+
        ','+GetTColor(nt^.dwDrawColor)+
        ','+GetTColor(nt^.dwShadowColor)+
        ','+GetShadowStyle(nt^.dwShadowStyle);
     FDest.Append(s);
     inc(nt);
   end;
 end;

procedure TBPTextExport.TranslateLTTree;
Var s:String;
    i,cnt:Integer;
    lt:PBPLinkTypes;
 begin
  cnt:=FHeader.dwLinksTypesCount-1;
  lt:=FLTTree;
  for i:=0 to cnt do
   begin
     s:='LinkType,'+IntToStr(lt^.dwLinkIndex)+
        ',"'+DeleteQuotes(lt^.ssCaption)+'"'+
        ','+GetTColor(lt^.dwColor)+
        ','+GetPenStyle(lt^.dwPenStyle);
     FDest.Append(s);
     inc(lt);
   end;
 end;

procedure TBPTextExport.TranslateOpTree;
Var i,j,cnt,cnt1:Integer;
    p:PBPOpTree;
    OD:PBPOpDetail;
    s:String;
 begin
    if (FOpTree=nil) then Exit;
    p:=FOpTree;
    cnt:=FHeader.dwOpEntries-1;
    for i:=0 to cnt do
     begin
       s:='OpNode,'+IntToStr(p^.dwLevel)+','+IntToStr(p^.dwParent)+',"'+DeleteQuotes(p^.ssCaption)+
           '",'+IntToStr(p^.dwNodeIndex);
       FDest.Append(s);
       cnt1:=p^.dwOpDetailCount-1;
       od:=p^.lpOpDetail;
       for j:=0 to cnt1 do
        begin
          s:='OpDetail,"'+DeleteQuotes(od^.ssCaption)+'"';
          FDest.Append(s);
          inc(od);
        end;
       p:=p^.next;
     end;
 end;

procedure TBPTextExport.TranslateTree;
Var tmpTree:PBPTree;
    s:String;
    sLevel,sParent:String;
 begin
 {}
   if (FTree=nil) then Exit;
   tmpTree:=FTree;
   while (tmpTree<>nil) do
    begin
      FLinksCount:=tmpTree^.dwLinksCount;
      FNodesCount:=tmpTree^.dwNodesCount;
      FLinks:=tmpTree^.lpLinks;
      FNodes:=tmpTree^.lpNodes;
      FLevel:=tmpTree^.dwLevel;
      sLevel:=tmpTree^.ssLevel;
      sParent:='"'+GetParentLevel(sLevel)+'"';
      sLevel:='"'+sLevel+'"';
      FParentLevel:=tmpTree^.dwParent;
      s:='Level,'+sLevel+','+sParent+',"'
         +DeleteQuotes(tmpTree^.ssCaption)+'","'+DeleteQuotes(tmpTree^.ssNotes)+'",'+
         IntToStr(tmpTree^.dwRevision)+',';
      if (dsWorking in tmpTree^.dwStatus) then
       s:=s+'Working'
       else
       begin
       if (dsDraft in tmpTree^.dwStatus) then
        s:=s+'Draft'
        else
        begin
        if (dsRecomended in tmpTree^.dwStatus) then
         s:=s+'Recomended'
         else
          if (dsPublication in tmpTree^.dwStatus) then
           s:=s+'Publication';
        end;
       end;
      FDest.Append(s);
      TranslateNodes;
      TranslateLinks;
      tmpTree:=tmpTree^.next;
    end;
 end;


procedure TBPTextExport.TranslateIPNode;
Var node:PBPInProcessNode;
    cnt,i:Integer;
    s,s1:String;
    com,path:String;
 begin
   if (FHeader.dwIPEntries=0) then Exit;
   node:=FIPNode;
   cnt:=FHeader.dwIPEntries-1;
   for i:=0 to cnt do
    begin
      s1:=IntToStr(node^.dwNodeNumber);
      com:=node^.ssComment;
      if length(com)>0 then com:='"'+com+'"';
      path:=node^.ssPath;
      if length(path)>0 then path:='"'+path+'"';
      s:='InProcessNode,"'+DeleteQuotes(node^.ssCaption)+'","'
                          +DeleteQuotes(node^.ssMech)+'","'
                          +DeleteQuotes(node^.ssUpr)+'","'+s1+'",'+
         IntToStr(node^.dwNodeIndex)+','+com+','+path;
      FDest.Append(s);
      inc(node);
    end;
 end;

procedure TBPTextExport.TranslateNodes;
Var s,s1,abc:String;
    i,cnt:Integer;
    node:PBPNode;
    num:Integer;
    com:String;
 begin
   {}
   if (FNodesCount=0) then Exit;
   node:=FNodes;
   cnt:=FNodesCount-1;
   for i:=0 to cnt do
    begin
      s1:=Trim(node^.ssNodeNumber);
      delete(s1,1,1);
{      num:=StrToInt(s1);}
      abc:=FloatToStr(node^.fABCValue);
      if pos(',',abc)>0 then
         abc[pos(',',abc)]:='.';
      com:=node^.ssComment;
      if length(com)>0 then com:='"'+com+'"';
      s:='Node,"'+DeleteQuotes(node^.ssCaption)+'","'+s1+'",'+
         IntToStr(node^.dwNodeIndex)+','+abc+','+com;

      FDest.Append(s);
      inc(node);
    end;
 end;

function TBPTextExport.GetIndexedLink(index:Integer):PBPLink;
  Var linkP:PBPLink;
  begin
    Result:=nil;
    if (index>=FLinksCount) then Exit;
    linkP:=FLinks;
    if (index>0) then Inc(linkP,index);
    Result:=linkP;
  end;



procedure TBPTextExport.TranslateLinks;
var s:String;
    i,cnt:Integer;
    link,mflink:PBPLink;
    fromnode:String;
    tonode:String;
    fromtoSide:String;
    fromcap,tocap,gcap:String;
    indF,indT:String;
    addon:String;
    FromPath:String;
    ToPath:String;
 begin
   {}
    if (FLinksCount=0) then Exit;
    link:=FLinks;
    cnt:=FLinksCount-1;
    for i:=0 to cnt do
     begin
       if ((link^.dwFromNodeNumber and BPGlobalMask)<>0) then
        begin
          case (link^.dwFromNodeNumber and BPGlobalMask) of
           BPGlobalLeftNodeNumber:   fromnode:='GlobalLeft';
           BPGlobalRightNodeNumber:  fromnode:='GlobalRight';
           BPGlobalTopNodeNumber:    fromnode:='GlobalTop';
           BPGlobalBottomNodeNumber: fromnode:='GlobalBottom';
           BPGlobalNodeOperation:    fromnode:='GlobalOperation='+IntToStr(link^.dwFromNodeNumber and  BPNotGlobalMask);
           BPGlobalNodeInProcess:    fromnode:='GlobalInProcess='+IntToStr(link^.dwFromNodeNumber and  BPNotGlobalMask);
          end;
        end
       else
        begin
          fromnode:='"'+GetNodeNumber(link^.dwFromNodeNumber)+'"';
        end;
       if ((link^.dwToNodeNumber and BPGlobalMask)<>0) then
        begin
          case (link^.dwToNodeNumber and BPGlobalMask) of
           BPGlobalLeftNodeNumber:   tonode:='GlobalLeft';
           BPGlobalRightNodeNumber:  tonode:='GlobalRight';
           BPGlobalTopNodeNumber:    tonode:='GlobalTop';
           BPGlobalBottomNodeNumber: tonode:='GlobalBottom';
           BPGlobalNodeOperation:    tonode:='GlobalOperation='+IntToStr(link^.dwToNodeNumber and  BPNotGlobalMask);
           BPGlobalNodeInProcess:    tonode:='GlobalInProcess='+IntToStr(link^.dwToNodeNumber and  BPNotGlobalMask);
          end;
        end
       else
        begin
          tonode:='"'+GetNodeNumber(link^.dwToNodeNumber)+'"';
        end;
       case (link^.dwFlags and BP_FROMTOMASK) of
        BP_LEFT_LEFT:       fromtoside:='LeftSide,LeftSide';
        BP_LEFT_RIGHT:      fromtoside:='LeftSide,RightSide';
        BP_LEFT_TOP:        fromtoside:='LeftSide,TopSide';
        BP_LEFT_BOTTOM:      fromtoside:='LeftSide,BottomSide';
        BP_RIGHT_LEFT:      fromtoside:='RightSide,LeftSide';
        BP_RIGHT_RIGHT:      fromtoside:='RightSide,RightSide';
        BP_RIGHT_TOP:      fromtoside:='RightSide,TopSide';
        BP_RIGHT_BOTTOM:      fromtoside:='RightSide,BottomSide';
        BP_TOP_LEFT:      fromtoside:='TopSide,LeftSide';
        BP_TOP_RIGHT:      fromtoside:='TopSide,RightSide';
        BP_TOP_TOP:      fromtoside:='TopSide,TopSide';
        BP_TOP_BOTTOM:      fromtoside:='TopSide,BottomSide';
        BP_BOTTOM_LEFT:      fromtoside:='BottomSide,LeftSide';
        BP_BOTTOM_RIGHT:      fromtoside:='BottomSide,RightSide';
        BP_BOTTOM_TOP:      fromtoside:='BottomSide,TopSide';
        BP_BOTTOM_BOTTOM:      fromtoside:='BottomSide,BottomSide';
       else
         fromtoside:='BadSide,BadSide';
       end;
       if ((link^.dwFlags and BP_SHOW_FROMCAPTION)<>0) then
        fromcap:='"'+DeleteQuotes(link^.ssFromCaption)+'"'
       else
        fromcap:='';
       if ((link^.dwFlags and BP_SHOW_TOCAPTION)<>0) then
        Tocap:='"'+DeleteQuotes(link^.ssToCaption)+'"'
       else
        tocap:='';
       if ((link^.dwFlags and BP_SHOW_GLOBALCAPTION)<>0) then
         GCap:='"'+DeleteQuotes(link^.ssGlobalCaption)+'"'
       else
        gcap:='';
        if fromcap='' then
          begin
          mflink:=GetMultipleFirst(i,False);
          if (mflink<>nil) then
            begin
              if ((mflink^.dwFlags and BP_SHOW_FROMCAPTION)<>0) then
               fromcap:='"'+DeleteQuotes(mflink^.ssFromCaption)+'"';
            end;
          end;
        if tocap='' then
          begin
           mflink:=GetMultipleFirst(i,True);
           if (mflink<>nil) then
            begin
              if ((mflink^.dwFlags and BP_SHOW_TOCAPTION)<>0) then
               tocap:='"'+DeleteQuotes(mflink^.ssToCaption)+'"';
            end;
          end;
       indF:=IntToStr(link^.dwFromLinkNumber);
       indT:=IntToStr(link^.dwToLinkNumber);
       if ((link^.dwFlags and BP_NOTUSED_FROM)<>0) then
        begin
         if ((link^.dwFlags and BP_NOTUSED_TO)<>0) then
          addon:='NotUsedAll'
         else
          addon:='NotUsedFrom';
        end
       else
        begin
         if ((link^.dwFlags and BP_NOTUSED_TO)<>0) then
          addon:='NotUsedTo'
         else
          addon:='';
        end;
       FromPath:=link^.ssFromPath;
       ToPath:=link^.ssToPath;
       if length(FromPath)>0 then FromPath:='"'+FromPath+'"';
       if length(ToPath)>0 then ToPath:='"'+ToPath+'"';
       s:='Link,'+fromNode+','+tonode+','+fromtoside+','+fromcap+','+tocap+','+gcap+','+
           IntToStr(link^.dwLinkIndex)+','+indF+','+indT+','+addon+','+FromPath+','+ToPath;
       FDest.Append(s);
       inc(link);
     end;
 end;

constructor TBPTextExport.Create;
 begin
   inherited Create;
   FDest:=TStringList.Create;
   FLinks:=nil;
   FNodes:=nil;
   FLinksCount:=0;
   FNodesCount:=0;
 end;

destructor  TBPTextExport.Destroy;
 begin
   FTree:=nil;
   FNTTree:=nil;
   FLTTree:=nil;
   FLinks:=nil;
   FNodes:=nil;
   FOpTree:=nil;
   FIPNode:=nil;
   FDest.Free;
   inherited Destroy;
 end;

procedure   TBPTextExport.InitTranslator(Tree:PBPTree;NTTree:PBPNodeTypes;LTTree:PBPLinkTypes;OpTree:PBPOpTree;Header:TBPHeader;IPArray:PBPInProcessNode;FileName:String);
 begin
   { Init translator }
   FTree:=Tree;
   FNTTree:=NTTree;
   FLTTree:=LTTree;
   FOpTree:=OpTree;
   FHeader:=Header;
   FIPNode:=IPArray;
   FFileName:=FileName;
   FDest.Clear;
 end;

function    TBPTextExport.Translate:Boolean;
 begin
  Result:=True;
  TranslateFileToken;
  TranslateModelToken;
  TranslateNTTree;
  TranslateLTTree;
  TranslateOpTree;
  TranslateIPNode;
  TranslateTree;
 end;

procedure   TBPTextExport.DoneTranslator;
 begin
  {}
   FTree:=nil;
   FNTTree:=nil;
   FLTTree:=nil;
   FLinks:=nil;
   FNodes:=nil;
   FOpTree:=nil;
   FFileName:='';
   FDest.Clear;
 end;

procedure   TBPTextExport.SaveResults(fileName:String);
 begin
   FDest.SaveToFile(FileName);
 end;


end.
