Unit Debug;

Interface

Uses Windows,SysUtils,BPDef;

procedure WriteLog(const logstring:ShortString);
procedure WriteNode(node:PBPNode);
procedure WriteNodes(StartNode:PBPNode;NodesCount:integer);
procedure WriteLink(link:PBPLink);
procedure WriteLinks(StartLink:PBPLink;LinksCount:integer);
procedure WriteLevel(tree:PBPTree);
procedure WriteTree(tree:PBPTree);

Const filename='bpd.txt';

Implementation

procedure WriteLog(const logstring:ShortString);
Var F:TextFile;
    S:String;
 begin
  try
   AssignFile(F,filename);
 {$I-}
  Append(F);
 {$I+}
 if IOResult<>0 then rewrite(F);
 s:=LogString;
 WriteLn(F,S);
 CloseFile(F);
  except
  try
   AssignFile(F,'fio.err');
   Rewrite(F);
   WriteLn(F,'Unable to write log file');
  finally
   CloseFile(F);
  end;
 end;
 end;

procedure WriteNode(node:PBPNode);
Var s:ShortString;
 begin
    WriteLog('');
    WriteLog('NodeBegin');
    if (node<>nil) then
     begin
       s:=node^.ssCaption;
       WriteLog('Caption: '+s);
       s:=IntToStr(node^.dwNodeNumber);
       WriteLog('Number: '+s);
       s:=node^.ssNodeNumber;
       WriteLog('DrawNumber: '+s);
     end
    else WriteLog('EmptyNode');
    WriteLog('NodeEnd');
    WriteLog('');
 end;

procedure WriteNodes(StartNode:PBPNode;NodesCount:integer);
Var i:integer;
    tmpNode:PBPNode;
 begin
   tmpNode:=StartNode;
   for i:=0 to (NodesCount-1) do
    begin
      WriteNode(tmpNode);
      inc(tmpNode);
    end;
 end;

procedure WriteLink(link:PBPLink);
Var s:ShortString;
 begin
   WriteLog('');
   WriteLog('LinkBegin');
    if (link<>nil) then
     begin
        if ((link^.dwFlags and BP_SHOW_FROMCAPTION)<>0) then
        WriteLog('FromCaption: '+link^.ssFromCaption);
        if ((link^.dwFlags and BP_SHOW_TOCAPTION)<>0) then
        WriteLog('ToCaption: '+link^.ssToCaption);
        if ((link^.dwFlags and BP_SHOW_GLOBALCAPTION)<>0) then
        WriteLog('GlobalCaption: '+link^.ssGlobalCaption);
        if ((link^.dwFromNodeNumber and BPGlobalMask)=0) then
         WriteLog('FromNode: '+IntToStr(link^.dwFromNodeNumber))
        else
         begin
          case (link^.dwFromNodeNumber and BPGlobalMask) of
           BPGlobalNodeLeft:  WriteLog('FromGlobalNumber: GlobalLeft');
           BPGlobalNodeRight: WriteLog('FromGlobalNumber: GlobalRight');
           BPGlobalNodeTop:   WriteLog('FromGlobalNumber: GlobalTop');
           BPGlobalNodeBottom:WriteLog('FromGlobalNumber: GlobalBottom');
           BPGlobalNodeOperation: WriteLog('FromGlobalNumber: Operation');
          end;
          WriteLog('FromParentNode: '+IntToStr(link^.dwFromNodeNumber and BPNotGlobalMask));
         end;
        case (link^.dwFlags and BP_FROMMASK) of
          BP_FROMNODE_LEFT:    WriteLog('FromSide: Left');
          BP_FROMNODE_RIGHT:   WriteLog('FromSide: Right');
          BP_FROMNODE_TOP:     WriteLog('FromSide: Top');
          BP_FROMNODE_BOTTOM:  WriteLog('FromSide: Bottom');
        end;
        WriteLog('FromICAM: '+IntToStr(link^.dwFromLinkNumber));
        if ((link^.dwToNodeNumber and BPGlobalMask)=0) then
         WriteLog('ToNode: '+IntToStr(link^.dwToNodeNumber))
        else
         begin
          case (link^.dwToNodeNumber and BPGlobalMask) of
           BPGlobalNodeLeft:  WriteLog('ToGlobalNumber: GlobalLeft');
           BPGlobalNodeRight: WriteLog('ToGlobalNumber: GlobalRight');
           BPGlobalNodeTop:   WriteLog('ToGlobalNumber: GlobalTop');
           BPGlobalNodeBottom:WriteLog('ToGlobalNumber: GlobalBottom');
           BPGlobalNodeOperation: WriteLog('ToGlobalNumber: Operation');
          end;
          WriteLog('ToParentNode: '+IntToStr(link^.dwToNodeNumber and BPNotGlobalMask));
         end;
        case (link^.dwFlags and BP_TOMASK) of
          BP_TONODE_LEFT:    WriteLog('ToSide: Left');
          BP_TONODE_RIGHT:   WriteLog('ToSide: Right');
          BP_TONODE_TOP:     WriteLog('ToSide: Top');
          BP_TONODE_BOTTOM:  WriteLog('ToSide: Bottom');
        end;
        WriteLog('ToICAM: '+IntToStr(link^.dwToLinkNumber));
        if ((link^.dwFlags and BP_NOTUSED_FROM)<>0) then
        WriteLog('NotUsed: From');
        if ((link^.dwFlags and BP_NOTUSED_TO)<>0) then
        WriteLog('NotUsed: To');
     end
    else WriteLog('EmptyLink');
    WriteLog('LinkEnd');
    WriteLog('');
 end;

procedure WriteLinks(StartLink:PBPLink;LinksCount:integer);
Var tmpLink:PBPLink;
    i:integer;
 begin
    tmpLink:=StartLink;
    for i:=0 to (LinksCount-1) do
     begin
       WriteLink(tmpLink);
       inc(tmpLink);
     end;
 end;

procedure WriteLevel(tree:PBPTree);
 begin
   WriteLog('');
   WriteLog('LevelBegin');
    if (tree<>nil) then
     begin
       WriteLog('Caption: '+tree^.ssCaption);
       WriteLog('Level: '+IntToStr(tree^.dwLevel));
       WriteLog('ParentLevel: '+IntToStr(tree^.dwParent));
       WriteNodes(tree^.lpNodes,tree^.dwNodesCount);
       WriteLinks(tree^.lpLinks,tree^.dwLinksCount);
     end
    else
     WriteLog('EmptyLevel');
   WriteLog('LevelEnd');
   WriteLog('');
 end;

procedure WriteTree(tree:PBPTree);
Var tmpTree:PBPTree;
 begin
   WriteLog('');
   WriteLog('TreeBegin');
   if (tree<>nil) then
     begin
       tmpTree:=tree;
       while (tmpTree<>nil) do
        begin
          WriteLevel(tmpTree);
          tmpTree:=tmpTree^.next;
        end;
     end
    else
     WriteLog('EmptyTree');
   WriteLog('TreeEnd');
   WriteLog('');
 end;

end.
