unit DSEdit;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs;

type
  TDSEdit = class;


  TDSEdit = class(TScrollingWinControl)
  private
    { Private declarations }
    FCanvas:TCanvas;
    procedure WMPaint(var Message: TWMPaint); message WM_PAINT;
  protected
    { Protected declarations }
    procedure Paint; virtual;
    procedure PaintWindow(DC: HDC); override;
    procedure PaintTo(DC: HDC; X, Y: Integer);
    procedure PaintPageTo(DC: HDC; X, Y: Integer; PageClipRect:TRect);
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    { Published declarations }
    property AutoScroll;
    property HorzScrollBar;
    property VertScrollBar;
    property HelpContext;
    property TabOrder;
    property TabStop;
    property Align default alNone;
    property Height default 100;
    property Hint;
    property Left;
    property Name;
    property ShowHint;
    property Top;
    property Visible;
    property Width default 100;
    property Canvas:TCanvas read FCanvas;
  end;

procedure Register;

implementation

constructor TDSEdit.Create(AOwner: TComponent);
begin
 inherited Create(AOwner);
 Parent:=(AOwner as TWinControl);
 FCanvas := TControlCanvas.Create;
 TControlCanvas(FCanvas).Control := Self;
 FCanvas.Brush.Color:=clWhite;
end;

destructor TDSEdit.Destroy;
begin
 if (FCanvas<>nil) then FCanvas.Destroy;
 inherited Destroy;
end;

procedure TDSEdit.WMPaint(var Message: TWMPaint);
begin
   PaintHandler(Message);
end;

procedure TDSEdit.PaintWindow(DC: HDC);
begin
  inherited PaintWindow(DC);
  FCanvas.Lock;
  try
    FCanvas.Handle := DC;
    try
      Paint;
    finally
      FCanvas.Handle := 0;
    end;
  finally
    FCanvas.Unlock;
  end;
end;

procedure TDSEdit.Paint;
begin
   FCanvas.FillRect(ClientRect);
{  Nodes.Root.SetUpTree;}
end;

procedure TDSEdit.PaintTo(DC: HDC; X, Y: Integer);
var
  SaveIndex: Integer;
begin
{  Include(self.FControlState, csPaintCopy);}
  SaveIndex := SaveDC(DC);
  MoveWindowOrg(DC, X, Y);
  Perform(WM_ERASEBKGND, DC, 0);
  Perform(WM_PAINT, DC, 0);
  RestoreDC(DC, SaveIndex);
{  Exclude(FControlState, csPaintCopy);}
end;

procedure TDSEdit.PaintPageTo(DC: HDC; X, Y: Integer; PageClipRect:TRect);
var
  I, EdgeFlags, BorderFlags, SaveIndex: Integer;
  R: TRect;
begin
  SaveIndex := SaveDC(DC);
  MoveWindowOrg(DC, X-PageClipRect.Left, Y-PageClipRect.Top);
  IntersectClipRect(DC, PageClipRect.Left, PageClipRect.Top, PageClipRect.Right, PageClipRect.Bottom);
  Perform(WM_ERASEBKGND, DC, 0);
  Perform(WM_PAINT, DC, 0);
  RestoreDC(DC, SaveIndex);
end;


procedure Register;
begin
  RegisterComponents('DeadSouls', [TDSEdit]);
end;

end.
