
{*******************************************************}
{                                                       }
{       Delphi Visual Component Library                 }
{                                                       }
{       Copyright (c) 1996,97 Borland International     }
{                                                       }
{       Rewritten by Dr.Khumen (c) 97 Dead Souls Inc.   }
{       (Added Support for RichEdit 2.0 format )        }
{                                                       }
{*******************************************************}

unit DSRichEd;

{$R-}

interface

uses Messages, Windows, SysUtils, CommCtrl, Classes, Controls, Forms,
  Menus, Graphics, StdCtrls, DSREdit, ToolWin;

Type  { TDSTextAttributes }

  TDSCustomRichEdit = class;

  TDSAttributeType = (atSelected, atDefaultText,atAll,atWord,atUseUIRules);
  TDSConsistentAttribute = (caBold, caColor, caFace, caItalic,
    caSize, caStrikeOut, caUnderline, caProtected,caLink,caOffset,caCharset,
    caSmallCaps,caAllCaps,caHidden,caOutline,caShadow,caEmboss,
    caImPrint,caDisabled,caRevised,caBackColor,caLCID,caUnderlineType,
    caWeight,caSpacing,caKerning,caStyle,caAnimation,caRevauthor);
  TDSConsistentAttributes = set of TDSConsistentAttribute;
  TDSUnderlineType = (utNone,utSingle,utWord,utDouble,utDotted,utInvert,utCF1);

  TDSTextAttributes = class(TPersistent)
  private
    RichEdit: TDSCustomRichEdit;
    FType: TDSAttributeType;
    procedure GetAttributes(var Format: TCharFormat2);
    function GetCharset: TFontCharset;
    function GetColor: TColor;
    function GetBgColor: TColor;
    function GetConsistentAttributes: TDSConsistentAttributes;
    function GetHeight: Integer;
    function GetName: TFontName;
    function GetPitch: TFontPitch;
    function GetProtected: Boolean;
    function GetSize: Integer;
    function GetStyle: TFontStyles;
    function GetUnderlineType: TDSUnderlineType;
    function GetWeight: Word;
    function GetSpacing: Smallint;
    function GetLID : LCID;
    function GetSStyle : SmallInt;
    function GetKerning : Word;
    function GetAnimation : Byte;
    function GetRevAuthor : Byte;
    procedure SetAttributes(var Format: TCharFormat2);
    procedure SetCharset(Value: TFontCharset);
    procedure SetColor(Value: TColor);
    procedure SetBgColor(Value: TColor);
    procedure SetHeight(Value: Integer);
    procedure SetName(Value: TFontName);
    procedure SetPitch(Value: TFontPitch);
    procedure SetProtected(Value: Boolean);
    procedure SetSize(Value: Integer);
    procedure SetStyle(Value: TFontStyles);
    procedure SetUnderlineType(Value: TDSUnderlineType);
    procedure SetWeight(Value: Word);
    procedure SetSpacing(Value: Smallint);
    procedure SetLID(Value: LCID);
    procedure SetSStyle(Value: SmallInt);
    procedure SetKerning(Value: Word);
    procedure SetAnimation(Value: Byte);
    procedure SetRevAuthor(Value: Byte);
  protected
    procedure InitFormat(var Format: TCharFormat2);
    procedure AssignTo(Dest: TPersistent); override;
  public
    constructor Create(AOwner: TDSCustomRichEdit; AttributeType: TDSAttributeType);
    procedure Assign(Source: TPersistent); override;
    property Charset: TFontCharset read GetCharset write SetCharset;
    property Color: TColor read GetColor write SetColor;
    property BgColor: TColor read GetBgColor write SetBgColor;
    property ConsistentAttributes: TDSConsistentAttributes read GetConsistentAttributes;
    property Name: TFontName read GetName write SetName;
    property Pitch: TFontPitch read GetPitch write SetPitch;
    property Protected: Boolean read GetProtected write SetProtected;
    property Size: Integer read GetSize write SetSize;
    property Style: TFontStyles read GetStyle write SetStyle;
    property Height: Integer read GetHeight write SetHeight;
    property UnderlineType: TDSUnderlineType read GetUnderlineType write SetUnderlineType;
    property Weight: Word read GetWeight write SetWeight;
    property Spacing: Smallint read GetSpacing write SetSpacing;
    property LID: LCID   read GetLID write SetLID;
    property SStyle: Smallint  read GetSStyle write SetSStyle;
    property Kerning: Word  read GetKerning write SetKerning;
    property Animation: Byte  read GetAnimation write SetAnimation;
    property RevAuthor: Byte  read GetRevAuthor write SetRevAuthor;
  end;

{ TDSParaAttributes }

  TDSNumberingStyle = (nsNone,nsLCLetter,nsUCLetter,nsLCRoman,nsUCRoman,nsSymbol,nsNumber,nsSequence);
  TDSAlignment = (taLeftJustify, taRightJustify, taCenter,taJustify);

  TDSLineSpacing = record
    dyLineSpacing: Longint;     { Line spacing depending on Rule                 }
    bLineSpacingRule: Byte;     { Rule for line spacing (see tom.doc)    }
  end;

  TDSBorder = record
    wBorderSpace: Word;         { Space between border and text (twips) }
    wBorderWidth: Word;         { Border pen width (twips)                               }
    wBorders: Word;             { Byte 0: bits specify which borders     }
                                { Nibble 2: border style, 3: color index }
  end;

  TDSShading = record
    wShadingWeight: Word;       { Shading in hundredths of a per cent    }
    wShadingStyle: Word;        { Nibble 0: style, 1: cfpat, 2: cbpat    }
  end;

  TDSParaAttributes = class(TPersistent)
  private
    RichEdit: TDSCustomRichEdit;
    procedure GetAttributes(var Paragraph: TParaFormat2);
    function GetAlignment: TDSAlignment;
    function GetFirstIndent: Longint;
    function GetLeftIndent: Longint;
    function GetRightIndent: Longint;
    function GetNumbering: TDSNumberingStyle;
    function GetTab(Index: Byte): Longint;
    function GetTabCount: Integer;
    function GetSpaceBefore : LongInt;
    function GetSpaceAfter : LongInt;
    function GetLineSpacing : TDSLineSpacing;
    function GetSStyle : SmallInt;
    function GetShading : TDSShading;
    function GetBorder : TDSBorder;
    function GetNumberingStyle : Word;
    function GetNumberingStart : Word;
    function GetNumberingTab : Word;
    procedure InitPara(var Paragraph: TParaFormat2);
    procedure SetAlignment(Value: TDSAlignment);
    procedure SetAttributes(var Paragraph: TParaFormat2);
    procedure SetFirstIndent(Value: Longint);
    procedure SetLeftIndent(Value: Longint);
    procedure SetRightIndent(Value: Longint);
    procedure SetNumbering(Value: TDSNumberingStyle);
    procedure SetTab(Index: Byte; Value: Longint);
    procedure SetTabCount(Value: Integer);
    procedure SetSpaceBefore(Value: LongInt);
    procedure SetSpaceAfter(Value: LongInt);
    procedure SetLineSpacing(Value: TDSLineSpacing);
    procedure SetShading(Value: TDSShading);
    procedure SetBorder(Value: TDSBorder);
    procedure SetSStyle(Value: Smallint);
    procedure SetNumberingStyle(Value: Word);
    procedure SetNumberingStart(Value: Word);
    procedure SetNumberingTab(Value: Word);

  public
    constructor Create(AOwner: TDSCustomRichEdit);
    procedure Assign(Source: TPersistent); override;
    property Alignment: TDSAlignment read GetAlignment write SetAlignment;
    property FirstIndent: Longint read GetFirstIndent write SetFirstIndent;
    property LeftIndent: Longint read GetLeftIndent write SetLeftIndent;
    property Numbering: TDSNumberingStyle read GetNumbering write SetNumbering;
    property RightIndent: Longint read GetRightIndent write SetRightIndent;
    property Tab[Index: Byte]: Longint read GetTab write SetTab;
    property TabCount: Integer read GetTabCount write SetTabCount;
    property SpaceBefore: LongInt read GetSpaceBefore write SetSpaceBefore;
    property SpaceAfter: LongInt read GetSpaceAfter write SetSpaceAfter;
    property LineSpacing: TDSLineSpacing read GetLineSpacing write SetLineSpacing;
    property Shading: TDSShading read GetShading write SetShading;
    property Border: TDSBorder read GetBorder write SetBorder;
    property SStyle: Smallint read GetSStyle write SetSStyle;
    property NumberingStyle: Word read GetNumberingStyle write SetNumberingStyle;
    property NumberingStart: Word read GetNumberingStart write SetNumberingStart;
    property NumberingTab: Word read GetNumberingTab write SetNumberingTab;
  end;

{ TDSCustomRichEdit }

  TDSRichEditResizeEvent = procedure(Sender: TObject; Rect: TRect) of object;
  TDSRichEditProtectChange = procedure(Sender: TObject;
    StartPos, EndPos: Integer; var AllowChange: Boolean) of object;
  TDSRichEditSaveClipboard = procedure(Sender: TObject;
    NumObjects, NumChars: Integer; var SaveClipboard: Boolean) of object;
  TDSSearchType = (stWholeWord, stMatchCase);
  TDSSearchTypes = set of TDSSearchType;

  TDSConversion = class(TObject)
  public
    function ConvertReadStream(Stream: TStream; Buffer: PChar; BufSize: Integer): Integer; virtual;
    function ConvertWriteStream(Stream: TStream; Buffer: PChar; BufSize: Integer): Integer; virtual;
  end;

  TDSConversionClass = class of TDSConversion;

  PDSConversionFormat = ^TDSConversionFormat;
  TDSConversionFormat = record
    ConversionClass: TDSConversionClass;
    Extension: string;
    Next: PDSConversionFormat;
  end;

  PDSRichEditStreamInfo = ^TDSRichEditStreamInfo;
  TDSRichEditStreamInfo = record
    Converter: TDSConversion;
    Stream: TStream;
  end;

  TDSCustomRichEdit = class(TCustomMemo)
  private
    FLibHandle: THandle;
    FHideScrollBars: Boolean;
    FSelAttributes: TDSTextAttributes;
    FDefAttributes: TDSTextAttributes;
    FParagraph: TDSParaAttributes;
    FScreenLogPixels: Integer;
    FRichEditStrings: TStrings;
    FMemStream: TMemoryStream;
    FOnSelChange: TNotifyEvent;
    FHideSelection: Boolean;
    FModified: Boolean;
    FDefaultConverter: TDSConversionClass;
    FOnResizeRequest: TDSRichEditResizeEvent;
    FOnProtectChange: TDSRichEditProtectChange;
    FOnSaveClipboard: TDSRichEditSaveClipboard;
    FPageRect: TRect;
    procedure CMColorChanged(var Message: TMessage); message CM_COLORCHANGED;
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    procedure CNNotify(var Message: TWMNotify); message CN_NOTIFY;
    function GetPlainText: Boolean;
    function ProtectChange(StartPos, EndPos: Integer): Boolean;
    function SaveClipboard(NumObj, NumChars: Integer): Boolean;
    procedure SetHideScrollBars(Value: Boolean);
    procedure SetHideSelection(Value: Boolean);
    procedure SetPlainText(Value: Boolean);
    procedure SeTDSRichEditStrings(Value: TStrings);
    procedure SetDefAttributes(Value: TDSTextAttributes);
    procedure SetSelAttributes(Value: TDSTextAttributes);
    procedure WMNCDestroy(var Message: TWMNCDestroy); message WM_NCDESTROY;
    procedure WMSetCursor(var Message: TWMSetCursor); message WM_SETCURSOR;
    procedure WMPaint(var Message: TWMPaint); message WM_PAINT;
    procedure WMSetFont(var Message: TWMSetFont); message WM_SETFONT;
  protected
    procedure CreateParams(var Params: TCreateParams); override;
    procedure CreateWnd; override;
    procedure DestroyWnd; override;
    procedure RequestSize(const Rect: TRect); virtual;
    procedure SelectionChange; dynamic;
    procedure DoSetMaxLength(Value: Integer); override;
    function GetSelLength: Integer; override;
    function GetSelStart: Integer; override;
    function GetSelText: string; override;
    procedure SetSelLength(Value: Integer); override;
    procedure SetSelStart(Value: Integer); override;
    property HideSelection: Boolean read FHideSelection write SetHideSelection default True;
    property HideScrollBars: Boolean read FHideScrollBars
      write SetHideScrollBars default True;
    property Lines: TStrings read FRichEditStrings write SeTDSRichEditStrings;
    property OnSaveClipboard: TDSRichEditSaveClipboard read FOnSaveClipboard
      write FOnSaveClipboard;
    property OnSelectionChange: TNotifyEvent read FOnSelChange write FOnSelChange;
    property OnProtectChange: TDSRichEditProtectChange read FOnProtectChange
      write FOnProtectChange;
    property OnResizeRequest: TDSRichEditResizeEvent read FOnResizeRequest
      write FOnResizeRequest;
    property PlainText: Boolean read GetPlainText write SetPlainText default False;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Clear; override;
    function FindText(const SearchStr: string;
      StartPos, Length: Integer; Options: TDSSearchTypes): Integer;
    function GetSelTextBuf(Buffer: PChar; BufSize: Integer): Integer; override;
    procedure Print(const Caption: string);
    class procedure RegisterConversionFormat(const AExtension: string;
      AConversionClass: TDSConversionClass);
    property DefaultConverter: TDSConversionClass
      read FDefaultConverter write FDefaultConverter;
    property DefAttributes: TDSTextAttributes read FDefAttributes write SetDefAttributes;
    property SelAttributes: TDSTextAttributes read FSelAttributes write SetSelAttributes;
    property PageRect: TRect read FPageRect write FPageRect;
    property Paragraph: TDSParaAttributes read FParagraph;
  end;

  TDSRichEdit = class(TDSCustomRichEdit)
  published
    property Align;
    property Alignment;
    property BorderStyle;
    property Color;
    property Ctl3D;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Font;
    property HideSelection;
    property HideScrollBars;
    property ImeMode;
    property ImeName;
    property Lines;
    property MaxLength;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PlainText;
    property PopupMenu;
    property ReadOnly;
    property ScrollBars;
    property ShowHint;
    property TabOrder;
    property TabStop default True;
    property Visible;
    property WantTabs;
    property WantReturns;
    property WordWrap;
    property OnChange;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnResizeRequest;
    property OnSelectionChange;
    property OnStartDrag;
    property OnProtectChange;
    property OnSaveClipboard;
  end;

function InitCommonControl(CC: Integer): Boolean;
procedure CheckCommonControl(CC: Integer);

procedure Register;

implementation

uses Printers, Consts, ComStrs;

const
  SectionSizeArea = 8;
  RTFConversionFormat: TDSConversionFormat = (
    ConversionClass: TDSConversion;
    Extension: 'rtf';
    Next: nil);
  TexTDSConversionFormat: TDSConversionFormat = (
    ConversionClass: TDSConversion;
    Extension: 'txt';
    Next: @RTFConversionFormat);

var
  ConversionFormatList: PDSConversionFormat = @TexTDSConversionFormat;

function InitCommonControl(CC: Integer): Boolean;
var
  ICC: TInitCommonControlsEx;
begin
  ICC.dwSize := SizeOf(TInitCommonControlsEx);
  ICC.dwICC := CC;
  Result := InitCommonControlsEx(ICC);
  if not Result then InitCommonControls;
end;

procedure CheckCommonControl(CC: Integer);
begin
  if not InitCommonControl(CC) then
    raise EComponentError.Create(SInvalidComCtl32);
end;

procedure Register;
begin
  RegisterComponents('Samples', [TDSRichEdit]);
end;

{ TDSTextAttributes }

constructor TDSTextAttributes.Create(AOwner: TDSCustomRichEdit;
  AttributeType: TDSAttributeType);
begin
  inherited Create;
  RichEdit := AOwner;
  FType := AttributeType;
end;

procedure TDSTextAttributes.InitFormat(var Format: TCharFormat2);
begin
  FillChar(Format, SizeOf(TCharFormat2), 0);
  Format.cbSize := SizeOf(TCharFormat2);
end;

function TDSTextAttributes.GetConsistentAttributes: TDSConsistentAttributes;
var
  Format: TCharFormat2;
begin
  Result := [];
  if RichEdit.HandleAllocated and (FType = atSelected) then
  begin
    InitFormat(Format);
    SendMessage(RichEdit.Handle, EM_GETCHARFORMAT,
      WPARAM(FType = atSelected), LPARAM(@Format));
    with Format do
    begin
      if (dwMask and CFM_BOLD) <> 0 then Include(Result, caBold);
      if (dwMask and CFM_COLOR) <> 0 then Include(Result, caColor);
      if (dwMask and CFM_FACE) <> 0 then Include(Result, caFace);
      if (dwMask and CFM_ITALIC) <> 0 then Include(Result, caItalic);
      if (dwMask and CFM_SIZE) <> 0 then Include(Result, caSize);
      if (dwMask and CFM_STRIKEOUT) <> 0 then Include(Result, caStrikeOut);
      if (dwMask and CFM_UNDERLINE) <> 0 then Include(Result, caUnderline);
      if (dwMask and CFM_PROTECTED) <> 0 then Include(Result, caProtected);
      if (dwMask and CFM_SMALLCAPS) <> 0 then Include(Result, caSmallCaps);
      if (dwMask and CFM_ALLCAPS) <> 0 then Include(Result, caAllCaps);
      if (dwMask and CFM_HIDDEN) <> 0 then Include(Result, caHidden);
      if (dwMask and CFM_OUTLINE) <> 0 then Include(Result, caOutline);
      if (dwMask and CFM_SHADOW) <> 0 then Include(Result, caShadow);
      if (dwMask and CFM_EMBOSS) <> 0 then Include(Result, caEmboss);
      if (dwMask and CFM_IMPRINT) <> 0 then Include(Result, caImPrint);
      if (dwMask and CFM_DISABLED) <> 0 then Include(Result, caDisabled);
      if (dwMask and CFM_REVISED) <> 0 then Include(Result, caRevised);
      if (dwMask and CFM_BACKCOLOR) <> 0 then Include(Result, caBackColor);
      if (dwMask and CFM_LCID) <> 0 then Include(Result, caLCID);
      if (dwMask and CFM_UNDERLINETYPE) <> 0 then Include(Result, caUnderlineType);
      if (dwMask and CFM_WEIGHT) <> 0 then Include(Result, caWeight);
      if (dwMask and CFM_SPACING) <> 0 then Include(Result, caSpacing);
      if (dwMask and CFM_KERNING) <> 0 then Include(Result, caKerning);
      if (dwMask and CFM_STYLE) <> 0 then Include(Result, caStyle);
      if (dwMask and CFM_ANIMATION) <> 0 then Include(Result, caAnimation);
      if (dwMask and CFM_REVAUTHOR) <> 0 then Include(Result, caRevauthor);
      if (dwMask and CFM_LINK) <> 0 then Include(Result, caLink);
      if (dwMask and CFM_OFFSET) <> 0 then Include(Result, caOffset);
      if (dwMask and CFM_CHARSET) <> 0 then Include(Result, caCharset);
    end;
  end;
end;

procedure TDSTextAttributes.GetAttributes(var Format: TCharFormat2);
begin
  InitFormat(Format);
  if RichEdit.HandleAllocated then
    SendMessage(RichEdit.Handle, EM_GETCHARFORMAT,
      WPARAM(FType = atSelected), LPARAM(@Format));
end;

procedure TDSTextAttributes.SetAttributes(var Format: TCharFormat2);
var
  Flag: Longint;
begin
  case (FType) of
    atSelected:   Flag:=SCF_SELECTION;
    atWord:       Flag:=SCF_WORD;
    atAll:        Flag:=SCF_ALL;
    atUseUIRules: Flag:=SCF_USEUIRULES;
   else Flag := 0;
  end;
  if RichEdit.HandleAllocated then
    SendMessage(RichEdit.Handle, EM_SETCHARFORMAT, Flag, LPARAM(@Format))
end;

function TDSTextAttributes.GetCharset: TFontCharset;
var
  Format: TCharFormat2;
begin
  GetAttributes(Format);
  Result := Format.bCharset;
end;

procedure TDSTextAttributes.SetCharset(Value: TFontCharset);
var
  Format: TCharFormat2;
begin
  InitFormat(Format);
  with Format do
  begin
    dwMask := CFM_CHARSET;
    bCharSet := Value;
  end;
  SetAttributes(Format);
end;

function TDSTextAttributes.GetProtected: Boolean;
var
  Format: TCharFormat2;
begin
  GetAttributes(Format);
  with Format do
    if (dwEffects and CFE_PROTECTED) <> 0 then
      Result := True else
      Result := False;
end;

procedure TDSTextAttributes.SetProtected(Value: Boolean);
var
  Format: TCharFormat2;
begin
  InitFormat(Format);
  with Format do
  begin
    dwMask := CFM_PROTECTED;
    if Value then dwEffects := CFE_PROTECTED;
  end;
  SetAttributes(Format);
end;

function TDSTextAttributes.GetColor: TColor;
var
  Format: TCharFormat2;
begin
  GetAttributes(Format);
  with Format do
    if (dwEffects and CFE_AUTOCOLOR) <> 0 then
      Result := clWindowText else
      Result := crTextColor;
end;

procedure TDSTextAttributes.SetColor(Value: TColor);
var
  Format: TCharFormat2;
begin
  InitFormat(Format);
  with Format do
  begin
    dwMask := CFM_COLOR;
    if Value = clWindowText then
      dwEffects := CFE_AUTOCOLOR else
      crTextColor := ColorToRGB(Value);
  end;
  SetAttributes(Format);
end;

function TDSTextAttributes.GetWeight: Word;
var
  Format: TCharFormat2;
begin
  GetAttributes(Format);
  with Format do Result := wWeight;
end;

procedure TDSTextAttributes.SetWeight(Value: Word);
var
  Format: TCharFormat2;
begin
  InitFormat(Format);
  with Format do
  begin
    dwMask := CFM_WEIGHT;
    wWeight := Value;
  end;
  SetAttributes(Format);
end;

function TDSTextAttributes.GetSpacing: Smallint;
var
  Format: TCharFormat2;
begin
  GetAttributes(Format);
  with Format do Result := sSpacing;
end;

procedure TDSTextAttributes.SetSpacing(Value: Smallint);
var
  Format: TCharFormat2;
begin
  InitFormat(Format);
  with Format do
  begin
    dwMask := CFM_SPACING;
    sSpacing := Value;
  end;
  SetAttributes(Format);
end;


function TDSTextAttributes.GetLID: LCID;
var
  Format: TCharFormat2;
begin
  GetAttributes(Format);
  with Format do Result := lid;
end;

procedure TDSTextAttributes.SetLID(Value: LCID);
var
  Format: TCharFormat2;
begin
  InitFormat(Format);
  with Format do
  begin
    dwMask := CFM_LCID;
    lid := Value;
  end;
  SetAttributes(Format);
end;


function TDSTextAttributes.GetSStyle: Smallint;
var
  Format: TCharFormat2;
begin
  GetAttributes(Format);
  with Format do Result := sStyle;
end;

procedure TDSTextAttributes.SetSStyle(Value: Smallint);
var
  Format: TCharFormat2;
begin
  InitFormat(Format);
  with Format do
  begin
    dwMask := CFM_STYLE;
    sStyle := Value;
  end;
  SetAttributes(Format);
end;


function TDSTextAttributes.GetKerning: word;
var
  Format: TCharFormat2;
begin
  GetAttributes(Format);
  with Format do Result := wKerning;
end;

procedure TDSTextAttributes.SetKerning(Value: Word);
var
  Format: TCharFormat2;
begin
  InitFormat(Format);
  with Format do
  begin
    dwMask := CFM_KERNING;
    wKerning := Value;
  end;
  SetAttributes(Format);
end;


function TDSTextAttributes.GetAnimation: Byte;
var
  Format: TCharFormat2;
begin
  GetAttributes(Format);
  with Format do Result := bAnimation;
end;

procedure TDSTextAttributes.SetAnimation(Value: Byte);
var
  Format: TCharFormat2;
begin
  InitFormat(Format);
  with Format do
  begin
    dwMask := CFM_ANIMATION;
    bAnimation := Value;
  end;
  SetAttributes(Format);
end;


function TDSTextAttributes.GetRevAuthor: Byte;
var
  Format: TCharFormat2;
begin
  GetAttributes(Format);
  with Format do Result := bRevAuthor;
end;

procedure TDSTextAttributes.SetRevAuthor(Value: Byte);
var
  Format: TCharFormat2;
begin
  InitFormat(Format);
  with Format do
  begin
    dwMask := CFM_REVAUTHOR;
    bRevAuthor := Value;
  end;
  SetAttributes(Format);
end;


function TDSTextAttributes.GetBgColor: TColor;
var
  Format: TCharFormat2;
begin
  GetAttributes(Format);
  with Format do
    if (dwEffects and CFE_AUTOCOLOR) <> 0 then
      Result := clWindowText else
      Result := crBackColor;
end;

procedure TDSTextAttributes.SetBgColor(Value: TColor);
var
  Format: TCharFormat2;
begin
  InitFormat(Format);
  with Format do
  begin
    dwMask := CFM_BACKCOLOR;
    if Value = clWindowText then
      dwEffects := CFE_AUTOCOLOR else
      crBackColor := ColorToRGB(Value);
  end;
  SetAttributes(Format);
end;


function TDSTextAttributes.GetName: TFontName;
var
  Format: TCharFormat2;
begin
  GetAttributes(Format);
  Result := Format.szFaceName;
end;

procedure TDSTextAttributes.SetName(Value: TFontName);
var
  Format: TCharFormat2;
begin
  InitFormat(Format);
  with Format do
  begin
    dwMask := CFM_FACE;
    StrPLCopy(szFaceName, Value, SizeOf(szFaceName));
  end;
  SetAttributes(Format);
end;

function TDSTextAttributes.GetStyle: TFontStyles;
var
  Format: TCharFormat2;
begin
  Result := [];
  GetAttributes(Format);
  with Format do
  begin
    if (dwEffects and CFE_BOLD) <> 0 then Include(Result, fsBold);
    if (dwEffects and CFE_ITALIC) <> 0 then Include(Result, fsItalic);
    if (dwEffects and CFE_UNDERLINE) <> 0 then Include(Result, fsUnderline);
    if (dwEffects and CFE_STRIKEOUT) <> 0 then Include(Result, fsStrikeOut);
  end;
end;

procedure TDSTextAttributes.SetStyle(Value: TFontStyles);
var
  Format: TCharFormat2;
begin
  InitFormat(Format);
  with Format do
  begin
    dwMask := CFM_BOLD or CFM_ITALIC or CFM_UNDERLINE or CFM_STRIKEOUT;
    if fsBold in Value then dwEffects := dwEffects or CFE_BOLD;
    if fsItalic in Value then dwEffects := dwEffects or CFE_ITALIC;
    if fsUnderline in Value then dwEffects := dwEffects or CFE_UNDERLINE;
    if fsStrikeOut in Value then dwEffects := dwEffects or CFE_STRIKEOUT;
  end;
  SetAttributes(Format);
end;

function TDSTextAttributes.GetUnderlineType: TDSUnderlineType;
var
  Format: TCharFormat2;
  r:TDSUnderlineType;
begin
  r:=utNone;
  GetAttributes(Format);
  with Format do
  begin
    if (bUnderlineType and CFU_UNDERLINENONE) <> 0 then r:=utNone;
    if (bUnderlineType and CFU_UNDERLINE) <> 0 then r:=utSingle;
    if (bUnderlineType and CFU_UNDERLINEWORD) <> 0 then r:=utWord;
    if (bUnderlineType and CFU_UNDERLINEDOUBLE) <> 0 then r:=utDouble;
    if (bUnderlineType and CFU_INVERT) <> 0 then r:=utInvert;
    if (bUnderlineType and CFU_UNDERLINEDOTTED) <> 0 then r:=utDotted;
    if (bUnderlineType and CFU_CF1UNDERLINE) <> 0 then r:=utCF1;
  end;
  Result:=r;
end;

procedure TDSTextAttributes.SetUnderlineType(Value: TDSUnderlineType);
var
  Format: TCharFormat2;
begin
   InitFormat(Format);
   with Format do
     begin
      case (Value) of
       utNone: bUnderlineType:=CFU_UNDERLINENONE;
       utSingle: bUnderlineType:=CFU_UNDERLINE;
       utWord: bUnderlineType:=CFU_UNDERLINEWORD;
       utDouble: bUnderlineType:=CFU_UNDERLINEDOUBLE;
       utDotted: bUnderlineType:=CFU_UNDERLINEDOTTED;
       utInvert: bUnderlineType:=CFU_INVERT;
       utCF1: bUnderlineType:=CFU_CF1UNDERLINE;
      end;
     end;
   SetAttributes(Format);
end;

function TDSTextAttributes.GetSize: Integer;
var
  Format: TCharFormat2;
begin
  GetAttributes(Format);
  Result := Format.yHeight div 20;
end;

procedure TDSTextAttributes.SetSize(Value: Integer);
var
  Format: TCharFormat2;
begin
  InitFormat(Format);
  with Format do
  begin
    dwMask := CFM_SIZE;
    yHeight := Value * 20;
  end;
  SetAttributes(Format);
end;

function TDSTextAttributes.GetHeight: Integer;
begin
  Result := MulDiv(Size, RichEdit.FScreenLogPixels, 72);
end;

procedure TDSTextAttributes.SetHeight(Value: Integer);
begin
  Size := MulDiv(Value, 72, RichEdit.FScreenLogPixels);
end;

function TDSTextAttributes.GetPitch: TFontPitch;
var
  Format: TCharFormat2;
begin
  GetAttributes(Format);
  case (Format.bPitchAndFamily and $03) of
    DEFAULT_PITCH: Result := fpDefault;
    VARIABLE_PITCH: Result := fpVariable;
    FIXED_PITCH: Result := fpFixed;
  else
    Result := fpDefault;
  end;
end;

procedure TDSTextAttributes.SetPitch(Value: TFontPitch);
var
  Format: TCharFormat2;
begin
  InitFormat(Format);
  with Format do
  begin
    case Value of
      fpVariable: Format.bPitchAndFamily := VARIABLE_PITCH;
      fpFixed: Format.bPitchAndFamily := FIXED_PITCH;
    else
      Format.bPitchAndFamily := DEFAULT_PITCH;
    end;
  end;
  SetAttributes(Format);
end;

procedure TDSTextAttributes.Assign(Source: TPersistent);
begin
  if Source is TFont then
  begin
    Color := TFont(Source).Color;
    Name := TFont(Source).Name;
    Charset := TFont(Source).Charset;
    Style := TFont(Source).Style;
    Size := TFont(Source).Size;
    Pitch := TFont(Source).Pitch;
  end
  else if Source is TDSTextAttributes then
  begin
    Color := TDSTextAttributes(Source).Color;
    Name := TDSTextAttributes(Source).Name;
    Charset := TDSTextAttributes(Source).Charset;
    Style := TDSTextAttributes(Source).Style;
    Pitch := TDSTextAttributes(Source).Pitch;
  end
  else inherited Assign(Source);
end;

procedure TDSTextAttributes.AssignTo(Dest: TPersistent);
begin
  if Dest is TFont then
  begin
    TFont(Dest).Color := Color;
    TFont(Dest).Name := Name;
    TFont(Dest).Charset := Charset;
    TFont(Dest).Style := Style;
    TFont(Dest).Size := Size;
    TFont(Dest).Pitch := Pitch;
  end
  else if Dest is TDSTextAttributes then
  begin
    TDSTextAttributes(Dest).Color := Color;
    TDSTextAttributes(Dest).Name := Name;
    TDSTextAttributes(Dest).Charset := Charset;
    TDSTextAttributes(Dest).Style := Style;
    TDSTextAttributes(Dest).Pitch := Pitch;
  end
  else inherited AssignTo(Dest);
end;

{ TDSParaAttributes }

constructor TDSParaAttributes.Create(AOwner: TDSCustomRichEdit);
begin
  inherited Create;
  RichEdit := AOwner;
end;

procedure TDSParaAttributes.InitPara(var Paragraph: TParaFormat2);
begin
  FillChar(Paragraph, SizeOf(TParaFormat2), 0);
  Paragraph.cbSize := SizeOf(TParaFormat2);
end;

procedure TDSParaAttributes.GetAttributes(var Paragraph: TParaFormat2);
begin
  InitPara(Paragraph);
  if RichEdit.HandleAllocated then
    SendMessage(RichEdit.Handle, EM_GETPARAFORMAT, 0, LPARAM(@Paragraph));
end;

procedure TDSParaAttributes.SetAttributes(var Paragraph: TParaFormat2);
begin
  if RichEdit.HandleAllocated then
    SendMessage(RichEdit.Handle, EM_SETPARAFORMAT, 0, LPARAM(@Paragraph))
end;

function TDSParaAttributes.GetAlignment: TDSAlignment;
var
  Paragraph: TParaFormat2;
begin
  GetAttributes(Paragraph);
  Result := TDSAlignment(Paragraph.wAlignment - 1);
end;

procedure TDSParaAttributes.SetAlignment(Value: TDSAlignment);
var
  Paragraph: TParaFormat2;
begin
  InitPara(Paragraph);
  with Paragraph do
  begin
    dwMask := PFM_ALIGNMENT;
    wAlignment := Ord(Value) + 1;
  end;
  SetAttributes(Paragraph);
end;

function TDSParaAttributes.GetSpaceBefore: LongInt;
var
  Paragraph: TParaFormat2;
begin
  GetAttributes(Paragraph);
  Result := Paragraph.dySpaceBefore;
end;

procedure TDSParaAttributes.SetSpaceBefore(Value: LongInt);
var
  Paragraph: TParaFormat2;
begin
  InitPara(Paragraph);
  with Paragraph do
  begin
    dwMask := PFM_SPACEBEFORE;
    dySpaceBefore := Value;
  end;
  SetAttributes(Paragraph);
end;

function TDSParaAttributes.GetSpaceAfter: LongInt;
var
  Paragraph: TParaFormat2;
begin
  GetAttributes(Paragraph);
  Result := Paragraph.dySpaceAfter;
end;

procedure TDSParaAttributes.SetSpaceAfter(Value: LongInt);
var
  Paragraph: TParaFormat2;
begin
  InitPara(Paragraph);
  with Paragraph do
  begin
    dwMask := PFM_SPACEAFTER;
    dySpaceAfter := Value;
  end;
  SetAttributes(Paragraph);
end;

function TDSParaAttributes.GetLineSpacing: TDSLineSpacing;
var
  Paragraph: TParaFormat2;
  r:TDSLineSpacing;
begin
  GetAttributes(Paragraph);
  with (Paragraph) do
   begin
    r.dyLineSpacing :=dyLineSpacing;
    r.bLineSpacingRule := bLineSpacingRule;
   end;
  Result := r;
end;

procedure TDSParaAttributes.SetLineSpacing(Value: TDSLineSpacing);
var
  Paragraph: TParaFormat2;
begin
  InitPara(Paragraph);
  with Paragraph do
  begin
    dwMask := PFM_LINESPACING;
    dyLineSpacing := Value.dyLineSpacing;
    bLineSpacingRule := Value.bLineSpacingRule;
  end;
  SetAttributes(Paragraph);
end;

function TDSParaAttributes.GetSStyle: Smallint;
var
  Paragraph: TParaFormat2;
begin
  GetAttributes(Paragraph);
  Result := Paragraph.sStyle;
end;

procedure TDSParaAttributes.SetSStyle(Value: Smallint);
var
  Paragraph: TParaFormat2;
begin
  InitPara(Paragraph);
  with Paragraph do
  begin
    dwMask := PFM_STYLE;
    sStyle := Value;
  end;
  SetAttributes(Paragraph);
end;

function TDSParaAttributes.GetShading: TDSShading;
var
  Paragraph: TParaFormat2;
  r:TDSShading;
begin
  GetAttributes(Paragraph);
  with (Paragraph) do
   begin
    r.wShadingWeight := wShadingWeight;
    r.wShadingStyle := wShadingStyle;
   end;
  Result := r;
end;

procedure TDSParaAttributes.SetShading(Value: TDSShading);
var
  Paragraph: TParaFormat2;
begin
  InitPara(Paragraph);
  with Paragraph do
  begin
    dwMask := PFM_SHADING;
    wShadingWeight := Value.wShadingWeight;
    wShadingStyle := Value.wShadingStyle;
  end;
  SetAttributes(Paragraph);
end;

function TDSParaAttributes.GetBorder: TDSBorder;
var
  Paragraph: TParaFormat2;
  r:TDSBorder;
begin
  GetAttributes(Paragraph);
  with (Paragraph) do
   begin
    r.wBorderSpace := wBorderSpace;
    r.wBorderWidth := wBorderWidth;
    r.wBorders := wBorders;
   end;
  Result := r;
end;

procedure TDSParaAttributes.SetBorder(Value: TDSBorder);
var
  Paragraph: TParaFormat2;
begin
  InitPara(Paragraph);
  with Paragraph do
  begin
    dwMask := PFM_BORDER;
    wBorderSpace := Value.wBorderSpace;
    wBorderWidth := Value.wBorderWidth;
    wBorders := Value.wBorders;
  end;
  SetAttributes(Paragraph);
end;

function TDSParaAttributes.GetNumberingStyle: Word;
var
  Paragraph: TParaFormat2;
begin
  GetAttributes(Paragraph);
  Result := Paragraph.wNumberingStyle;
end;

procedure TDSParaAttributes.SetNumberingStyle(Value: Word);
var
  Paragraph: TParaFormat2;
begin
  InitPara(Paragraph);
  with Paragraph do
  begin
    dwMask := PFM_NUMBERINGSTYLE;
    wNumberingStyle := Value;
  end;
  SetAttributes(Paragraph);
end;

function TDSParaAttributes.GetNumberingStart: Word;
var
  Paragraph: TParaFormat2;
begin
  GetAttributes(Paragraph);
  Result := Paragraph.wNumberingStart;
end;

procedure TDSParaAttributes.SetNumberingStart(Value: Word);
var
  Paragraph: TParaFormat2;
begin
  InitPara(Paragraph);
  with Paragraph do
  begin
    dwMask := PFM_NUMBERINGSTART;
    wNumberingStart := Value;
  end;
  SetAttributes(Paragraph);
end;

function TDSParaAttributes.GetNumberingTab: Word;
var
  Paragraph: TParaFormat2;
begin
  GetAttributes(Paragraph);
  Result := Paragraph.wNumberingTab;
end;

procedure TDSParaAttributes.SetNumberingTab(Value: Word);
var
  Paragraph: TParaFormat2;
begin
  InitPara(Paragraph);
  with Paragraph do
  begin
    dwMask := PFM_NUMBERINGTAB;
    wNumberingTab := Value;
  end;
  SetAttributes(Paragraph);
end;


function TDSParaAttributes.GetNumbering: TDSNumberingStyle;
var
  Paragraph: TParaFormat2;
begin
  GetAttributes(Paragraph);
  Result := TDSNumberingStyle(Paragraph.wNumbering);
end;

procedure TDSParaAttributes.SetNumbering(Value: TDSNumberingStyle);
var
  Paragraph: TParaFormat2;
begin
  case Value of
    nsLCLetter..nsSequence : if LeftIndent < 10 then LeftIndent := 10;
    nsNone: LeftIndent := 0;
  end;
  InitPara(Paragraph);
  with Paragraph do
  begin
    dwMask := PFM_NUMBERING;
    wNumbering := Ord(Value);
  end;
  SetAttributes(Paragraph);
end;

function TDSParaAttributes.GetFirstIndent: Longint;
var
  Paragraph: TParaFormat2;
begin
  GetAttributes(Paragraph);
  Result := Paragraph.dxStartIndent div 20
end;

procedure TDSParaAttributes.SetFirstIndent(Value: Longint);
var
  Paragraph: TParaFormat2;
begin
  InitPara(Paragraph);
  with Paragraph do
  begin
    dwMask := PFM_STARTINDENT;
    dxStartIndent := Value * 20;
  end;
  SetAttributes(Paragraph);
end;

function TDSParaAttributes.GetLeftIndent: Longint;
var
  Paragraph: TParaFormat2;
begin
  GetAttributes(Paragraph);
  Result := Paragraph.dxOffset div 20;
end;

procedure TDSParaAttributes.SetLeftIndent(Value: Longint);
var
  Paragraph: TParaFormat2;
begin
  InitPara(Paragraph);
  with Paragraph do
  begin
    dwMask := PFM_OFFSET;
    dxOffset := Value * 20;
  end;
  SetAttributes(Paragraph);
end;

function TDSParaAttributes.GetRightIndent: Longint;
var
  Paragraph: TParaFormat2;
begin
  GetAttributes(Paragraph);
  Result := Paragraph.dxRightIndent div 20;
end;

procedure TDSParaAttributes.SetRightIndent(Value: Longint);
var
  Paragraph: TParaFormat2;
begin
  InitPara(Paragraph);
  with Paragraph do
  begin
    dwMask := PFM_RIGHTINDENT;
    dxRightIndent := Value * 20;
  end;
  SetAttributes(Paragraph);
end;

function TDSParaAttributes.GetTab(Index: Byte): Longint;
var
  Paragraph: TParaFormat2;
begin
  GetAttributes(Paragraph);
  Result := Paragraph.rgxTabs[Index] div 20;
end;

procedure TDSParaAttributes.SetTab(Index: Byte; Value: Longint);
var
  Paragraph: TParaFormat2;
begin
  GetAttributes(Paragraph);
  with Paragraph do
  begin
    rgxTabs[Index] := Value * 20;
    dwMask := PFM_TABSTOPS;
    if cTabCount < Index then cTabCount := Index;
    SetAttributes(Paragraph);
  end;
end;

function TDSParaAttributes.GetTabCount: Integer;
var
  Paragraph: TParaFormat2;
begin
  GetAttributes(Paragraph);
  Result := Paragraph.cTabCount;
end;

procedure TDSParaAttributes.SetTabCount(Value: Integer);
var
  Paragraph: TParaFormat2;
begin
  GetAttributes(Paragraph);
  with Paragraph do
  begin
    dwMask := PFM_TABSTOPS;
    cTabCount := Value;
    SetAttributes(Paragraph);
  end;
end;

procedure TDSParaAttributes.Assign(Source: TPersistent);
var
  I: Integer;
begin
  if Source is TDSParaAttributes then
  begin
    Alignment := TDSParaAttributes(Source).Alignment;
    FirstIndent := TDSParaAttributes(Source).FirstIndent;
    LeftIndent := TDSParaAttributes(Source).LeftIndent;
    RightIndent := TDSParaAttributes(Source).RightIndent;
    Numbering := TDSParaAttributes(Source).Numbering;
    for I := 0 to MAX_TAB_STOPS - 1 do
      Tab[I] := TDSParaAttributes(Source).Tab[I];
  end
  else inherited Assign(Source);
end;

{ TDSConversion }

function TDSConversion.ConvertReadStream(Stream: TStream; Buffer: PChar; BufSize: Integer): Integer;
begin
  Result := Stream.Read(Buffer^, BufSize);
end;

function TDSConversion.ConvertWriteStream(Stream: TStream; Buffer: PChar; BufSize: Integer): Integer;
begin
  Result := Stream.Write(Buffer^, BufSize);
end;

{ TDSRichEditStrings }

const
  ReadError = $0001;
  WriteError = $0002;
  NoError = $0000;

type
  TSelection = record
    StartPos, EndPos: Integer;
  end;

  TDSRichEditStrings = class(TStrings)
  private
    RichEdit: TDSCustomRichEdit;
    FPlainText: Boolean;
    FConverter: TDSConversion;
    procedure EnableChange(const Value: Boolean);
  protected
    function Get(Index: Integer): string; override;
    function GetCount: Integer; override;
    procedure Put(Index: Integer; const S: string); override;
    procedure SetUpdateState(Updating: Boolean); override;
    procedure SetTextStr(const Value: string); override;
  public
    procedure Clear; override;
    procedure AddStrings(Strings: TStrings); override;
    procedure Delete(Index: Integer); override;
    procedure Insert(Index: Integer; const S: string); override;
    procedure LoadFromFile(const FileName: string); override;
    procedure LoadFromStream(Stream: TStream); override;
    procedure SaveToFile(const FileName: string); override;
    procedure SaveToStream(Stream: TStream); override;
    property PlainText: Boolean read FPlainText write FPlainText;
  end;

procedure TDSRichEditStrings.AddStrings(Strings: TStrings);
var
  SelChange: TNotifyEvent;
begin
  SelChange := RichEdit.OnSelectionChange;
  RichEdit.OnSelectionChange := nil;
  try
    inherited AddStrings(Strings);
  finally
    RichEdit.OnSelectionChange := SelChange;
  end;
end;

function TDSRichEditStrings.GetCount: Integer;
begin
  Result := SendMessage(RichEdit.Handle, EM_GETLINECOUNT, 0, 0);
  if SendMessage(RichEdit.Handle, EM_LINELENGTH, SendMessage(RichEdit.Handle,
    EM_LINEINDEX, Result - 1, 0), 0) = 0 then Dec(Result);
end;

function TDSRichEditStrings.Get(Index: Integer): string;
var
  Text: array[0..4095] of Char;
  L: Integer;
begin
  Word((@Text)^) := SizeOf(Text);
  L := SendMessage(RichEdit.Handle, EM_GETLINE, Index, Longint(@Text));
  if (Text[L - 2] = #13) and (Text[L - 1] = #10) then Dec(L, 2);
  SetString(Result, Text, L);
end;

procedure TDSRichEditStrings.Put(Index: Integer; const S: string);
var
  Selection: TCharRange;
begin
  if Index >= 0 then
  begin
    Selection.cpMin := SendMessage(RichEdit.Handle, EM_LINEINDEX, Index, 0);
    if Selection.cpMin <> -1 then
    begin
      Selection.cpMax := Selection.cpMin +
        SendMessage(RichEdit.Handle, EM_LINELENGTH, Selection.cpMin, 0);
      SendMessage(RichEdit.Handle, EM_EXSETSEL, 0, Longint(@Selection));
      SendMessage(RichEdit.Handle, EM_REPLACESEL, 0, Longint(PChar(S)));
    end;
  end;
end;

procedure TDSRichEditStrings.Insert(Index: Integer; const S: string);
var
  L: Integer;
  Selection: TCharRange;
  Fmt: PChar;
  Str: string;
begin
  if Index >= 0 then
  begin
    Selection.cpMin := SendMessage(RichEdit.Handle, EM_LINEINDEX, Index, 0);
    if Selection.cpMin >= 0 then Fmt := '%s'#13#10
    else begin
      Selection.cpMin :=
        SendMessage(RichEdit.Handle, EM_LINEINDEX, Index - 1, 0);
      if Selection.cpMin < 0 then Exit;
      L := SendMessage(RichEdit.Handle, EM_LINELENGTH, Selection.cpMin, 0);
      if L = 0 then Exit;
      Inc(Selection.cpMin, L);
      Fmt := #13#10'%s';
    end;
    Selection.cpMax := Selection.cpMin;
    SendMessage(RichEdit.Handle, EM_EXSETSEL, 0, Longint(@Selection));
    Str := Format(Fmt, [S]);
    SendMessage(RichEdit.Handle, EM_REPLACESEL, 0, LongInt(PChar(Str)));
    {if RichEdit.SelStart <> (Selection.cpMax + Length(Str)) then
      raise EOutOfResources.Create(sRichEditInsertError);}
  end;
end;

procedure TDSRichEditStrings.Delete(Index: Integer);
const
  Empty: PChar = '';
var
  Selection: TCharRange;
begin
  if Index < 0 then Exit;
  Selection.cpMin := SendMessage(RichEdit.Handle, EM_LINEINDEX, Index, 0);
  if Selection.cpMin <> -1 then
  begin
    Selection.cpMax := SendMessage(RichEdit.Handle, EM_LINEINDEX, Index + 1, 0);
    if Selection.cpMax = -1 then
      Selection.cpMax := Selection.cpMin +
        SendMessage(RichEdit.Handle, EM_LINELENGTH, Selection.cpMin, 0);
    SendMessage(RichEdit.Handle, EM_EXSETSEL, 0, Longint(@Selection));
    SendMessage(RichEdit.Handle, EM_REPLACESEL, 0, Longint(Empty));
  end;
end;

procedure TDSRichEditStrings.Clear;
begin
  RichEdit.Clear;
end;

procedure TDSRichEditStrings.SetUpdateState(Updating: Boolean);
begin
  SendMessage(RichEdit.Handle, WM_SETREDRAW, Ord(not Updating), 0);
  if not Updating then begin
    RichEdit.Refresh;
    RichEdit.Perform(CM_TEXTCHANGED, 0, 0);
  end;
end;

procedure TDSRichEditStrings.EnableChange(const Value: Boolean);
var
  EventMask: Longint;
begin
  with RichEdit do
  begin
    if Value then
      EventMask := SendMessage(Handle, EM_GETEVENTMASK, 0, 0) or ENM_CHANGE
    else
      EventMask := SendMessage(Handle, EM_GETEVENTMASK, 0, 0) and not ENM_CHANGE;
    SendMessage(Handle, EM_SETEVENTMASK, 0, EventMask);
  end;
end;

procedure TDSRichEditStrings.SetTextStr(const Value: string);
begin
  EnableChange(False);
  try
    inherited SetTextStr(Value);
  finally
    EnableChange(True);
  end;
end;

function AdjustLineBreaks(Dest, Source: PChar): Integer; assembler;
asm
        PUSH    ESI
        PUSH    EDI
        MOV     EDI,EAX
        MOV     ESI,EDX
        MOV     EDX,EAX
        CLD
@@1:    LODSB
@@2:    OR      AL,AL
        JE      @@4
        CMP     AL,0AH
        JE      @@3
        STOSB
        CMP     AL,0DH
        JNE     @@1
        MOV     AL,0AH
        STOSB
        LODSB
        CMP     AL,0AH
        JE      @@1
        JMP     @@2
@@3:    MOV     EAX,0A0DH
        STOSW
        JMP     @@1
@@4:    STOSB
        LEA     EAX,[EDI-1]
        SUB     EAX,EDX
        POP     EDI
        POP     ESI
end;

function StreamSave(dwCookie: Longint; pbBuff: PByte;
  cb: Longint; var pcb: Longint): Longint; stdcall;
var
  StreamInfo: PDSRichEditStreamInfo;
begin
  Result := NoError;
  StreamInfo := PDSRichEditStreamInfo(Pointer(dwCookie));
  try
    pcb := 0;
    if StreamInfo^.Converter <> nil then
      pcb := StreamInfo^.Converter.ConvertWriteStream(StreamInfo^.Stream, PChar(pbBuff), cb);
  except
    Result := WriteError;
  end;
end;

function StreamLoad(dwCookie: Longint; pbBuff: PByte;
  cb: Longint; var pcb: Longint): Longint; stdcall;
var
  Buffer, pBuff: PChar;
  StreamInfo: PDSRichEditStreamInfo;
begin
  Result := NoError;
  StreamInfo := PDSRichEditStreamInfo(Pointer(dwCookie));
  Buffer := StrAlloc(cb + 1);
  try
    cb := cb div 2;
    pcb := 0;
    pBuff := Buffer + cb;
    try
      if StreamInfo^.Converter <> nil then
        pcb := StreamInfo^.Converter.ConvertReadStream(StreamInfo^.Stream, pBuff, cb);
      if pcb > 0 then
      begin
        pBuff[pcb] := #0;
        if pBuff[pcb - 1] = #13 then pBuff[pcb - 1] := #0;
        pcb := AdjustLineBreaks(Buffer, pBuff);
        Move(Buffer^, pbBuff^, pcb);
      end;
    except
      Result := ReadError;
    end;
  finally
    StrDispose(Buffer);
  end;
end;

procedure TDSRichEditStrings.LoadFromStream(Stream: TStream);
var
  EditStream: TEditStream;
  Position: Longint;
  TextType: Longint;
  StreamInfo: TDSRichEditStreamInfo;
  Converter: TDSConversion;
begin
  StreamInfo.Stream := Stream;
  if FConverter <> nil then
    Converter := FConverter else
    Converter := RichEdit.DefaultConverter.Create;
  StreamInfo.Converter := Converter;
  try
    with EditStream do
    begin
      dwCookie := LongInt(Pointer(@StreamInfo));
      pfnCallBack := @StreamLoad;
      dwError := 0;
    end;
    Position := Stream.Position;
    if PlainText then TextType := SF_TEXT
    else TextType := SF_RTF;
    SendMessage(RichEdit.Handle, EM_STREAMIN, TextType, Longint(@EditStream));
    if (TextType = SF_RTF) and (EditStream.dwError <> 0) then
    begin
      Stream.Position := Position;
      if PlainText then TextType := SF_RTF
      else TextType := SF_TEXT;
      SendMessage(RichEdit.Handle, EM_STREAMIN, TextType, Longint(@EditStream));
      if EditStream.dwError <> 0 then
        raise EOutOfResources.Create(sRichEditLoadFail);
    end;
  finally
    if FConverter = nil then Converter.Free;
  end;
end;

procedure TDSRichEditStrings.SaveToStream(Stream: TStream);
var
  EditStream: TEditStream;
  TextType: Longint;
  StreamInfo: TDSRichEditStreamInfo;
  Converter: TDSConversion;
begin
  if FConverter <> nil then
    Converter := FConverter else
    Converter := RichEdit.DefaultConverter.Create;
  StreamInfo.Stream := Stream;
  StreamInfo.Converter := Converter;
  try
    with EditStream do
    begin
      dwCookie := LongInt(Pointer(@StreamInfo));
      pfnCallBack := @StreamSave;
      dwError := 0;
    end;
    if PlainText then TextType := SF_TEXT
    else TextType := SF_RTF;
    SendMessage(RichEdit.Handle, EM_STREAMOUT, TextType, Longint(@EditStream));
    if EditStream.dwError <> 0 then
      raise EOutOfResources.Create(sRichEditSaveFail);
  finally
    if FConverter = nil then Converter.Free;
  end;
end;

procedure TDSRichEditStrings.LoadFromFile(const FileName: string);
var
  Ext: string;
  Convert: PDSConversionFormat;
begin
  Ext := AnsiLowerCaseFileName(ExtractFileExt(Filename));
  System.Delete(Ext, 1, 1);
  Convert := ConversionFormatList;
  while Convert <> nil do
    with Convert^ do
      if Extension <> Ext then Convert := Next
      else Break;
  if Convert = nil then
    Convert := @TexTDSConversionFormat;
  FConverter := Convert^.ConversionClass.Create;
  try
    inherited LoadFromFile(FileName);
  except
    FConverter.Free;
    FConverter := nil;
    raise;
  end;
end;

procedure TDSRichEditStrings.SaveToFile(const FileName: string);
var
  Ext: string;
  Convert: PDSConversionFormat;
begin
  Ext := AnsiLowerCaseFileName(ExtractFileExt(Filename));
  System.Delete(Ext, 1, 1);
  Convert := ConversionFormatList;
  while Convert <> nil do
    with Convert^ do
      if Extension <> Ext then Convert := Next
      else Break;
  if Convert = nil then
    Convert := @TexTDSConversionFormat;
  FConverter := Convert^.ConversionClass.Create;
  try
    inherited SaveToFile(FileName);
  except
    FConverter.Free;
    FConverter := nil;
    raise;
  end;
end;

{ TDSRichEdit }

constructor TDSCustomRichEdit.Create(AOwner: TComponent);
var
  DC: HDC;
begin
  inherited Create(AOwner);
  FSelAttributes := TDSTextAttributes.Create(Self, atSelected);
  FDefAttributes := TDSTextAttributes.Create(Self, atDefaultText);
  FParagraph := TDSParaAttributes.Create(Self);
  FRichEditStrings := TDSRichEditStrings.Create;
  TDSRichEditStrings(FRichEditStrings).RichEdit := Self;
  TabStop := True;
  Width := 185;
  Height := 89;
  AutoSize := False;
  FHideSelection := True;
  HideScrollBars := True;
  DC := GetDC(0);
  FScreenLogPixels := GetDeviceCaps(DC, LOGPIXELSY);
  DefaultConverter := TDSConversion;
  ReleaseDC(0, DC);
end;

destructor TDSCustomRichEdit.Destroy;
begin
  FSelAttributes.Free;
  FDefAttributes.Free;
  FParagraph.Free;
  FRichEditStrings.Free;
  FMemStream.Free;
  inherited Destroy;
end;

procedure TDSCustomRichEdit.Clear;
begin
  inherited Clear;
  Modified := False;
end;

procedure TDSCustomRichEdit.CreateParams(var Params: TCreateParams);
const
  RichEditModuleName = 'RICHED20.DLL';
  HideScrollBars: array[Boolean] of Longint = (ES_DISABLENOSCROLL, 0);
  HideSelections: array[Boolean] of Longint = (ES_NOHIDESEL, 0);
var
  OldError: Longint;
begin
  OldError := SetErrorMode(SEM_NOOPENFILEERRORBOX);
  FLibHandle := LoadLibrary(RichEditModuleName);
  if (FLibHandle > 0) and (FLibHandle < HINSTANCE_ERROR) then FLibHandle := 0;
  SetErrorMode(OldError);
  inherited CreateParams(Params);
  CreateSubClass(Params, RICHEDIT_CLASS);
  with Params do
  begin
    Style := Style or HideScrollBars[FHideScrollBars] or
      HideSelections[HideSelection];
    WindowClass.style := WindowClass.style and not (CS_HREDRAW or CS_VREDRAW);
  end;
end;

procedure TDSCustomRichEdit.CreateWnd;
var
  Plain: Boolean;
begin
  inherited CreateWnd;
  if (SysLocale.FarEast) and not (SysLocale.PriLangID = LANG_JAPANESE) then
    Font.Charset := GetDefFontCharSet;
  SendMessage(Handle, EM_SETEVENTMASK, 0,
    ENM_CHANGE or ENM_SELCHANGE or ENM_REQUESTRESIZE or
    ENM_PROTECTED);
  SendMessage(Handle, EM_SETBKGNDCOLOR, 0, ColorToRGB(Color));
  if FMemStream <> nil then
  begin
    Plain := PlainText;
    PlainText := False;
    try
      Lines.LoadFromStream(FMemStream);
      FMemStream.Free;
      FMemStream := nil;
    finally
      PlainText := Plain;
    end;
  end;
  Modified := FModified;
end;

procedure TDSCustomRichEdit.DestroyWnd;
var
  Plain: Boolean;
begin
  FModified := Modified;
  FMemStream := TMemoryStream.Create;
  Plain := PlainText;
  PlainText := False;
  try
    Lines.SaveToStream(FMemStream);
    FMemStream.Position := 0;
  finally
    PlainText := Plain;
  end;
  inherited DestroyWnd;
end;

procedure TDSCustomRichEdit.WMNCDestroy(var Message: TWMNCDestroy);
begin
  inherited;
  if FLibHandle <> 0 then FreeLibrary(FLibHandle);
end;

procedure TDSCustomRichEdit.WMSetFont(var Message: TWMSetFont);
begin
  FDefAttributes.Assign(Font);
end;

procedure TDSCustomRichEdit.CMFontChanged(var Message: TMessage);
begin
  FDefAttributes.Assign(Font);
end;

procedure TDSCustomRichEdit.DoSetMaxLength(Value: Integer);
begin
  SendMessage(Handle, EM_EXLIMITTEXT, 0, Value);
end;

function TDSCustomRichEdit.GetSelLength: Integer;
var
  CharRange: TCharRange;
begin
  SendMessage(Handle, EM_EXGETSEL, 0, Longint(@CharRange));
  Result := CharRange.cpMax - CharRange.cpMin;
end;

function TDSCustomRichEdit.GetSelStart: Integer;
var
  CharRange: TCharRange;
begin
  SendMessage(Handle, EM_EXGETSEL, 0, Longint(@CharRange));
  Result := CharRange.cpMin;
end;

function TDSCustomRichEdit.GetSelTextBuf(Buffer: PChar; BufSize: Integer): Integer;
var
  S: string;
begin
  S := GetSelText;
  Result := Length(S);
  if BufSize < Length(S) then Result := BufSize;
  StrPLCopy(Buffer, S, Result);
end;

function TDSCustomRichEdit.GetSelText: string;
var
  Length: Integer;
begin
  SetLength(Result, GetSelLength + 1);
  Length := SendMessage(Handle, EM_GETSELTEXT, 0, Longint(PChar(Result)));
  SetLength(Result, Length);
end;

procedure TDSCustomRichEdit.SetHideScrollBars(Value: Boolean);
begin
  if HideScrollBars <> Value then
  begin
    FHideScrollBars := value;
    RecreateWnd;
  end;
end;

procedure TDSCustomRichEdit.SetHideSelection(Value: Boolean);
begin
  if HideSelection <> Value then
  begin
    FHideSelection := Value;
    SendMessage(Handle, EM_HIDESELECTION, Ord(HideSelection), LongInt(True));
  end;
end;

procedure TDSCustomRichEdit.SetSelAttributes(Value: TDSTextAttributes);
begin
  SelAttributes.Assign(Value);
end;

procedure TDSCustomRichEdit.SetSelLength(Value: Integer);
var
  CharRange: TCharRange;
begin
  SendMessage(Handle, EM_EXGETSEL, 0, Longint(@CharRange));
  CharRange.cpMax := CharRange.cpMin + Value;
  SendMessage(Handle, EM_EXSETSEL, 0, Longint(@CharRange));
  SendMessage(Handle, EM_SCROLLCARET, 0, 0);
end;

procedure TDSCustomRichEdit.SetDefAttributes(Value: TDSTextAttributes);
begin
  DefAttributes.Assign(Value);
end;

function TDSCustomRichEdit.GetPlainText: Boolean;
begin
  Result := TDSRichEditStrings(Lines).PlainText;
end;

procedure TDSCustomRichEdit.SetPlainText(Value: Boolean);
begin
  TDSRichEditStrings(Lines).PlainText := Value;
end;

procedure TDSCustomRichEdit.CMColorChanged(var Message: TMessage);
begin
  inherited;
  SendMessage(Handle, EM_SETBKGNDCOLOR, 0, ColorToRGB(Color))
end;

procedure TDSCustomRichEdit.SeTDSRichEditStrings(Value: TStrings);
begin
  FRichEditStrings.Assign(Value);
end;

procedure TDSCustomRichEdit.SetSelStart(Value: Integer);
var
  CharRange: TCharRange;
begin
  CharRange.cpMin := Value;
  CharRange.cpMax := Value;
  SendMessage(Handle, EM_EXSETSEL, 0, Longint(@CharRange));
end;

procedure TDSCustomRichEdit.Print(const Caption: string);
var
  Range: TFormatRange;
  LastChar, MaxLen, LogX, LogY, OldMap: Integer;
begin
  FillChar(Range, SizeOf(TFormatRange), 0);
  with Printer, Range do
  begin
    BeginDoc;
    hdc := Handle;
    hdcTarget := hdc;
    LogX := GetDeviceCaps(Handle, LOGPIXELSX);
    LogY := GetDeviceCaps(Handle, LOGPIXELSY);
    if IsRectEmpty(PageRect) then
    begin
      rc.right := PageWidth * 1440 div LogX;
      rc.bottom := PageHeight * 1440 div LogY;
    end
    else begin
      rc.left := PageRect.Left * 1440 div LogX;
      rc.top := PageRect.Top * 1440 div LogY;
      rc.right := PageRect.Right * 1440 div LogX;
      rc.bottom := PageRect.Bottom * 1440 div LogY;
    end;
    rcPage := rc;
    Title := Caption;
    LastChar := 0;
    MaxLen := GetTextLen;
    chrg.cpMax := -1;
    // ensure printer DC is in text map mode
    OldMap := SetMapMode(hdc, MM_TEXT);
    SendMessage(Handle, EM_FORMATRANGE, 0, 0);    // flush buffer
    try
      repeat
        chrg.cpMin := LastChar;
        LastChar := SendMessage(Self.Handle, EM_FORMATRANGE, 1, Longint(@Range));
        if (LastChar < MaxLen) and (LastChar <> -1) then NewPage;
      until (LastChar >= MaxLen) or (LastChar = -1);
      EndDoc;
    finally
      SendMessage(Handle, EM_FORMATRANGE, 0, 0);  // flush buffer
      SetMapMode(hdc, OldMap);       // restore previous map mode
    end;
  end;
end;

var
  Painting: Boolean = False;

procedure TDSCustomRichEdit.WMPaint(var Message: TWMPaint);
var
  R, R1: TRect;
begin
  if GetUpdateRect(Handle, R, True) then
  begin
    with ClientRect do R1 := Rect(Right - 3, Top, Right, Bottom);
    if IntersectRect(R, R, R1) then InvalidateRect(Handle, @R1, True);
  end;
  if Painting then
    Invalidate
  else begin
    Painting := True;
    try
      inherited;
    finally
      Painting := False;
    end;
  end;
end;

procedure TDSCustomRichEdit.WMSetCursor(var Message: TWMSetCursor);
var
  P: TPoint;
begin
  inherited;
  if Message.Result = 0 then
  begin
    Message.Result := 1;
    GetCursorPos(P);
    with PointToSmallPoint(P) do
      case Perform(WM_NCHITTEST, 0, MakeLong(X, Y)) of
        HTVSCROLL,
        HTHSCROLL:
          Windows.SetCursor(Screen.Cursors[crArrow]);
        HTCLIENT:
          Windows.SetCursor(Screen.Cursors[crIBeam]);
      end;
  end;
end;

procedure TDSCustomRichEdit.CNNotify(var Message: TWMNotify);
begin
  with Message.NMHdr^ do
    case code of
      EN_SELCHANGE: SelectionChange;
      EN_REQUESTRESIZE: RequestSize(PReqSize(Pointer(Message.NMHdr))^.rc);
      EN_SAVECLIPBOARD:
        with PENSaveClipboard(Pointer(Message.NMHdr))^ do
          if not SaveClipboard(cObjectCount, cch) then Message.Result := 1;
      EN_PROTECTED:
        with PENProtected(Pointer(Message.NMHdr))^.chrg do
          if not ProtectChange(cpMin, cpMax) then Message.Result := 1;
    end;
end;

function TDSCustomRichEdit.SaveClipboard(NumObj, NumChars: Integer): Boolean;
begin
  Result := True;
  if Assigned(OnSaveClipboard) then OnSaveClipboard(Self, NumObj, NumChars, Result);
end;

function TDSCustomRichEdit.ProtectChange(StartPos, EndPos: Integer): Boolean;
begin
  Result := False;
  if Assigned(OnProtectChange) then OnProtectChange(Self, StartPos, EndPos, Result);
end;

procedure TDSCustomRichEdit.SelectionChange;
begin
  if Assigned(OnSelectionChange) then OnSelectionChange(Self);
end;

procedure TDSCustomRichEdit.RequestSize(const Rect: TRect);
begin
  if Assigned(OnResizeRequest) then OnResizeRequest(Self, Rect);
end;

function TDSCustomRichEdit.FindText(const SearchStr: string;
  StartPos, Length: Integer; Options: TDSSearchTypes): Integer;
var
  Find: TFindText;
  Flags: Integer;
begin
  with Find.chrg do
  begin
    cpMin := StartPos;
    cpMax := cpMin + Length;
  end;
  Flags := 0;
  if stWholeWord in Options then Flags := Flags or FT_WHOLEWORD;
  if stMatchCase in Options then Flags := Flags or FT_MATCHCASE;
  Find.lpstrText := PChar(SearchStr);
  Result := SendMessage(Handle, EM_FINDTEXT, Flags, LongInt(@Find));
end;

procedure AppendConversionFormat(const Ext: string; AClass: TDSConversionClass);
var
  NewRec: PDSConversionFormat;
begin
  New(NewRec);
  with NewRec^ do
  begin
    Extension := AnsiLowerCaseFileName(Ext);
    ConversionClass := AClass;
    Next := ConversionFormatList;
  end;
  ConversionFormatList := NewRec;
end;

class procedure TDSCustomRichEdit.RegisterConversionFormat(const AExtension: string;
  AConversionClass: TDSConversionClass);
begin
  AppendConversionFormat(AExtension, AConversionClass);
end;


initialization

finalization

end.
