unit QRTView;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  quickrpt,OrgChart;

type

  TQRMOrgChart= class(TCustomOrgChart)
  private FBorderStyle:TBorderStyle;
  public
    property BorderStyle:TBorderStyle read FBorderStyle write FBorderStyle;
    property Alignment;
  end;

  TQRCustomOrgChart = class(TQRPrintable)
  private
    { Private declarations }
    FOrgChart:TQRMOrgChart;
    FParentOrgChart:TCustomOrgChart;
    FAutoStretch : boolean;
    function GetAlignment : TAlignment;
    procedure SetAlignment(Value : TAlignment); override;
    procedure SetParentOrgChart(Value : TCustomOrgChart);
  protected
    { Protected declarations }
    procedure Print(OfsX, OfsY : integer); override;
    procedure SetBounds(ALeft, ATop, AWidth, AHeight : integer); override;
    property  ParentOrgChart : TCustomOrgChart read FParentOrgChart write SetParentOrgChart;
  public
    { Public declarations }
    constructor Create(AOwner : TComponent); override;
    destructor Destroy; override;
  published
    { Published declarations }
    property Alignment : TAlignment read GetAlignment write SetAlignment;
    property AutoStretch : boolean read FAutoStretch write FAutoStretch;
  end;

procedure Register;

implementation

constructor TQRCustomOrgChart.Create(AOwner : TComponent);
begin
  FOrgChart := nil;
  inherited Create(AOwner);
  FOrgChart := TQRMOrgChart.Create(Self);
  FOrgChart.Parent := self;
  FOrgChart.BorderStyle := bsNone;
  AutoStretch := false;
  Width := 100;
  Height := 100;
end;

destructor TQRCustomOrgChart.Destroy;
begin
  FOrgChart.Free;
  inherited Destroy;
end;

function TQRCustomOrgChart.GetAlignment : TAlignment;
begin
  result := FOrgChart.Alignment;
end;

procedure TQRCustomOrgChart.Print(OfsX, OfsY : integer);
var
  Range: TFormatRange;
  LastChar, MaxLen, LogX, LogY: Integer;
  AOrgChart : TCustomOrgChart;
  ChangeX, Y : extended;
  VSize : extended;
begin
  if assigned(FParentOrgChart) then
    AOrgChart := FParentOrgChart
  else
    AOrgChart := FOrgChart;
  FillChar(Range, SizeOf(TFormatRange), 0);
  with Range do
  begin
    hdc := ParentReport.QRPrinter.Canvas.Handle;
    hdcTarget := hdc;
    LogX := GetDeviceCaps(hdc, LOGPIXELSX);
    LogY := GetDeviceCaps(hdc, LOGPIXELSY);
    rc.Left := QRPrinter.XPos(OfsX + Size.Left)* 1440 div LogX;
    rc.Top := QRPrinter.YPos(OfsY + Size.Top) * 1440 div LogY;
    rc.Right := QRPrinter.XPos(OfsX + Size.Width + Size.Left) *1440 div LogX;
    rc.Bottom := QRPrinter.YPos(OfsY + Size.Height + Size.Top) * 1440 div LogY;
    rcPage := rc;
    LastChar := 0;
    MaxLen := AOrgChart.GetTextLen;
    chrg.cpMax := -1;
    chrg.cpMin := LastChar;
    Y := OfsY + Size.Top;
    VSize := OfsY+Size.Top + Size.Height;
    if AutoStretch then
    begin
      LastChar := SendMessage(AOrgChart.Handle, EM_FORMATRANGE, 0, Longint(@Range));
      while (LastChar < MaxLen) and (LastChar <> -1) do
      begin
        if TQRCustomBand(Parent).CanExpand(50) then
        begin
          TQRCustomBand(Parent).ExpandBand(50, Y, ChangeX);
          VSize := VSize + 50;
          rc.Bottom := QRPrinter.YPos(VSize) * 1440 div LogY;
          LastChar := SendMessage(AOrgChart.Handle, EM_FORMATRANGE, 0, Longint(@Range));
          if (LastChar >= MaxLen) or (LastChar = -1) then
          begin
            LastChar := SendMessage(AOrgChart.Handle, EM_FORMATRANGE, 1, Longint(@Range));
            chrg.cpMin := LastChar;
          end;
        end else
        begin
          LastChar := SendMessage(AOrgChart.Handle, EM_FORMATRANGE, 1, Longint(@Range));
          TQRCustomBand(Parent).ExpandBand(50, Y, ChangeX);
          VSize := Y + 50;
          rc.Top := QRPrinter.YPos(Y) * 1440 div LogY;
          rc.Bottom := QRPrinter.YPos(VSize) * 1440 div LogY;
          rc.Left := rc.Left + QRPrinter.XSize(ChangeX) * 1440 div LogX;
          rc.Right := rc.Right + QRPrinter.XSize(ChangeX) * 1440 div LogX;
          chrg.cpMin := LastChar;
          hdc := ParentReport.QRPrinter.Canvas.Handle;
          hdcTarget := hdc;
         end;
      end;
    end else
      SendMessage(AOrgChart.Handle, EM_FORMATRANGE, 1, Longint(@Range));
  end;
  SendMessage(FOrgChart.Handle, EM_FORMATRANGE, 0, 0);
end;

procedure TQRCustomOrgChart.SetAlignment(Value : TAlignment);
begin
  FOrgChart.Alignment := Value;
end;

procedure TQRCustomOrgChart.SetBounds(ALeft, ATop, AWidth, AHeight : integer);
begin
  inherited SetBounds(ALeft, ATop, AWidth, AHeight);
  if FOrgChart <> nil then FOrgChart.SetBounds(1, 1, AWidth - 2, AHeight - 2);
end;

procedure TQRCustomOrgChart.SetParentOrgChart(Value : TCustomOrgChart);
begin
  FParentOrgChart := Value;
{  if Value<>nil then
    FOrgChart.Lines := Value.Lines; }
end;


procedure Register;
begin
  RegisterComponents('Samples', [TQRCustomTreeView]);
end;

end.
