Unit Glu;

Interface

Uses Windows,Gl;

{ OpenGL Utility routines (glu.h) =======================================}

function gluErrorString (errCode: GLenum): PChar; stdcall;
function gluErrorUnicodeStringEXT (errCode: GLenum): PWChar; stdcall;
function gluGetString (name: GLenum): PChar; stdcall;

procedure gluLookAt(eyex, eyey, eyez,
                    centerx, centery, centerz,
                    upx, upy, upz: GLdouble); stdcall;
procedure gluOrtho2D(left, right, bottom, top: GLdouble); stdcall;
procedure gluPerspective(fovy, aspect, zNear, zFar: GLdouble); stdcall;
procedure gluPickMatrix (x, y, width, height: GLdouble; viewport: PGLint); stdcall;
function  gluProject (objx, objy, obyz: GLdouble;
                      modelMatrix: PGLdouble;
                      projMatrix: PGLdouble;
                      viewport: PGLint;
                      var winx, winy, winz: GLDouble): Integer; stdcall;
function  gluUnProject(winx, winy, winz: GLdouble;
                      modelMatrix: PGLdouble;
                      projMatrix: PGLdouble;
                      viewport: PGLint;
                      var objx, objy, objz: GLdouble): Integer; stdcall;
function  gluScaleImage(format: GLenum;
   widthin, heightin: GLint; typein: GLenum; datain: Pointer;
   widthout, heightout: GLint; typeout: GLenum; dataout: Pointer): Integer; stdcall;

function  gluBuild1DMipmaps (target: GLenum; components, width: GLint;
                             format, atype: GLenum; data: Pointer): Integer; stdcall;
function  gluBuild2DMipmaps (target: GLenum; components, width: GLint;
                             format, atype: GLenum; data: Pointer): Integer; stdcall;

type
  _GLUquadricObj = record end;
  GLUquadricObj = ^_GLUquadricObj;

  TGLUquadricErrorProc = procedure (error: GLenum) stdcall;

function  gluNewQuadric: GLUquadricObj; stdcall;
procedure gluDeleteQuadric (state: GLUquadricObj); stdcall;
procedure gluQuadricNormals (quadObject: GLUquadricObj; normals: GLenum);  stdcall;
procedure gluQuadricTexture (quadObject: GLUquadricObj; textureCoords: GLboolean );stdcall;
procedure gluQuadricOrientation (quadObject: GLUquadricObj; orientation: GLenum); stdcall;
procedure gluQuadricDrawStyle (quadObject: GLUquadricObj; drawStyle: GLenum); stdcall;
procedure gluCylinder (quadObject: GLUquadricObj;
  baseRadius, topRadius, height: GLdouble; slices, stacks: GLint); stdcall;
procedure gluDisk (quadObject: GLUquadricObj;
  innerRadius, outerRadius: GLdouble; slices, loops: GLint); stdcall;
procedure gluPartialDisk (quadObject: GLUquadricObj;
  innerRadius, outerRadius: GLdouble; slices, loops: GLint;
  startAngle, sweepAngle: GLdouble); stdcall;
procedure gluSphere (quadObject: GLUquadricObj; radius: GLdouble; slices, loops: GLint); stdcall;
procedure gluQuadricCallback (quadObject: GLUquadricObj; which: GLenum;
  callback: Pointer); stdcall;

type
  _GLUtesselator = record end;
  GLUtesselator = ^_GLUtesselator;

  // tesselator callback procedure types
  TGLUtessBeginProc = procedure (a: GLenum) stdcall;
  TGLUtessEdgeFlagProc = procedure (flag: GLboolean) stdcall;
  TGLUtessVertexProc = procedure (p: Pointer) stdcall;
  TGLUtessEndProc = procedure stdcall;
  TGLUtessErrorProc = TGLUquadricErrorProc;
  TGLUtessCombineProc = procedure (a: PGLdouble; b: Pointer;
                                   c: PGLfloat; var d: Pointer) stdcall;

function gluNewTess: GLUtesselator; stdcall;
procedure gluDeleteTess( tess: GLUtesselator ); stdcall;
procedure gluTessBeginPolygon( tess: GLUtesselator ); stdcall;
procedure gluTessBeginContour( tess: GLUtesselator ); stdcall;
procedure gluTessVertex( tess: GLUtesselator; coords: PGLdouble; data: Pointer ); stdcall;
procedure gluTessEndContour( tess: GLUtesselator ); stdcall;
procedure gluTessEndPolygon( tess: GLUtesselator ); stdcall;
procedure gluTessProperty( tess: GLUtesselator; which: GLenum; value: GLdouble); stdcall;
procedure gluTessNormal( tess: GLUtesselator; x,y,z: GLdouble); stdcall;
procedure gluTessCallback( tess: GLUtesselator; which: GLenum; callback: pointer); stdcall;

type
  TGLUnurbsObj = record end;
  GLUnurbsObj = ^TGLUnurbsObj;

  TGLUnurbsErrorProc = TGLUquadricErrorProc;

function gluNewNurbsRenderer: GLUnurbsObj; stdcall;
procedure gluDeleteNurbsRenderer (nobj: GLUnurbsObj); stdcall;
procedure gluBeginSurface (nobj: GLUnurbsObj); stdcall;
procedure gluBeginCurve (nobj: GLUnurbsObj); stdcall;
procedure gluEndCurve (nobj: GLUnurbsObj); stdcall;
procedure gluEndSurface (nobj: GLUnurbsObj); stdcall;
procedure gluBeginTrim (nobj: GLUnurbsObj); stdcall;
procedure gluEndTrim (nobj: GLUnurbsObj); stdcall;
procedure gluPwlCurve (nobj: GLUnurbsObj; count: GLint; points: PGLfloat;
  stride: GLint; _type: GLenum); stdcall;
procedure gluNurbsCurve (nobj: GLUnurbsObj; nknots: GLint; knot: PGLfloat;
  stride: GLint; ctlpts: PGLfloat; order: GLint; _type: GLenum); stdcall;
procedure gluNurbsSurface (nobj: GLUnurbsObj;
  sknot_count: GLint; sknot: PGLfloat;
  tknot_count: GLint; tknot: PGLfloat;
  s_stride, t_stride: GLint;
  ctlpts: PGLfloat; sorder, torder: GLint; _type: GLenum); stdcall;
procedure gluLoadSamplingMatrices (nobj: GLUnurbsObj;
  modelMatrix: PGLdouble; projMatrix: PGLdouble; viewport: PGLint); stdcall;
procedure gluNurbsProperty (nobj: GLUnurbsObj; prop: GLenum; value: GLfloat); stdcall;
procedure gluGetNurbsProperty (nobj: GLUnurbsObj; prop: GLenum; var value: GLfloat); stdcall;
procedure gluNurbsCallback (nobj: GLUnurbsObj; which: GLenum; callback: pointer); stdcall;
procedure gluDeleteNurbsTessellatorEXT (nurb: GLUnurbsObj); stdcall;
procedure gluGetTessProperty (tess: GLUtesselator;
                     which: GLenum; data: PGLdouble); stdcall;
function  gluNewNurbsTessellatorEXT:GLUnurbsObj; stdcall;


{****           Generic constants               ****}
const

{ Extensions }
  GLU_EXT_object_space_tess          = 1;
  GLU_EXT_nurbs_tessellator          = 1;

{ Version }
  GLU_VERSION_1_2  =               1;
  GLU_VERSION_1_1  =               1;

{ Errors: (return value 0 = no error) }
  GLU_INVALID_ENUM       = 100900;
  GLU_INVALID_VALUE      = 100901;
  GLU_OUT_OF_MEMORY      = 100902;
  GLU_INCOMPATIBLE_GL_VERSION  =   100903;
  GLU_INVALID_OPERATION        =   100904;

{ gets }
  GLU_VERSION            = 100800;
  GLU_EXTENSIONS         = 100801;

{ For laughs: }
  GLU_TRUE               = GL_TRUE;
  GLU_FALSE              = GL_FALSE;

{***           Quadric constants               ***}

{ Types of normals: }
  GLU_SMOOTH             = 100000;
  GLU_FLAT               = 100001;
  GLU_NONE               = 100002;

{ DrawStyle types: }
  GLU_POINT              = 100010;
  GLU_LINE               = 100011;
  GLU_FILL               = 100012;
  GLU_SILHOUETTE         = 100013;

{ Orientation types: }
  GLU_OUTSIDE            = 100020;
  GLU_INSIDE             = 100021;

{ Callback types: }
{      GLU_ERROR               100103 }


{***           Tesselation constants           ***}

  GLU_TESS_MAX_COORD     =         1.0e150;

{ Property types: }
  GLU_TESS_WINDING_RULE  =         100110;
  GLU_TESS_BOUNDARY_ONLY =         100111;
  GLU_TESS_TOLERANCE     =         100112;

{ Possible winding rules: }
  GLU_TESS_WINDING_ODD          =  100130;
  GLU_TESS_WINDING_NONZERO      =  100131;
  GLU_TESS_WINDING_POSITIVE     =  100132;
  GLU_TESS_WINDING_NEGATIVE     =  100133;
  GLU_TESS_WINDING_ABS_GEQ_TWO  =  100134;


{ Callback types: }
  GLU_TESS_BEGIN     = 100100 ;     { void (*)(GLenum    type)         }
  GLU_BEGIN          = 100100 ;
  GLU_TESS_VERTEX    = 100101 ;     { void (*)(void      *data)        }
  GLU_VERTEX         = 100101 ;
  GLU_TESS_END       = 100102 ;     { void (*)(void)                   }
  GLU_END            = 100102 ;
  GLU_TESS_ERROR     = 100103 ;     { void (*)(GLenum    errno)        }
  GLU_TESS_EDGE_FLAG = 100104 ;     { void (*)(GLboolean boundaryEdge) }
  GLU_EDGE_FLAG      = 100104 ;
  GLU_TESS_COMBINE   = 100105 ;     { void (*)(GLdouble  coords[3],;
                                                    void      *data[4],;
                                                    GLfloat   weight[4],;
                                                    void      **dataOut)    }

  GLU_TESS_BEGIN_DATA     =             100106;
  GLU_TESS_VERTEX_DATA    =             100107;
  GLU_TESS_END_DATA       =             100108;
  GLU_TESS_ERROR_DATA     =             100109;
  GLU_TESS_EDGE_FLAG_DATA =             100110;
  GLU_TESS_COMBINE_DATA   =             100111;
  GLU_NURBS_MODE_EXT               =    100160;
  GLU_NURBS_TESSELLATOR_EXT        =    100161;
  GLU_NURBS_RENDERER_EXT           =    100162;
  GLU_NURBS_BEGIN_EXT              =    100164;
  GLU_NURBS_VERTEX_EXT             =    100165;
  GLU_NURBS_NORMAL_EXT             =    100166;
  GLU_NURBS_COLOR_EXT              =    100167;
  GLU_NURBS_TEX_COORD_EXT          =    100168;
  GLU_NURBS_END_EXT                =    100169;
  GLU_NURBS_BEGIN_DATA_EXT         =    100170;
  GLU_NURBS_VERTEX_DATA_EXT        =    100171;
  GLU_NURBS_NORMAL_DATA_EXT        =    100172;
  GLU_NURBS_COLOR_DATA_EXT         =    100173;
  GLU_NURBS_TEX_COORD_DATA_EXT     =    100174;
  GLU_NURBS_END_DATA_EXT           =    100175;

{ TessContour }
  GLU_CW                           =    100120;
  GLU_CCW                          =    100121;
  GLU_INTERIOR                     =    100122;
  GLU_EXTERIOR                     =    100123;
  GLU_UNKNOWN                      =    100124;


{ Errors: }
  GLU_TESS_ERROR1    = 100151;
  GLU_TESS_ERROR2    = 100152;
  GLU_TESS_ERROR3    = 100153;
  GLU_TESS_ERROR4    = 100154;
  GLU_TESS_ERROR5    = 100155;
  GLU_TESS_ERROR6    = 100156;
  GLU_TESS_ERROR7    = 100157;
  GLU_TESS_ERROR8    = 100158;

  GLU_TESS_MISSING_BEGIN_POLYGON  = GLU_TESS_ERROR1;
  GLU_TESS_MISSING_BEGIN_CONTOUR  = GLU_TESS_ERROR2;
  GLU_TESS_MISSING_END_POLYGON    = GLU_TESS_ERROR3;
  GLU_TESS_MISSING_END_CONTOUR    = GLU_TESS_ERROR4;
  GLU_TESS_COORD_TOO_LARGE        = GLU_TESS_ERROR5;
  GLU_TESS_NEED_COMBINE_CALLBACK  = GLU_TESS_ERROR6;

{***           NURBS constants                 ***}

{ Properties: }
  GLU_AUTO_LOAD_MATRIX          =  100200;
  GLU_CULLING                   =  100201;
  GLU_SAMPLING_TOLERANCE        =  100203;
  GLU_DISPLAY_MODE              =  100204;
  GLU_PARAMETRIC_TOLERANCE      =  100202;
  GLU_SAMPLING_METHOD           =  100205;
  GLU_U_STEP                    =  100206;
  GLU_V_STEP                    =  100207;

{ Sampling methods: }
  GLU_OBJECT_PARAMETRIC_ERROR_EXT = 100208;
  GLU_OBJECT_PATH_LENGTH_EXT      = 100209;
  GLU_PATH_LENGTH               =  100215;
  GLU_PARAMETRIC_ERROR          =  100216;
  GLU_DOMAIN_DISTANCE           =  100217;

{ Trimming curve types }
  GLU_MAP1_TRIM_2       =  100210;
  GLU_MAP1_TRIM_3       =  100211;

{ Display modes: }
{      GLU_FILL                100012 }
  GLU_OUTLINE_POLYGON    = 100240;
  GLU_OUTLINE_PATCH      = 100241;

{ Callbacks: }
{      GLU_ERROR               100103 }
    GLU_ERROR            = 100103;

{ Errors: }
  GLU_NURBS_ERROR1       = 100251;
  GLU_NURBS_ERROR2       = 100252;
  GLU_NURBS_ERROR3       = 100253;
  GLU_NURBS_ERROR4       = 100254;
  GLU_NURBS_ERROR5       = 100255;
  GLU_NURBS_ERROR6       = 100256;
  GLU_NURBS_ERROR7       = 100257;
  GLU_NURBS_ERROR8       = 100258;
  GLU_NURBS_ERROR9       = 100259;
  GLU_NURBS_ERROR10      = 100260;
  GLU_NURBS_ERROR11      = 100261;
  GLU_NURBS_ERROR12      = 100262;
  GLU_NURBS_ERROR13      = 100263;
  GLU_NURBS_ERROR14      = 100264;
  GLU_NURBS_ERROR15      = 100265;
  GLU_NURBS_ERROR16      = 100266;
  GLU_NURBS_ERROR17      = 100267;
  GLU_NURBS_ERROR18      = 100268;
  GLU_NURBS_ERROR19      = 100269;
  GLU_NURBS_ERROR20      = 100270;
  GLU_NURBS_ERROR21      = 100271;
  GLU_NURBS_ERROR22      = 100272;
  GLU_NURBS_ERROR23      = 100273;
  GLU_NURBS_ERROR24      = 100274;
  GLU_NURBS_ERROR25      = 100275;
  GLU_NURBS_ERROR26      = 100276;
  GLU_NURBS_ERROR27      = 100277;
  GLU_NURBS_ERROR28      = 100278;
  GLU_NURBS_ERROR29      = 100279;
  GLU_NURBS_ERROR30      = 100280;
  GLU_NURBS_ERROR31      = 100281;
  GLU_NURBS_ERROR32      = 100282;
  GLU_NURBS_ERROR33      = 100283;
  GLU_NURBS_ERROR34      = 100284;
  GLU_NURBS_ERROR35      = 100285;
  GLU_NURBS_ERROR36      = 100286;
  GLU_NURBS_ERROR37      = 100287;

{
/****           Backwards compatibility for old tesselator           ****/

typedef GLUtesselator GLUtriangulatorObj; }

procedure   gluBeginPolygon( tess: GLUtesselator ); stdcall;

procedure   gluNextContour( tess: GLUtesselator;
                                 _type: GLenum); stdcall;

procedure   gluEndPolygon( tess: GLUtesselator );  stdcall;

{ /* Contours types -- obsolete! */
#define GLU_CW          100120
#define GLU_CCW         100121
#define GLU_INTERIOR    100122
#define GLU_EXTERIOR    100123
#define GLU_UNKNOWN     100124

/* Names without "TESS_" prefix */
#define GLU_BEGIN       GLU_TESS_BEGIN
#define GLU_VERTEX      GLU_TESS_VERTEX
#define GLU_END         GLU_TESS_END
#define GLU_ERROR       GLU_TESS_ERROR
#define GLU_EDGE_FLAG   GLU_TESS_EDGE_FLAG
}


const
  glu32 = 'glu32.dll';

Implementation

{ OpenGL Utility routines (glu.h) =======================================}

function gluErrorString;                     external glu32;
function gluErrorUnicodeStringEXT;           external glu32;
function gluGetString;                       external glu32;
procedure gluLookAt;                         external glu32;
procedure gluOrtho2D;                        external glu32;
procedure gluPerspective;                    external glu32;
procedure gluPickMatrix;                     external glu32;
function  gluProject;                        external glu32;
function  gluUnProject;                      external glu32;
function  gluScaleImage;                     external glu32;
function  gluBuild1DMipmaps;                 external glu32;
function  gluBuild2DMipmaps;                 external glu32;
function  gluNewQuadric;                     external glu32;
procedure gluDeleteQuadric;                  external glu32;
procedure gluQuadricNormals;                 external glu32;
procedure gluQuadricTexture;                 external glu32;
procedure gluQuadricOrientation;             external glu32;
procedure gluQuadricDrawStyle;               external glu32;
procedure gluCylinder;                       external glu32;
procedure gluDisk;                           external glu32;
procedure gluPartialDisk;                    external glu32;
procedure gluSphere;                         external glu32;
procedure gluQuadricCallback;                external glu32;

function gluNewTess                         ;external glu32;
procedure gluDeleteTess                     ;external glu32;
procedure gluTessBeginPolygon               ;external glu32;
procedure gluTessBeginContour               ;external glu32;
procedure gluTessVertex                     ;external glu32;
procedure gluTessEndContour                 ;external glu32;
procedure gluTessEndPolygon                 ;external glu32;
procedure gluTessProperty                   ;external glu32;
procedure gluTessNormal                     ;external glu32;
procedure gluTessCallback                   ;external glu32;

function gluNewNurbsRenderer                ;external glu32;
procedure gluDeleteNurbsRenderer            ;external glu32;
procedure gluBeginSurface                   ;external glu32;
procedure gluBeginCurve                     ;external glu32;
procedure gluEndCurve                       ;external glu32;
procedure gluEndSurface                     ;external glu32;
procedure gluBeginTrim                      ;external glu32;
procedure gluEndTrim                        ;external glu32;
procedure gluPwlCurve                       ;external glu32;
procedure gluNurbsCurve                     ;external glu32;
procedure gluNurbsSurface                   ;external glu32;
procedure gluLoadSamplingMatrices           ;external glu32;
procedure gluNurbsProperty                  ;external glu32;
procedure gluGetNurbsProperty               ;external glu32;
procedure gluNurbsCallback                  ;external glu32;
procedure gluDeleteNurbsTessellatorEXT      ;external glu32;
procedure gluGetTessProperty                ;external glu32;
function  gluNewNurbsTessellatorEXT         ;external glu32;
procedure gluBeginPolygon                   ;external glu32;
procedure gluNextContour                    ;external glu32;
procedure gluEndPolygon                     ;external glu32;


begin
 Set8087CW($133F);
end.
