(*************************     TSComboBox97     *******************************

 TSComboBox97 is a Simple Static ComboBox with Office 97 Style
  like a ComboBox with csDropDownList style.
 This component can be freely used and distributed in commercial and private
 environments, provided this notice is not modified in any way.
 Feel free to contact us if you have any questions, comments or suggestions
 at TRSOFT@Menden.net

    Copyright  1998 by TRSOFT  All Rights Reserved.
    Thomas Radtke Software Entwicklung.
    http://www.abcnet.de/TRSOFT/

  THIS SOFTWARE IS PROVIDED AS IS AND WITHOUT WARRANTY OF ANY KIND,
  EITHER EXPRESSED OR IMPLIED.

------------------- History ------------------------

Ver 1.0
     Main Component Build;

Ver 1.1
     Disabled Font Color added.
     Btn3State property added
     Little Bug removed ( If Item was not changed and FPopForm closed
           FButton.Frame.Color was COLOR_WINDOW instead of COLOR_3DHILIGHT.
           
*****************************************************************************)


unit SCombo97;

{$R-}

interface

uses
    SysUtils, Windows, Messages, Classes, Graphics, Controls,
    Forms, Dialogs, StdCtrls, ExtCtrls, Buttons, Menus;

type

// ********************************* TXButton *********************************

  TXButton = class(TGraphicControl)
   private
        FBitmap: TBitmap;
        Pushed : Boolean;
        MouseIn: Boolean;
        procedure CMButtonPressed(var Message: TMessage); message CM_BUTTONPRESSED;
        procedure CMMouseEnter(var msg: TMessage); message CM_MOUSEENTER;
        procedure CMMouseLeave(var msg: TMessage); message CM_MOUSELEAVE;
        procedure BitmapChange(Sender: TObject);
       	procedure SetBitmap(value: TBitmap);
   protected
        procedure Loaded; override;
        procedure Paint; override;
        procedure MouseDown(Button: TMouseButton; Shift: TShiftState;X, Y: Integer); override;
        procedure MouseUp(Button: TMouseButton; Shift: TShiftState;X, Y: Integer); override;
        procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
   public
       	constructor Create(aOwner: TComponent); override;
        destructor Destroy; override;
        procedure Click; override;
   published
       	property Bitmap: TBitmap read FBitmap write SetBitmap stored True;
        property Enabled;
        property ParentShowHint;
        property ShowHint;
        property Visible;
        property OnClick;
        property OnMouseMove;
        property OnMouseDown;
        property OnMouseUp;
   end;

// ***************************** TSComboBox97 *********************************
  TLocator = ( loNone, loFill );

  TPopForm = class(TForm)
   protected
        procedure CreateParams( var Params: TCreateParams ); override;
   end;

  TSComboBox97 = class(TCustomPanel)
   private
        F3State       : Boolean;
        FItemHeight   : Integer;
        FText         : String;
        FDropDownCount: Integer;
        FOnDropDown   : TNotifyEvent;
        FOnDrawItem   : TDrawItemEvent;
        MouseInControl: Boolean;
        FItems        : TStrings;
        FItemIndex    : smallint;
        FBPanel       : TXButton;
        FButton       : TSpeedButton;
        FPanel        : TPanel;
        FPopForm      : TPopForm;
        FListBox      : TListBox;
        FPopped       : boolean;
        FSorted       : Boolean;
        FSelItem      : boolean;
        FOnChange     : TNotifyEvent;
        FLocator      : TLocator;
        FLocColor     : TColor;
        FOldColor     : TColor;
        procedure ButtonDeselect;
        procedure DropList;
        procedure FreeList;
        procedure ButtonClick(Sender: TObject);
        procedure BoxClick(Sender: TObject);
        procedure ListMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
        procedure ListDblClick(Sender: TObject);
        procedure ListKeyPress(Sender: TObject; var Key: Char);
        procedure ListDeAct(Sender : TObject);
        procedure SetItems(Value : TStrings);
        function  GetItemHeight: Integer;
        procedure SetItemHeight(Value: Integer);
   protected
        procedure CreateParams( var Params: TCreateParams ); override;
        procedure CreateWindowHandle(const Params: TCreateParams); override;
        procedure WMSize(var Message: TWMSize); message WM_SIZE;
        procedure WMSetFocus(var Message: TWMSetFocus); message WM_SETFOCUS;
        procedure CMEnabledChanged (var Message: TMessage); message CM_ENABLEDCHANGED;
        procedure CMFontChanged (var Message: TMessage); message CM_FONTCHANGED;
        procedure CMMouseEnter (var Message: TMessage); message CM_MOUSEENTER;
        procedure CMMouseLeave (var Message: TMessage); message CM_MOUSELEAVE;
        procedure WMKillFocus (var Message: TWMKillFocus); message WM_KILLFOCUS;
        procedure WMNCPaint (var Message: TMessage); message WM_NCPAINT;
        procedure RedrawBorder (const Clip: HRGN);
        procedure NewAdjustHeight;
        procedure KeyDown(var key: Word; Shift: TShiftState); override;
        procedure CreateWnd; override;
        procedure Loaded; override;
        procedure DoEnter; override;
        procedure DoExit; override;
        procedure Change;
   public
        constructor Create(AOwner: TComponent); override;
        destructor Destroy; override;
        procedure Clear;
        property Canvas;
   published
        property Items:         TStrings       read FItems         write SetItems;
        property ItemIndex:     Smallint       read FItemIndex     write FItemIndex;
        property Sorted:        Boolean        read FSorted        write FSorted        default False;
        property Locator:       TLocator       read FLocator       write FLocator       default loNone;
        property LocColor:      TColor         read FLocColor      write FLocColor      default clAqua;
        property DropDownCount: Integer        read FDropDownCount write FDropDownCount default 8;
        property OnDropDown:    TNotifyEvent   read FOnDropDown    write FOnDropDown;
        property OnDrawItem:    TDrawItemEvent read FOnDrawItem    write FOnDrawItem;
        property ItemHeight:    Integer        read GetItemHeight  write SetItemHeight;
        property Btn3State:     Boolean        read F3State        write F3State        default False;
        property Alignment;
        property Align;
        property Cursor;
        property ImeMode;
        property ImeName;
        property Text;
        property Color;
        property DragCursor;
        property DragMode;
        property Enabled;
        property Font;
        property ParentColor;
        property ParentCtl3D;
        property ParentFont;
        property ParentShowHint;
        property PopupMenu;
        property ShowHint;
        property TabOrder;
        property TabStop;
        property Visible;
        property OnChange: TNotifyEvent read FOnChange write FOnChange;
        property OnClick;
        property OnDblClick;
        property OnDragDrop;
        property OnDragOver;
        property OnEndDrag;
        property OnEnter;
        property OnExit;
        property OnMouseDown;
        property OnMouseMove;
        property OnMouseUp;
        property OnKeyDown;
   end;

procedure Register;

implementation

procedure Register;
begin
     RegisterComponents('TRSOFT', [TSComboBox97]);
end;

// ****************** TXButton ( Only as Background for SpeedButton ) *******

constructor TXButton.Create(aOwner: TComponent);
begin
     inherited Create(aOwner);
     FBitmap := TBitmap.Create;
     FBitmap.OnChange := BitmapChange;
     ControlStyle := [csCaptureMouse, csOpaque];
     Pushed := False;
     MouseIn := False;
     Width := 25;
     Height := 25;
     Visible:=True;
     Enabled := True;
end;

destructor TXButton.Destroy;
begin
     FBitmap.Destroy;
     inherited Destroy;
end;

procedure TXButton.BitmapChange(Sender: TObject);
begin
     if not FBitmap.Empty then FBitmap.Dormant;
     Repaint;
end;

procedure TXButton.Loaded;
begin
     inherited Loaded;
     if (csDesigning in ComponentState) then Repaint;
end;

procedure TXButton.Paint;
var
   a,b,c,d:Integer;
   ARect, BRect : TRect;
   OffScreen: TBitmap;
begin
     if FBitmap.Height > 0 then
       begin
            a:=FBitmap.Width;
            b:=FBitmap.Height;
            c:=(Width - a)+ a;
            d:=(Height - b)+ b;
            ARect := Rect(a,b,c,d);
       end;


     OffScreen := TBitmap.Create;
     OffScreen.Width := Width;
     OffScreen.Height := Height;
     OffScreen.Canvas.Brush.Color := clBtnFace;
     OffScreen.Canvas.FillRect(Rect(0,0,Width,Height));
     OffScreen.Canvas.Brush.Style := bsClear;

     if not MouseIn then
       begin
	   OffScreen.Canvas.Pen.Color := clBtnHighlight;
        OffScreen.Canvas.MoveTo(0, Height-1);
	   OffScreen.Canvas.LineTo(0, 0);
	   OffScreen.Canvas.LineTo(Width-1, 0);
        OffScreen.Canvas.Pen.Color := clBtnHighlight;
      	OffScreen.Canvas.LineTo(Width-1, Height-1);
	   OffScreen.Canvas.LineTo(0, Height-1);
       end
	else
       if Pushed=False then
         begin
              OffScreen.Canvas.Pen.Color := clBtnHighlight;
              OffScreen.Canvas.MoveTo(0, Height-1);
     	      OffScreen.Canvas.LineTo(0, 0);
	         OffScreen.Canvas.LineTo(Width-1, 0);
              OffScreen.Canvas.Pen.Color := clBtnShadow;
              OffScreen.Canvas.LineTo(Width-1, Height-1);
	         OffScreen.Canvas.LineTo(0, Height-1);
         end
       else
         begin
              OffScreen.Canvas.Pen.Color := clBtnShadow;
              OffScreen.Canvas.MoveTo(0, Height-1);
              OffScreen.Canvas.LineTo(0, 0);
              OffScreen.Canvas.LineTo(Width-1, 0);
              OffScreen.Canvas.Pen.Color := clBtnHighlight;
              OffScreen.Canvas.LineTo(Width-1, Height-1);
              OffScreen.Canvas.LineTo(0, Height-1);
         end;
     Canvas.CopyRect(Rect(0,0,Width,Height), OffScreen.Canvas, Rect(0,0,Width,Height));
     OffScreen.Free;
     if FBitmap.Height > 0 then
       begin
            BRect := Rect(0,0,FBitmap.Width,FBitmap.Height);
            Canvas.CopyRect(ARect, FBitmap.Canvas, BRect);
       end;
end;

procedure TXButton.SetBitmap(value: TBitmap);
begin
	FBitmap.Assign(value);
     if not FBitmap.Empty then FBitmap.Dormant;
end;

procedure TXButton.MouseDown(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
begin
     inherited MouseDown(Button, Shift, X, Y);
     if (Button = mbLeft) and Enabled then
     begin
          if not Pushed then
            begin
                 MouseIn:=True;
                 Pushed:=True;
                 Repaint;
            end;
     end;
end;

procedure TXButton.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
     inherited MouseMove(Shift, X, Y);
end;

procedure TXButton.MouseUp(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
var
   DoClick: Boolean;
begin
     inherited MouseUp(Button, Shift, X, Y);
     begin
          DoClick := (X >= 0) and (X < ClientWidth) and (Y >= 0) and (Y <= ClientHeight);
          MouseIn := True;
          Pushed:=False;
          if DoClick  then Invalidate;
          if DoClick then Click;
    end;
end;

procedure TXButton.CMButtonPressed(var Message: TMessage);
var
   Sender: TXButton;
begin
     if Message.WParam = 0 then
     begin
          Sender := TXButton(Message.LParam);
          if Sender <> Self then
            begin
                 Pushed := False;
                 Invalidate;
            end;
     end;
end;

procedure TXButton.CMMouseEnter(var msg: TMessage);
begin
     MouseIn := True;
     Repaint;
end;

procedure TXButton.CMMouseLeave(var msg: TMessage);
begin
     MouseIn := False;
     Invalidate;
end;

procedure TXButton.Click;
begin
     inherited Click;
end;

// *************************** TSComboBox97 **********************************

procedure TPopForm.CreateParams( var Params: TCreateParams );
begin
     inherited CreateParams( Params );
end;

procedure TSComboBox97.CreateParams( var Params: TCreateParams );
begin
     inherited CreateParams( Params );
     Params.Style := Params.Style or WS_CLIPCHILDREN or WS_TABSTOP
                     or WS_CLIPSIBLINGS or ES_MULTILINE;
end;

procedure TSComboBox97.CreateWindowHandle(const Params: TCreateParams);
begin
     inherited CreateWindowHandle(Params);
end;

constructor TSComboBox97.Create(AOwner: TComponent);
begin
     inherited Create(AOwner);
     ControlStyle:= ControlStyle + [csAcceptsControls,csSetCaption];
     Ctl3D := False;
     Height:= 21;
     Width:=145;
     FPopped:= False;
     F3State:= False;
     FBPanel:=TXButton.Create(Self);
     FButton:= TSpeedButton.Create(Self);
     FItems:= TStringList.Create;
     FItemHeight := Height - 9;
     FItemIndex:= -1;
     TabStop:=True;
     Color:=clWindow;
     FSorted:= False;
     FSelItem:= False;
     FButton.Flat := True;
     ImeMode:=imDontCare;
     Cursor:=crDefault;
     FDropDownCount := 8;
     FLocColor:=clAqua;
     BorderStyle:=bsSingle;
     FText:=Text;
     BevelInner:=bvNone;
     BevelOuter:=bvNone;
     Alignment:=taLeftJustify;
     Align:=alNone;
     OnClick:=BoxClick;
end;

destructor TSComboBox97.Destroy;
begin
     FItems.Free;
     inherited Destroy;
end;

procedure TSComboBox97.Loaded;
begin
     inherited Loaded;
     FOldColor:= Color;
end;

procedure TSComboBox97.CreateWnd;
begin
     inherited CreateWnd;
     FBPanel.Parent:=Self;
     FButton.Parent:= Self;
     FButton.Flat:=True;
     SendMessage(Self.Handle, WM_SIZE, 0, 0);
end;

procedure TSComboBox97.DoEnter;
begin
     if FLocator = loFill then Color:= FLocColor;
     inherited DoEnter;
end;

procedure TSComboBox97.DoExit;
begin
     if FLocator = loFill then Color:= FOldColor;
     inherited DoExit;
end;

procedure TSComboBox97.NewAdjustHeight;
var
   DC: HDC;
   SaveFont: HFONT;
   Metrics: TTextMetric;
begin
     DC := GetDC(0);
     SaveFont := SelectObject(DC, Font.Handle);
     GetTextMetrics (DC, Metrics);
     SelectObject (DC, SaveFont);
     ReleaseDC (0, DC);
     Height := Metrics.tmHeight + 8;
end;

procedure TSComboBox97.RedrawBorder (const Clip: HRGN);
var
   DC: HDC;
   R: TRect;
   NewClipRgn: HRGN;
   BtnFaceBrush, WindowBrush, FrameBrush : HBRUSH;
begin
     DC := GetWindowDC(Handle);
     try
       if Clip <> 0 then
         begin
              GetWindowRect (Handle, R);
              if SelectClipRgn(DC, Clip) = ERROR then
                begin
                     NewClipRgn := CreateRectRgnIndirect(R);
                     SelectClipRgn (DC, NewClipRgn);
                     DeleteObject (NewClipRgn);
                end;
              OffsetClipRgn (DC, -R.Left, -R.Top);
         end;

       GetWindowRect (Handle, R);
       OffsetRect (R, -R.Left, -R.Top);
       BtnFaceBrush := GetSysColorBrush(COLOR_BTNFACE);
       WindowBrush := GetSysColorBrush(COLOR_WINDOW);
       FrameBrush := GetSysColorBrush(COLOR_3DHILIGHT);
       if ((csDesigning in ComponentState) and Enabled) or
          (not(csDesigning in ComponentState)
           and MouseInControl or Focused )
        then
          begin
               DrawEdge (DC, R, BDR_SUNKENOUTER, BF_RECT or BF_ADJUST);
               with R do
                   begin
                        FillRect (DC, Rect(Left, Top, Left+1, Bottom-1), BtnFaceBrush);
                        FillRect (DC, Rect(Left, Top, Right-1, Top+1), BtnFaceBrush);
                   end;
               DrawEdge (DC, R, BDR_SUNKENINNER, BF_BOTTOMRIGHT);
               InflateRect (R, -1, -1);
               FrameRect (DC, R, WindowBrush);
          end
        else
          begin
            FrameRect (DC, R, BtnFaceBrush);
            InflateRect (R, 0, 0);
            FrameRect (DC, R, BtnFaceBrush);
            InflateRect (R, -1, -1);
            FrameRect (DC, R, FrameBrush);
          end;
     finally
       ReleaseDC (Handle, DC);
     end;
end;

procedure TSComboBox97.CMMouseEnter (var Message: TMessage);
begin
     inherited;
     MouseInControl := True;
     FButton.Flat := False;
     RedrawBorder (0);
end;

procedure TSComboBox97.CMMouseLeave (var Message: TMessage);
begin
     inherited;
     MouseInControl := False;
     FButton.Flat := True;
     RedrawBorder (0);
end;

procedure TSComboBox97.CMEnabledChanged (var Message: TMessage);
const
  EnableColors: array[Boolean] of TColor = (clBtnFace, clWindow);
  EnableFontColors :array[Boolean] of TColor = (clGrayText,clBtnText);
var
   OldSize,NewSize:Integer;
begin
     inherited;
     NewSize :=Height;
     OldSize:=Font.Size;
     Font.Color:=EnableFontColors[Enabled];
     Font.Size:=OldSize;
     FBPanel.Visible := Enabled;
     Color := EnableColors[Enabled];
     Height:= NewSize;
end;

procedure TSComboBox97.CMFontChanged (var Message: TMessage);
begin
     inherited;
     if (not((csDesigning in ComponentState) and (csLoading in ComponentState)))
      then NewAdjustHeight;
end;

procedure TSComboBox97.WMKillFocus (var Message: TWMKillFocus);
begin
     inherited;
     if not(csDesigning in ComponentState) then
       begin
            MouseInControl:=False;
            RedrawBorder (0);
            ResetIme;
       end;
end;

procedure TSComboBox97.WMSetFocus(var Message: TWMSetFocus);
begin
     inherited;
     if not(csDesigning in ComponentState) then
       begin
            MouseInControl:=True;
            RedrawBorder (0);
            SetIme;
       end;
end;

procedure TSComboBox97.WMNCPaint (var Message: TMessage);
begin
     inherited;
     RedrawBorder (Message.WParam);
end;

procedure TSComboBox97.WMSize (var Message: TWMSize);
var
   ABMP : TBitmap;
   R: TRect;
begin
     SetItemHeight(Height - 9);
     inherited;
     FBPanel.Height:= Height-4;
     FBPanel.Width:= trunc(FBPanel.Height * 0.80);
     FBPanel.Left:= Width - FBPanel.Width - 4;
     FBPanel.Top:= 0;

     FButton.Height:= Height-4;
     FButton.Width:= trunc(FButton.Height * 0.80);
     FButton.Left:= Width - FButton.Width - 4;
     FButton.Top:= 0;
     FButton.Cursor:= crHandPoint;
     ABMP:= TBitmap.Create;
     with ABMP,Canvas do
         begin
              Width:=9;
              Height:=3;
              Brush.Style:= bsClear;
              Pen.Color:= clBlack;
              Pen.Style:= psSolid;
              Pen.Width:= 1;
              MoveTo(2,0);
              LineTo(7,0);
              MoveTo(3,1);
              LineTo(6,1);
              Pixels[4,2]:= Pen.Color;
        end;
     FButton.Glyph.Assign(ABMP);
     ABMP.Free;
     if F3State then
       begin
            FButton.GroupIndex:=1;
            FButton.AllowAllUp:=False;
       end;
     FButton.OnClick := ButtonClick;
     R:= Rect(0, 0, FButton.Left-2, Height);
     SendMessage(Self.Handle, EM_SETRECT, 0, Longint(@R));
end;

procedure TSComboBox97.SetItems(Value : TStrings);
begin
     FItems.Assign(Value);
end;

procedure TSComboBox97.KeyDown(var Key: Word; Shift: TShiftState);
var
   x:Integer;
begin
     inherited KeyDown(Key,Shift);
     if (Key = VK_DOWN) and (Shift = [ssALT]) then ButtonClick(Self);

     if (Key = VK_DOWN) and (Shift = [])then
       begin
            if FSelItem then
              begin
                   if FItems.Count = 0  then Exit;
                   x:=Fitems.IndexOf(Text);
                   if x > FItems.Count-2 then x:= FItems.Count-2;
                   Text:=FItems.Strings[x + 1]
              end;
            FSelItem:= True;
            SetFocus;
            if Assigned(FOnChange) then FOnChange(Self);
       end;

     if (Key = VK_UP) and (Shift = [])then
       begin
            if FSelItem then
              begin
                   x:=Fitems.IndexOf(Text);
                   if x > 0 then Text:=FItems.Strings[x - 1];
              end;
            FSelItem:= True;
            SetFocus;
            if Assigned(FOnChange) then FOnChange(Self);
       end;
end;

procedure TSComboBox97.ButtonClick(Sender: TObject);
begin
     if (not FPopped) and (FPopForm = NIL) then
       begin
            DropList;
            FSelItem:=False;
       end;
end;

procedure TSComboBox97.BoxClick(Sender: TObject);
begin
     if (not FPopped) and (FPopForm = NIL) then
       begin
            DropList;
            FSelItem:=False;
       end;
end;

procedure TSComboBox97.ButtonDeselect;
begin
     FButton.GroupIndex:=0;
     FButton.AllowAllUp:=True;
     FButton.Down:=False;
end;

procedure TSComboBox97.DropList;
var
   P: TPoint;
   AForm: TForm;
begin
     FPopped:= true;
     FPopForm:= TPopForm.CreateNew(Application);
     P:= Self.ClientToScreen(Point(Left,Top));
     FPopForm.Left:= P.X - Left - 2;
     FPopForm.Top:= P.Y - Top + Height - 1;
     FPopForm.Width:= Width;
     FPopForm.Height:= 80;
     FPopForm.BorderStyle:= bsNone;
     FPopForm.FormStyle:= fsStayOnTop;
     FPopForm.Font.Assign(Font);
     FPopForm.OnDeactivate:= ListDeAct;

     FPanel:=TPanel.Create(FPopForm);
     FPanel.Align:=            alClient;
     FPanel.BevelInner:=       bvNone;
     FPanel.BevelOuter:=       bvLowered;
     FPanel.BorderStyle:=      bsNone;
     FPanel.BevelWidth:=       1;
     FPanel.Parent:=           FPopForm;

     FListBox := TListBox.Create(FPanel);
     FListBox.ParentFont:=      True;
     FListBox.Ctl3D:=           False;
     FListBox.BorderStyle:=     bsNone;
     FListBox.ItemHeight:=      Height - 6;
     FListBox.ExtendedSelect:=  False;
     FListBox.ImeMode:=         ImeMode;
     FListBox.Sorted:=          FSorted;
     FListBox.Align:=           alClient;
     FListBox.MultiSelect:=     False;
     FListBox.Style:=           lbOwnerDrawFixed;
     FListBox.Parent:=          FPanel;
     FListBox.Items.Assign(FItems);
     FListBox.OnMouseMove:=     ListMouseMove;
     FListBox.OnDblClick:=      ListDblClick;
     FListBox.OnKeyPress:=      ListKeyPress;
     if FListBox.Items.Count < FDropDownCount
       then
           FPopForm.Height:= (FListBox.ItemHeight * FListBox.Items.Count) + 4
       else
           FPopForm.Height:= (FListBox.ItemHeight * FDropDownCount) + 4;
     FPopForm.Show;
     if FListBox.Items.count > 0 then FListBox.Itemindex:= 0;
     FListBox.SetFocus;
     AForm:= (GetparentForm(Self) as TForm);
     if AForm <> nil then AForm.Perform(WM_NCACTIVATE, 1, 0);
end;

procedure TSComboBox97.FreeList;
begin
     if FSelItem then
       begin
            if FListBox.ItemIndex > -1
              then
                  begin
                       Text:= FListBox.Items[FListBox.ItemIndex];
                       FItemIndex:= FListBox.ItemIndex;
                  end
            else Text:=FText;
       end;
     FPopForm.Release;
     FPopForm:= NIL;
     if CanFocus then SetFocus;
     FPopped:= false;
     if Assigned(FOnChange) then FOnChange(Self);
     if F3State then
       begin
            ButtonDeselect;
            FButton.GroupIndex:=1;
            FButton.AllowAllUp:=False;
       end;
     Invalidate;  
end;

procedure TSComboBox97.ListMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
var
   CP:TPoint;
   V:Integer;
begin
     inherited;
     CP.X:=X;
     CP.y:=Y;
     V:=FListBox.ItemAtPos(CP,True);
     if V > -1 then FListBox.ItemIndex:=FListBox.ItemAtPos(CP,True);
     FSelItem:=False;
end;

procedure TSComboBox97.ListDblClick(Sender: TObject);
begin
     FSelItem:= true;
     SetFocus;
end;

procedure TSComboBox97.ListKeyPress(Sender: TObject; var Key: Char);
begin
     if ord(Key) = VK_RETURN then
       begin
            FSelItem:= true;
            SetFocus;
       end;
     if ord(Key) = VK_ESCAPE then
       begin
            FSelItem:= false;
            SetFocus;
       end;
end;

procedure TSComboBox97.ListDeAct(Sender : TObject);
begin
     FreeList;
end;

procedure TSComboBox97.Clear;
begin
     SetTextBuf('');
     FItems.Clear;
end;

function TSComboBox97.GetItemHeight: Integer;
begin
     Result := FItemHeight;
end;

procedure TSComboBox97.SetItemHeight(Value: Integer);
begin
     if Value > 0 then FItemHeight := Value;
     Height:=FItemHeight + 9;
     Invalidate;
end;

procedure TSComboBox97.Change;
begin
     inherited Changed;
     if Assigned(FOnChange) then FOnChange(Self);
     Invalidate;
end;

end.
