{ :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
  :: QuickReport 2.0 for Delphi 1.0/2.0/3.0                  ::
  ::                                                         ::
  :: QRBONUS - Bonus classes, only in QR Professional        ::
  ::                                                         ::
  :: Copyright (c) 1997 QuSoft AS                            ::
  :: All Rights Reserved                                     ::
  ::                                                         ::
  :: web: http://www.qusoft.no    mail: support@qusoft.no    ::
  ::                              fax: +47 22 41 74 91       ::
  ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: }

unit QRBonus;

interface

uses
  Sysutils, Classes, Graphics, Controls, Forms, QRPrntr, QRCtrls, QRPrev, QuickRpt;

type
  (***************************************************************************
    TQRPrintJob - TQRPrinter wrapper
    **************************************************************************
    This class works as a wrapper around TQRPrinter, making your life easier
    if you want to use this independently of the rest of QuickReport.
    Override the CreateOutput method to create the actual output and call the
    AQRPrinter.XPos, YPos, XSize and YSize to convert any coordinates from 0.1mm
    to actual device coordinates. The example bellow will write a text at
    a position 2 x 2 cm from the top left corner :

    procedure MyClass.CreateOutput(AQRPrinter : TQRPrinter);
    begin
      AQRPrinter.BeginDoc;
      AQRPrinter.NewPage;
      AQRPrinter.Canvas.TextOut(AQRPrinter.XPos(200), AQRPrinter.YPos(200),'Hello World!');
      AQRPrinter.EndDoc;
    end;

    A cleaner version might look like this:

    procedure MyClass.CreateOutput(AQRPrinter : TQRPrinter);
    begin
      with AQRPrinter do
      begin
        BeginDoc;
        NewPage;
        Canvas.TextOut(XPos(200), YPos(200),'Hello World!');
        EndDoc;
      end;
    end;

    **************************************************************************
    TQRPHandler - .QRP file manager class
    **************************************************************************
    TQRPHandler makes it easy to load, preview and print saved reports in
    .qrp files. The class has one property and two methods:

    property Filename : string
    Set the name of the .qrp file you want to work with, including extension

    procedure Preview
    Previews the file set by Filename

    procedure Print
    Prints the file set by Filename

    Example:

    with TQRPHandler.Create do
    try
      Filename := 'MYFILE.QRP';
      Preview;
    finally
      Free;
    end;

    **************************************************************************
    TQRMerger - String list with embeded expressions 'merger'
    **************************************************************************


    **************************************************************************
    TQRExprMemo - Memo component which can include expressions in the text
    **************************************************************************

    *)

  TQRPrintJob = class
  protected
    procedure CreateOutput(AQRPrinter : TQRPrinter); virtual;
  public
    procedure Preview;
    procedure Print;
  end;

  TQRPHandler = class
  private
    FFilename : string;
    FQRPrinter : TQRPrinter;
  protected
    procedure SetFilename(Value : string);
  public
    constructor Create;
    destructor Destroy; override;
    procedure Preview;
    procedure Print;
    property Filename : string read FFilename write SetFilename;
  end;

  TQRMerger = class
  private
    FOrgLines : TStrings;
    FStrippedLines : TStrings;
    FMergedLines : TStrings;
    FMerged : boolean;
    FPrepared : boolean;
    Expressions : TList;
    FDataSets : TList;
  protected
    function GetOrgLines : TStrings;
    function GetMergedLines : TStrings;
    procedure SetOrgLines(Value : TStrings);
  public
    constructor Create;
    destructor Destroy; override;
    procedure Prepare;
    procedure Merge;
    procedure Unprepare;
    property Lines : TStrings read GetOrgLines write SetOrgLines;
    property MergedLines : TStrings read GetMergedLines;
    property Merged : boolean read FMerged;
    property Prepared : boolean read FPrepared;
    property DataSets :TList read FDataSets write FDataSets;
  end;

  TQRExprMemo = class(TQRCustomLabel)
  private
    Merger : TQRMerger;
    FRemoveBlankLines : boolean;
  protected
    procedure Prepare; override;
    procedure Unprepare; override;
    procedure Print(OfsX, OfsY : integer); override;
    procedure Paint; override;
  public
    constructor Create(AOwner : TComponent); override;
  published
    property RemoveBlankLines : boolean read FRemoveBlankLines write FRemoveBlankLines;
    property Alignment;
    property AlignToBand;
    property AutoSize;
    property AutoStretch;
    property Color;
    property Font;
    property Lines;
    property ParentFont;
    property Transparent;
    property WordWrap;
  end;

{$ifndef ver100}
procedure Register;
{$endif}

implementation

procedure TQRPrintJob.CreateOutput(AQRPrinter : TQRPrinter);
begin
end;

procedure TQRPrintJob.Preview;
var
  aQRPrinter : TQRPrinter;
begin
  aQRPrinter := TQRPrinter.Create;
  with aQRPrinter do
  try
    Destination := qrdMetafile;
    OnGenerateToPrinter := Self.Print;
    Preview;
    Application.ProcessMessages;
    CreateOutput(aQRPrinter);
    repeat
      Application.ProcessMessages
    until not aQRPrinter.ShowingPreview;
  finally
    Free;
  end;
end;

procedure TQRPrintJob.Print;
var
  aQRPrinter : TQRPrinter;
begin
  aQRPrinter := TQRPrinter.Create;
  with aQRPrinter do
  try
    Destination := qrdPrinter;
    CreateOutput(aQRPrinter);
  finally
    Free;
  end;
end;

constructor TQRPHandler.Create;
begin
  FFilename := '';
  FQRPrinter := nil;
end;

destructor TQRPHandler.Destroy;
begin
  if assigned(FQRPrinter) then
    FQRPrinter.Free;
  inherited Destroy;
end;

procedure TQRPHandler.SetFilename(Value : string);
begin
  if FFilename <> Value then
  begin
    if assigned(FQRPrinter) then FQRPrinter.Free;
    FQRPrinter := TQRPrinter.Create;
    FFilename := Value;
    FQRPrinter.Load(Filename);
  end;
end;

procedure TQRPHandler.Preview;
begin
  if assigned(FQRPrinter) then
  begin
    with TQRStandardPreview.CreatePreview(Application, FQRPrinter) do
    begin
      Show;
      QRPreview.ZoomToWidth;
    end;
    repeat
      Application.ProcessMessages
    until not FQRPrinter.ShowingPreview;
  end;
end;

procedure TQRPHandler.Print;
begin
  if assigned(FQRPrinter) then FQRPrinter.Print;
end;

{ TQRMerger class }

type
  TMemoEvaluator = class(TQREvaluator)
  public
    Line : integer;
    Position : integer;
  end;

constructor TQRMerger.Create;
begin
  FOrgLines := TStringList.Create;
  FMergedLines := nil;
  FMerged := false;
  FPrepared := false;
  Expressions := nil;
  FDataSets := TList.Create;
end;

destructor TQRMerger.Destroy;
begin
  if Prepared then Unprepare;
  if FOrgLines <> nil then
    FOrgLines.Free;
  if FDataSets <> nil then
    FDataSets.Free;
  inherited Destroy;
end;

function TQRMerger.GetOrgLines : TStrings;
begin
  Result := FOrgLines;
end;

function TQRMerger.GetMergedLines : TStrings;
begin
  if Merged then
    Result := FMergedLines
  else
    Result := nil;
end;

procedure TQRMerger.SetOrgLines(Value : TStrings);
begin
  if FOrgLines <> nil then
    FOrgLines.Free;
  FOrgLines := Value;
end;

procedure TQRMerger.Prepare;
var
  I, start, stop, Len : integer;
  Expr : String;
  aLine : string;
  aEvaluator : TMemoEvaluator;
begin
  if Prepared then Unprepare;
  Expressions := TList.Create;
  FMergedLines := TStringList.Create;
  if Lines.Count > 0 then
  begin
    FStrippedLines := TStringList.Create;
    try
      for I := 0 to Lines.Count - 1 do
      begin
        aLine := FOrgLines[I];
        Start := Pos('{', aLine);
        while Start > 0 do
        begin
          stop := pos('}', aLine);
          Len := Stop - Start - 1;
          if Len > 0 then
          begin
            Expr := copy(aLine, start + 1, Len);
            Delete(aLine, Start, Len + 2);
            aEvaluator := TMemoEvaluator.Create;
            aEvaluator.DataSets := DataSets;
            aEvaluator.Prepare(Expr);
            aEvaluator.Line := I;
            aEvaluator.Position := Start;
            Expressions.Add(aEvaluator);
          end;
          Start := Pos('{', aLine);
        end;
        FStrippedLines.Add(aLine);
      end;
    finally
      FPrepared := true;
    end;
  end;
end;

procedure TQRMerger.Unprepare;
var
  I : integer;
begin
  if Prepared then
  try
    for I := 0 to Expressions.Count - 1 do
      TMemoEvaluator(Expressions[I]).Free;
    FStrippedLines.Free;
  finally
    FPrepared := false;
    FMerged := false;
  end;
  FMergedLines.Free;
  Expressions.Free;
end;

procedure TQRMerger.Merge;
var
  I : integer;
  aResult : TQREvResult;
  Replacement : string;
  aLine : string;
begin
  if (Expressions.Count > 0) then
  begin
    if Merged then
      FMergedLines.Clear;
    FMergedLines.Assign(FStrippedLines);
    for I := Expressions.Count - 1 downto 0 do
    begin
      with TMemoEvaluator(Expressions[I]) do
      begin
        aResult := Value;
        case aResult.Kind of
          resInt : Replacement := IntToStr(aResult.IntResult);
          resDouble : Replacement := FloatToStr(aResult.dblResult);
          resString : Replacement := aResult.StrResult;
          resError : Replacement := '';
        end;
        aLine := FMergedLines[Line];
        Insert(Replacement, aLine, Position);
        FMergedLines[Line] := aLine;
      end;
    end
  end else
    if not Merged then
      FMergedLines.Assign(FOrgLines);
  FMerged := true;
end;

{ TQRExrMemo component }

constructor TQRExprMemo.Create(AOwner : TComponent);
begin
  inherited Create(AOwner);
  FRemoveBlankLines := false;
end;

procedure TQRExprMemo.Prepare;
begin
  Merger := TQRMerger.Create;
  Merger.Lines.Assign(Lines);
  Merger.Datasets.Free;  { Free the object already there }
  Merger.Datasets := ParentReport.AllDatasets;
  Merger.Prepare;
  inherited Prepare;
end;

procedure TQRExprMemo.Unprepare;
begin
  Lines.Assign(Merger.Lines);
  Merger.DataSets := nil;
  Merger.Unprepare;
  Merger.Free;
  inherited Unprepare;
end;

procedure TQRExprMemo.Print(OfsX, OfsY : integer);
var
  I, J : integer;
begin
  Caption := '';
  Merger.Merge;
  if Merger.Prepared then
  begin
    J := Merger.MergedLines.Count - 1;
    I := 0;
    while I <= J do
    begin
      if Merger.MergedLines[I] = '' then
      begin
        Merger.MergedLines.Delete(I);
        dec(J);
      end;
      inc(I);
    end;
    Lines := Merger.MergedLines;
    if (Lines.Count > 0) and (Caption > '') then
      Caption := '';
  end;
  inherited Print(OfsX, OfsY);
  Caption := Name;
end;

procedure TQRExprMemo.Paint;
begin
  if (Lines.Count > 0) and (Caption > '') then
    Caption := '';
  inherited Paint;
end;

{$ifndef ver100}
procedure Register;
begin
  RegisterComponents('QReport', [TQRExprMemo]);
end;
{$endif}

end.
